/**
* Copyright (C) 2021 Xilinx, Inc
*
* Licensed under the Apache License, Version 2.0 (the "License"). You may
* not use this file except in compliance with the License. A copy of the
* License is located at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
* WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
* License for the specific language governing permissions and limitations
* under the License.
*/

#include "adf_api_message.h"
#include <iostream>
#include "core/common/error.h"

adf::err_code errorMsg(adf::err_code code, std::string msg)
{
    throw xrt_core::error(-((int)code), msg);
    return code;
}

void debugMsg(std::string msg)
{
}

void infoMsg(std::string msg)
{
}
