// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: rpc_messages.proto

#include "rpc_messages.pb.h"

#include <algorithm>
#include "google/protobuf/io/coded_stream.h"
#include "google/protobuf/extension_set.h"
#include "google/protobuf/wire_format_lite.h"
#include "google/protobuf/descriptor.h"
#include "google/protobuf/generated_message_reflection.h"
#include "google/protobuf/reflection_ops.h"
#include "google/protobuf/wire_format.h"
#include "google/protobuf/generated_message_tctable_impl.h"
// @@protoc_insertion_point(includes)

// Must be included last.
#include "google/protobuf/port_def.inc"
PROTOBUF_PRAGMA_INIT_SEG
namespace _pb = ::google::protobuf;
namespace _pbi = ::google::protobuf::internal;
namespace _fl = ::google::protobuf::internal::field_layout;

inline constexpr xclWriteQueue_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        written_size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteQueue_response::xclWriteQueue_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteQueue_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteQueue_responseDefaultTypeInternal() {}
  union {
    xclWriteQueue_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteQueue_responseDefaultTypeInternal _xclWriteQueue_response_default_instance_;

inline constexpr xclWriteQueue_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        src_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        q_handle_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        req_{::uint64_t{0u}},
        nonblocking_{false},
        eot_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteQueue_call::xclWriteQueue_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteQueue_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteQueue_callDefaultTypeInternal() {}
  union {
    xclWriteQueue_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteQueue_callDefaultTypeInternal _xclWriteQueue_call_default_instance_;

inline constexpr xclWriteAddrSpaceDeviceRam_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal() {}
  union {
    xclWriteAddrSpaceDeviceRam_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrSpaceDeviceRam_responseDefaultTypeInternal _xclWriteAddrSpaceDeviceRam_response_default_instance_;

inline constexpr xclWriteAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal() {}
  union {
    xclWriteAddrSpaceDeviceRam_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrSpaceDeviceRam_callDefaultTypeInternal _xclWriteAddrSpaceDeviceRam_call_default_instance_;

inline constexpr xclWriteAddrKernelCtrl_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteAddrKernelCtrl_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_responseDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_responseDefaultTypeInternal _xclWriteAddrKernelCtrl_response_default_instance_;

inline constexpr xclWriteAddrKernelCtrl_call_kernelInfo::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_call_kernelInfo _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_call_kernelInfoDefaultTypeInternal _xclWriteAddrKernelCtrl_call_kernelInfo_default_instance_;

inline constexpr xclSyncBOAIENB_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclSyncBOAIENB_response::xclSyncBOAIENB_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSyncBOAIENB_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSyncBOAIENB_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSyncBOAIENB_responseDefaultTypeInternal() {}
  union {
    xclSyncBOAIENB_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSyncBOAIENB_responseDefaultTypeInternal _xclSyncBOAIENB_response_default_instance_;

inline constexpr xclSyncBOAIENB_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gmioname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}},
        offset_{::uint64_t{0u}},
        boh_{::uint64_t{0u}},
        dir_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclSyncBOAIENB_call::xclSyncBOAIENB_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSyncBOAIENB_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSyncBOAIENB_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSyncBOAIENB_callDefaultTypeInternal() {}
  union {
    xclSyncBOAIENB_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSyncBOAIENB_callDefaultTypeInternal _xclSyncBOAIENB_call_default_instance_;

inline constexpr xclSlaveWriteReq_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclSlaveWriteReq_response::xclSlaveWriteReq_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSlaveWriteReq_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveWriteReq_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveWriteReq_responseDefaultTypeInternal() {}
  union {
    xclSlaveWriteReq_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveWriteReq_responseDefaultTypeInternal _xclSlaveWriteReq_response_default_instance_;

inline constexpr xclSlaveWriteReq_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclSlaveWriteReq_call::xclSlaveWriteReq_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSlaveWriteReq_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveWriteReq_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveWriteReq_callDefaultTypeInternal() {}
  union {
    xclSlaveWriteReq_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveWriteReq_callDefaultTypeInternal _xclSlaveWriteReq_call_default_instance_;

inline constexpr xclSlaveReadReq_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclSlaveReadReq_response::xclSlaveReadReq_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSlaveReadReq_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveReadReq_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveReadReq_responseDefaultTypeInternal() {}
  union {
    xclSlaveReadReq_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveReadReq_responseDefaultTypeInternal _xclSlaveReadReq_response_default_instance_;

inline constexpr xclSlaveReadReq_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        addr_{::uint64_t{0u}},
        size_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclSlaveReadReq_call::xclSlaveReadReq_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSlaveReadReq_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSlaveReadReq_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSlaveReadReq_callDefaultTypeInternal() {}
  union {
    xclSlaveReadReq_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSlaveReadReq_callDefaultTypeInternal _xclSlaveReadReq_call_default_instance_;

inline constexpr xclSetupInstance_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        success_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclSetupInstance_response::xclSetupInstance_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetupInstance_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_responseDefaultTypeInternal() {}
  union {
    xclSetupInstance_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_responseDefaultTypeInternal _xclSetupInstance_response_default_instance_;

inline constexpr xclSetupInstance_call_argflowpair::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        tag_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        arg_{::uint64_t{0u}},
        flow_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetupInstance_call_argflowpairDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_call_argflowpairDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_call_argflowpairDefaultTypeInternal() {}
  union {
    xclSetupInstance_call_argflowpair _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_call_argflowpairDefaultTypeInternal _xclSetupInstance_call_argflowpair_default_instance_;

inline constexpr xclSetEnvironment_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclSetEnvironment_response::xclSetEnvironment_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetEnvironment_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_responseDefaultTypeInternal() {}
  union {
    xclSetEnvironment_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_responseDefaultTypeInternal _xclSetEnvironment_response_default_instance_;

inline constexpr xclSetEnvironment_call_namevaluepair::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        name_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        value_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
PROTOBUF_CONSTEXPR xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetEnvironment_call_namevaluepairDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_call_namevaluepairDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_call_namevaluepairDefaultTypeInternal() {}
  union {
    xclSetEnvironment_call_namevaluepair _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_call_namevaluepairDefaultTypeInternal _xclSetEnvironment_call_namevaluepair_default_instance_;

inline constexpr xclRegWrite_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclRegWrite_response::xclRegWrite_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclRegWrite_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegWrite_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegWrite_responseDefaultTypeInternal() {}
  union {
    xclRegWrite_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegWrite_responseDefaultTypeInternal _xclRegWrite_response_default_instance_;

inline constexpr xclRegWrite_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        baseaddress_{0u},
        offset_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclRegWrite_call::xclRegWrite_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclRegWrite_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegWrite_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegWrite_callDefaultTypeInternal() {}
  union {
    xclRegWrite_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegWrite_callDefaultTypeInternal _xclRegWrite_call_default_instance_;

inline constexpr xclRegRead_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclRegRead_response::xclRegRead_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclRegRead_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegRead_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegRead_responseDefaultTypeInternal() {}
  union {
    xclRegRead_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegRead_responseDefaultTypeInternal _xclRegRead_response_default_instance_;

inline constexpr xclRegRead_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        baseaddress_{0u},
        offset_{0u},
        size_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclRegRead_call::xclRegRead_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclRegRead_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclRegRead_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclRegRead_callDefaultTypeInternal() {}
  union {
    xclRegRead_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclRegRead_callDefaultTypeInternal _xclRegRead_call_default_instance_;

inline constexpr xclReadQueue_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dest_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadQueue_response::xclReadQueue_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadQueue_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadQueue_responseDefaultTypeInternal() {}
  union {
    xclReadQueue_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadQueue_responseDefaultTypeInternal _xclReadQueue_response_default_instance_;

inline constexpr xclReadQueue_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dest_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        q_handle_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        req_{::uint64_t{0u}},
        nonblocking_{false},
        eot_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadQueue_call::xclReadQueue_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadQueue_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadQueue_callDefaultTypeInternal() {}
  union {
    xclReadQueue_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadQueue_callDefaultTypeInternal _xclReadQueue_call_default_instance_;

inline constexpr xclReadBusStatus_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        uncomplete_rd_cycles_{::uint64_t{0u}},
        uncomplete_wr_cycles_{::uint64_t{0u}},
        idle_bus_cycles_{::uint64_t{0u}},
        uncomplete_rd_id_{::uint64_t{0u}},
        uncomplete_wr_id_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadBusStatus_response::xclReadBusStatus_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadBusStatus_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadBusStatus_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadBusStatus_responseDefaultTypeInternal() {}
  union {
    xclReadBusStatus_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadBusStatus_responseDefaultTypeInternal _xclReadBusStatus_response_default_instance_;

inline constexpr xclReadBusStatus_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slot_n_{0} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadBusStatus_call::xclReadBusStatus_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadBusStatus_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadBusStatus_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadBusStatus_callDefaultTypeInternal() {}
  union {
    xclReadBusStatus_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadBusStatus_callDefaultTypeInternal _xclReadBusStatus_call_default_instance_;

inline constexpr xclReadAddrSpaceDeviceRam_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal() {}
  union {
    xclReadAddrSpaceDeviceRam_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrSpaceDeviceRam_responseDefaultTypeInternal _xclReadAddrSpaceDeviceRam_response_default_instance_;

inline constexpr xclReadAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadAddrSpaceDeviceRam_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrSpaceDeviceRam_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrSpaceDeviceRam_callDefaultTypeInternal() {}
  union {
    xclReadAddrSpaceDeviceRam_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrSpaceDeviceRam_callDefaultTypeInternal _xclReadAddrSpaceDeviceRam_call_default_instance_;

inline constexpr xclReadAddrKernelCtrl_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadAddrKernelCtrl_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrKernelCtrl_responseDefaultTypeInternal() {}
  union {
    xclReadAddrKernelCtrl_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrKernelCtrl_responseDefaultTypeInternal _xclReadAddrKernelCtrl_response_default_instance_;

inline constexpr xclReadAddrKernelCtrl_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclReadAddrKernelCtrl_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclReadAddrKernelCtrl_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclReadAddrKernelCtrl_callDefaultTypeInternal() {}
  union {
    xclReadAddrKernelCtrl_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclReadAddrKernelCtrl_callDefaultTypeInternal _xclReadAddrKernelCtrl_call_default_instance_;

inline constexpr xclPollCompletion_response_request::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dest_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclPollCompletion_response_request::xclPollCompletion_response_request(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPollCompletion_response_requestDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_response_requestDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_response_requestDefaultTypeInternal() {}
  union {
    xclPollCompletion_response_request _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_response_requestDefaultTypeInternal _xclPollCompletion_response_request_default_instance_;

inline constexpr xclPollCompletion_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        req_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclPollCompletion_call::xclPollCompletion_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPollCompletion_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_callDefaultTypeInternal() {}
  union {
    xclPollCompletion_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_callDefaultTypeInternal _xclPollCompletion_call_default_instance_;

inline constexpr xclPerfMonReadTrace_response_events::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        awlen_{0u},
        arlen_{0u},
        eventflags_{0u},
        timestamp_{0u},
        host_timestamp_{::uint64_t{0u}},
        rd_bytes_{0u},
        wr_bytes_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_response_eventsDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response_eventsDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_response_eventsDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_response_events _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_response_eventsDefaultTypeInternal _xclPerfMonReadTrace_response_events_default_instance_;

inline constexpr xclPerfMonReadTrace_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slotname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ack_{false},
        accel_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_callDefaultTypeInternal _xclPerfMonReadTrace_call_default_instance_;

inline constexpr xclPerfMonReadTrace_Streaming_response_events::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        eventflags_{0u},
        timestamp_{0u},
        host_timestamp_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_response_events _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_response_eventsDefaultTypeInternal _xclPerfMonReadTrace_Streaming_response_events_default_instance_;

inline constexpr xclPerfMonReadTrace_Streaming_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slotname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_Streaming_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_callDefaultTypeInternal _xclPerfMonReadTrace_Streaming_call_default_instance_;

inline constexpr xclPerfMonReadCounters_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        wr_byte_count_{0u},
        wr_trans_count_{0u},
        total_wr_latency_{0u},
        rd_byte_count_{0u},
        rd_trans_count_{0u},
        total_rd_latency_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadCounters_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_responseDefaultTypeInternal _xclPerfMonReadCounters_response_default_instance_;

inline constexpr xclPerfMonReadCounters_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slotname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        accel_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadCounters_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_callDefaultTypeInternal _xclPerfMonReadCounters_call_default_instance_;

inline constexpr xclPerfMonReadCounters_Streaming_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        str_num_tranx_{::uint64_t{0u}},
        str_data_bytes_{::uint64_t{0u}},
        str_busy_cycles_{::uint64_t{0u}},
        str_stall_cycles_{::uint64_t{0u}},
        str_starve_cycles_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_Streaming_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_Streaming_responseDefaultTypeInternal _xclPerfMonReadCounters_Streaming_response_default_instance_;

inline constexpr xclPerfMonReadCounters_Streaming_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slotname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadCounters_Streaming_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadCounters_Streaming_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadCounters_Streaming_callDefaultTypeInternal() {}
  union {
    xclPerfMonReadCounters_Streaming_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadCounters_Streaming_callDefaultTypeInternal _xclPerfMonReadCounters_Streaming_call_default_instance_;

inline constexpr xclPerfMonGetTraceCount_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        no_of_samples_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonGetTraceCount_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonGetTraceCount_responseDefaultTypeInternal() {}
  union {
    xclPerfMonGetTraceCount_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonGetTraceCount_responseDefaultTypeInternal _xclPerfMonGetTraceCount_response_default_instance_;

inline constexpr xclPerfMonGetTraceCount_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        slotname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ack_{false},
        accel_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonGetTraceCount_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonGetTraceCount_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonGetTraceCount_callDefaultTypeInternal() {}
  union {
    xclPerfMonGetTraceCount_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonGetTraceCount_callDefaultTypeInternal _xclPerfMonGetTraceCount_call_default_instance_;

inline constexpr xclLoadXclbinContent_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclLoadXclbinContent_response::xclLoadXclbinContent_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclLoadXclbinContent_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadXclbinContent_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadXclbinContent_responseDefaultTypeInternal() {}
  union {
    xclLoadXclbinContent_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadXclbinContent_responseDefaultTypeInternal _xclLoadXclbinContent_response_default_instance_;

inline constexpr xclLoadXclbinContent_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        xmlbuff_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        sharedbin_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        emuldata_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        xmlbuffsize_{::uint64_t{0u}},
        sharedbinsize_{::uint64_t{0u}},
        emuldatasize_{::uint64_t{0u}},
        keepdir_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclLoadXclbinContent_call::xclLoadXclbinContent_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclLoadXclbinContent_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadXclbinContent_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadXclbinContent_callDefaultTypeInternal() {}
  union {
    xclLoadXclbinContent_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadXclbinContent_callDefaultTypeInternal _xclLoadXclbinContent_call_default_instance_;

inline constexpr xclLoadBitstream_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclLoadBitstream_response::xclLoadBitstream_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclLoadBitstream_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_responseDefaultTypeInternal() {}
  union {
    xclLoadBitstream_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_responseDefaultTypeInternal _xclLoadBitstream_response_default_instance_;

inline constexpr xclLoadBitstream_call_ddrbank::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclLoadBitstream_call_ddrbankDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_call_ddrbankDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_call_ddrbankDefaultTypeInternal() {}
  union {
    xclLoadBitstream_call_ddrbank _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_call_ddrbankDefaultTypeInternal _xclLoadBitstream_call_ddrbank_default_instance_;

inline constexpr xclInterruptOccured_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclInterruptOccured_response::xclInterruptOccured_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclInterruptOccured_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclInterruptOccured_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclInterruptOccured_responseDefaultTypeInternal() {}
  union {
    xclInterruptOccured_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclInterruptOccured_responseDefaultTypeInternal _xclInterruptOccured_response_default_instance_;

inline constexpr xclInterruptOccured_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        interrupt_line_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclInterruptOccured_call::xclInterruptOccured_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclInterruptOccured_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclInterruptOccured_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclInterruptOccured_callDefaultTypeInternal() {}
  union {
    xclInterruptOccured_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclInterruptOccured_callDefaultTypeInternal _xclInterruptOccured_call_default_instance_;

inline constexpr xclImportBO_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclImportBO_response::xclImportBO_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclImportBO_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclImportBO_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclImportBO_responseDefaultTypeInternal() {}
  union {
    xclImportBO_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclImportBO_responseDefaultTypeInternal _xclImportBO_response_default_instance_;

inline constexpr xclImportBO_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dst_filename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        offset_{::uint64_t{0u}},
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclImportBO_call::xclImportBO_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclImportBO_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclImportBO_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclImportBO_callDefaultTypeInternal() {}
  union {
    xclImportBO_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclImportBO_callDefaultTypeInternal _xclImportBO_call_default_instance_;

inline constexpr xclGraphWait_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphWait_response::xclGraphWait_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphWait_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphWait_responseDefaultTypeInternal() {}
  union {
    xclGraphWait_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphWait_responseDefaultTypeInternal _xclGraphWait_response_default_instance_;

inline constexpr xclGraphWait_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphWait_call::xclGraphWait_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphWait_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphWait_callDefaultTypeInternal() {}
  union {
    xclGraphWait_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphWait_callDefaultTypeInternal _xclGraphWait_call_default_instance_;

inline constexpr xclGraphUpdateRTP_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphUpdateRTP_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphUpdateRTP_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphUpdateRTP_responseDefaultTypeInternal() {}
  union {
    xclGraphUpdateRTP_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphUpdateRTP_responseDefaultTypeInternal _xclGraphUpdateRTP_response_default_instance_;

inline constexpr xclGraphUpdateRTP_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        portname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        buffer_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphUpdateRTP_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphUpdateRTP_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphUpdateRTP_callDefaultTypeInternal() {}
  union {
    xclGraphUpdateRTP_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphUpdateRTP_callDefaultTypeInternal _xclGraphUpdateRTP_call_default_instance_;

inline constexpr xclGraphTimedWait_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphTimedWait_response::xclGraphTimedWait_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphTimedWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedWait_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedWait_responseDefaultTypeInternal() {}
  union {
    xclGraphTimedWait_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedWait_responseDefaultTypeInternal _xclGraphTimedWait_response_default_instance_;

inline constexpr xclGraphTimedWait_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        cycletimeout_{::uint64_t{0u}},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphTimedWait_call::xclGraphTimedWait_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphTimedWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedWait_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedWait_callDefaultTypeInternal() {}
  union {
    xclGraphTimedWait_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedWait_callDefaultTypeInternal _xclGraphTimedWait_call_default_instance_;

inline constexpr xclGraphTimedEnd_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphTimedEnd_response::xclGraphTimedEnd_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphTimedEnd_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedEnd_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedEnd_responseDefaultTypeInternal() {}
  union {
    xclGraphTimedEnd_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedEnd_responseDefaultTypeInternal _xclGraphTimedEnd_response_default_instance_;

inline constexpr xclGraphTimedEnd_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        cycletimeout_{::uint64_t{0u}},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphTimedEnd_call::xclGraphTimedEnd_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphTimedEnd_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphTimedEnd_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphTimedEnd_callDefaultTypeInternal() {}
  union {
    xclGraphTimedEnd_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphTimedEnd_callDefaultTypeInternal _xclGraphTimedEnd_call_default_instance_;

inline constexpr xclGraphRun_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphRun_response::xclGraphRun_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphRun_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphRun_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphRun_responseDefaultTypeInternal() {}
  union {
    xclGraphRun_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphRun_responseDefaultTypeInternal _xclGraphRun_response_default_instance_;

inline constexpr xclGraphRun_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gh_{0u},
        iterations_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphRun_call::xclGraphRun_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphRun_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphRun_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphRun_callDefaultTypeInternal() {}
  union {
    xclGraphRun_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphRun_callDefaultTypeInternal _xclGraphRun_call_default_instance_;

inline constexpr xclGraphResume_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphResume_response::xclGraphResume_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphResume_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphResume_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphResume_responseDefaultTypeInternal() {}
  union {
    xclGraphResume_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphResume_responseDefaultTypeInternal _xclGraphResume_response_default_instance_;

inline constexpr xclGraphResume_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphResume_call::xclGraphResume_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphResume_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphResume_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphResume_callDefaultTypeInternal() {}
  union {
    xclGraphResume_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphResume_callDefaultTypeInternal _xclGraphResume_call_default_instance_;

inline constexpr xclGraphReadRTP_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        buffer_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphReadRTP_response::xclGraphReadRTP_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphReadRTP_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphReadRTP_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphReadRTP_responseDefaultTypeInternal() {}
  union {
    xclGraphReadRTP_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphReadRTP_responseDefaultTypeInternal _xclGraphReadRTP_response_default_instance_;

inline constexpr xclGraphReadRTP_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        portname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        buffer_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphReadRTP_call::xclGraphReadRTP_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphReadRTP_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphReadRTP_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphReadRTP_callDefaultTypeInternal() {}
  union {
    xclGraphReadRTP_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphReadRTP_callDefaultTypeInternal _xclGraphReadRTP_call_default_instance_;

inline constexpr xclGraphInit_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphInit_response::xclGraphInit_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphInit_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphInit_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphInit_responseDefaultTypeInternal() {}
  union {
    xclGraphInit_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphInit_responseDefaultTypeInternal _xclGraphInit_response_default_instance_;

inline constexpr xclGraphInit_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        graphname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphInit_call::xclGraphInit_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphInit_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphInit_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphInit_callDefaultTypeInternal() {}
  union {
    xclGraphInit_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphInit_callDefaultTypeInternal _xclGraphInit_call_default_instance_;

inline constexpr xclGraphEnd_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphEnd_response::xclGraphEnd_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphEnd_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphEnd_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphEnd_responseDefaultTypeInternal() {}
  union {
    xclGraphEnd_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphEnd_responseDefaultTypeInternal _xclGraphEnd_response_default_instance_;

inline constexpr xclGraphEnd_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gh_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclGraphEnd_call::xclGraphEnd_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGraphEnd_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGraphEnd_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGraphEnd_callDefaultTypeInternal() {}
  union {
    xclGraphEnd_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGraphEnd_callDefaultTypeInternal _xclGraphEnd_call_default_instance_;

inline constexpr xclGetDeviceTimestamp_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        device_timestamp_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGetDeviceTimestamp_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDeviceTimestamp_responseDefaultTypeInternal() {}
  union {
    xclGetDeviceTimestamp_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDeviceTimestamp_responseDefaultTypeInternal _xclGetDeviceTimestamp_response_default_instance_;

inline constexpr xclGetDeviceTimestamp_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGetDeviceTimestamp_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDeviceTimestamp_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDeviceTimestamp_callDefaultTypeInternal() {}
  union {
    xclGetDeviceTimestamp_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDeviceTimestamp_callDefaultTypeInternal _xclGetDeviceTimestamp_call_default_instance_;

inline constexpr xclGetDebugMessages_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        display_msgs_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        log_msgs_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        stop_msgs_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
PROTOBUF_CONSTEXPR xclGetDebugMessages_response::xclGetDebugMessages_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGetDebugMessages_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDebugMessages_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDebugMessages_responseDefaultTypeInternal() {}
  union {
    xclGetDebugMessages_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDebugMessages_responseDefaultTypeInternal _xclGetDebugMessages_response_default_instance_;

inline constexpr xclGetDebugMessages_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false},
        force_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGetDebugMessages_call::xclGetDebugMessages_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGetDebugMessages_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGetDebugMessages_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGetDebugMessages_callDefaultTypeInternal() {}
  union {
    xclGetDebugMessages_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGetDebugMessages_callDefaultTypeInternal _xclGetDebugMessages_call_default_instance_;

inline constexpr xclGMIOWait_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclGMIOWait_response::xclGMIOWait_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGMIOWait_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGMIOWait_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGMIOWait_responseDefaultTypeInternal() {}
  union {
    xclGMIOWait_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGMIOWait_responseDefaultTypeInternal _xclGMIOWait_response_default_instance_;

inline constexpr xclGMIOWait_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        gmioname_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
PROTOBUF_CONSTEXPR xclGMIOWait_call::xclGMIOWait_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclGMIOWait_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclGMIOWait_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclGMIOWait_callDefaultTypeInternal() {}
  union {
    xclGMIOWait_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclGMIOWait_callDefaultTypeInternal _xclGMIOWait_call_default_instance_;

inline constexpr xclFreeDeviceBuffer_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclFreeDeviceBuffer_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclFreeDeviceBuffer_responseDefaultTypeInternal() {}
  union {
    xclFreeDeviceBuffer_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclFreeDeviceBuffer_responseDefaultTypeInternal _xclFreeDeviceBuffer_response_default_instance_;

inline constexpr xclFreeDeviceBuffer_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ddraddress_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclFreeDeviceBuffer_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclFreeDeviceBuffer_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclFreeDeviceBuffer_callDefaultTypeInternal() {}
  union {
    xclFreeDeviceBuffer_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclFreeDeviceBuffer_callDefaultTypeInternal _xclFreeDeviceBuffer_call_default_instance_;

inline constexpr xclDestroyQueue_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        success_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclDestroyQueue_response::xclDestroyQueue_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclDestroyQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclDestroyQueue_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclDestroyQueue_responseDefaultTypeInternal() {}
  union {
    xclDestroyQueue_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclDestroyQueue_responseDefaultTypeInternal _xclDestroyQueue_response_default_instance_;

inline constexpr xclDestroyQueue_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        q_handle_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclDestroyQueue_call::xclDestroyQueue_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclDestroyQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclDestroyQueue_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclDestroyQueue_callDefaultTypeInternal() {}
  union {
    xclDestroyQueue_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclDestroyQueue_callDefaultTypeInternal _xclDestroyQueue_call_default_instance_;

inline constexpr xclCreateQueue_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        q_handle_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclCreateQueue_response::xclCreateQueue_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCreateQueue_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCreateQueue_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCreateQueue_responseDefaultTypeInternal() {}
  union {
    xclCreateQueue_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCreateQueue_responseDefaultTypeInternal _xclCreateQueue_response_default_instance_;

inline constexpr xclCreateQueue_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        write_{false},
        type_{0u},
        state_{0u},
        route_{0u},
        flow_{::uint64_t{0u}},
        qsize_{0u},
        desc_size_{0u},
        flags_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclCreateQueue_call::xclCreateQueue_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCreateQueue_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCreateQueue_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCreateQueue_callDefaultTypeInternal() {}
  union {
    xclCreateQueue_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCreateQueue_callDefaultTypeInternal _xclCreateQueue_call_default_instance_;

inline constexpr xclCopyBufferHost2Device_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBufferHost2Device_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferHost2Device_responseDefaultTypeInternal() {}
  union {
    xclCopyBufferHost2Device_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferHost2Device_responseDefaultTypeInternal _xclCopyBufferHost2Device_response_default_instance_;

inline constexpr xclCopyBufferHost2Device_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        xcldevicehandle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        src_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        dest_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        seek_{::uint64_t{0u}},
        space_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBufferHost2Device_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferHost2Device_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferHost2Device_callDefaultTypeInternal() {}
  union {
    xclCopyBufferHost2Device_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferHost2Device_callDefaultTypeInternal _xclCopyBufferHost2Device_call_default_instance_;

inline constexpr xclCopyBufferDevice2Host_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dest_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBufferDevice2Host_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferDevice2Host_responseDefaultTypeInternal() {}
  union {
    xclCopyBufferDevice2Host_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferDevice2Host_responseDefaultTypeInternal _xclCopyBufferDevice2Host_response_default_instance_;

inline constexpr xclCopyBufferDevice2Host_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        xcldevicehandle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        dest_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        src_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        skip_{::uint64_t{0u}},
        space_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBufferDevice2Host_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBufferDevice2Host_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBufferDevice2Host_callDefaultTypeInternal() {}
  union {
    xclCopyBufferDevice2Host_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBufferDevice2Host_callDefaultTypeInternal _xclCopyBufferDevice2Host_call_default_instance_;

inline constexpr xclCopyBO_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBO_response::xclCopyBO_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBO_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBO_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBO_responseDefaultTypeInternal() {}
  union {
    xclCopyBO_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBO_responseDefaultTypeInternal _xclCopyBO_response_default_instance_;

inline constexpr xclCopyBO_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        dst_filename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        src_handle_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        src_offset_{::uint64_t{0u}},
        dst_offset_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclCopyBO_call::xclCopyBO_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclCopyBO_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclCopyBO_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclCopyBO_callDefaultTypeInternal() {}
  union {
    xclCopyBO_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclCopyBO_callDefaultTypeInternal _xclCopyBO_call_default_instance_;

inline constexpr xclClose_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        valid_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclClose_response::xclClose_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclClose_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclClose_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclClose_responseDefaultTypeInternal() {}
  union {
    xclClose_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclClose_responseDefaultTypeInternal _xclClose_response_default_instance_;

inline constexpr xclClose_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        xcldevicehandle_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        closeall_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclClose_call::xclClose_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclClose_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclClose_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclClose_callDefaultTypeInternal() {}
  union {
    xclClose_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclClose_callDefaultTypeInternal _xclClose_call_default_instance_;

inline constexpr xclAllocDeviceBuffer_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        filename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        ack_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclAllocDeviceBuffer_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclAllocDeviceBuffer_responseDefaultTypeInternal() {}
  union {
    xclAllocDeviceBuffer_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclAllocDeviceBuffer_responseDefaultTypeInternal _xclAllocDeviceBuffer_response_default_instance_;

inline constexpr xclAllocDeviceBuffer_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ddraddress_{::uint64_t{0u}},
        size_{::uint64_t{0u}},
        peertopeer_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclAllocDeviceBuffer_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclAllocDeviceBuffer_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclAllocDeviceBuffer_callDefaultTypeInternal() {}
  union {
    xclAllocDeviceBuffer_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclAllocDeviceBuffer_callDefaultTypeInternal _xclAllocDeviceBuffer_call_default_instance_;

inline constexpr swemuDriverVersion_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        success_{false} {}

template <typename>
PROTOBUF_CONSTEXPR swemuDriverVersion_response::swemuDriverVersion_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct swemuDriverVersion_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR swemuDriverVersion_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~swemuDriverVersion_responseDefaultTypeInternal() {}
  union {
    swemuDriverVersion_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 swemuDriverVersion_responseDefaultTypeInternal _swemuDriverVersion_response_default_instance_;

inline constexpr swemuDriverVersion_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        version_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()) {}

template <typename>
PROTOBUF_CONSTEXPR swemuDriverVersion_call::swemuDriverVersion_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct swemuDriverVersion_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR swemuDriverVersion_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~swemuDriverVersion_callDefaultTypeInternal() {}
  union {
    swemuDriverVersion_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 swemuDriverVersion_callDefaultTypeInternal _swemuDriverVersion_call_default_instance_;

inline constexpr response_packet_info::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        size_{::uint64_t{0u}},
        xcl_api_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR response_packet_info::response_packet_info(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct response_packet_infoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR response_packet_infoDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~response_packet_infoDefaultTypeInternal() {}
  union {
    response_packet_info _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 response_packet_infoDefaultTypeInternal _response_packet_info_default_instance_;

inline constexpr ddr_mem_msg::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        strb_data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        size_{::uint64_t{0u}},
        fullsize_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR ddr_mem_msg::ddr_mem_msg(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct ddr_mem_msgDefaultTypeInternal {
  PROTOBUF_CONSTEXPR ddr_mem_msgDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~ddr_mem_msgDefaultTypeInternal() {}
  union {
    ddr_mem_msg _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 ddr_mem_msgDefaultTypeInternal _ddr_mem_msg_default_instance_;

inline constexpr call_packet_info::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        size_{::uint64_t{0u}},
        xcl_api_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR call_packet_info::call_packet_info(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct call_packet_infoDefaultTypeInternal {
  PROTOBUF_CONSTEXPR call_packet_infoDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~call_packet_infoDefaultTypeInternal() {}
  union {
    call_packet_info _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 call_packet_infoDefaultTypeInternal _call_packet_info_default_instance_;

inline constexpr xclWriteAddrKernelCtrl_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        kernel_info_{},
        data_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        addr_{::uint64_t{0u}},
        size_{0u},
        pf_id_{0u},
        bar_id_{0u} {}

template <typename>
PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclWriteAddrKernelCtrl_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclWriteAddrKernelCtrl_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclWriteAddrKernelCtrl_callDefaultTypeInternal() {}
  union {
    xclWriteAddrKernelCtrl_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclWriteAddrKernelCtrl_callDefaultTypeInternal _xclWriteAddrKernelCtrl_call_default_instance_;

inline constexpr xclSetupInstance_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        setup_{},
        route_{::uint64_t{0u}} {}

template <typename>
PROTOBUF_CONSTEXPR xclSetupInstance_call::xclSetupInstance_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetupInstance_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetupInstance_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetupInstance_callDefaultTypeInternal() {}
  union {
    xclSetupInstance_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetupInstance_callDefaultTypeInternal _xclSetupInstance_call_default_instance_;

inline constexpr xclSetEnvironment_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : environment_{},
        _cached_size_{0} {}

template <typename>
PROTOBUF_CONSTEXPR xclSetEnvironment_call::xclSetEnvironment_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclSetEnvironment_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclSetEnvironment_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclSetEnvironment_callDefaultTypeInternal() {}
  union {
    xclSetEnvironment_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclSetEnvironment_callDefaultTypeInternal _xclSetEnvironment_call_default_instance_;

inline constexpr xclPollCompletion_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : fullrequest_{},
        _cached_size_{0} {}

template <typename>
PROTOBUF_CONSTEXPR xclPollCompletion_response::xclPollCompletion_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPollCompletion_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPollCompletion_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPollCompletion_responseDefaultTypeInternal() {}
  union {
    xclPollCompletion_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPollCompletion_responseDefaultTypeInternal _xclPollCompletion_response_default_instance_;

inline constexpr xclPerfMonReadTrace_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : output_data_{},
        _cached_size_{0} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_responseDefaultTypeInternal _xclPerfMonReadTrace_response_default_instance_;

inline constexpr xclPerfMonReadTrace_Streaming_response::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : output_data_{},
        _cached_size_{0} {}

template <typename>
PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal() {}
  union {
    xclPerfMonReadTrace_Streaming_response _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclPerfMonReadTrace_Streaming_responseDefaultTypeInternal _xclPerfMonReadTrace_Streaming_response_default_instance_;

inline constexpr xclLoadBitstream_call::Impl_::Impl_(
    ::_pbi::ConstantInitialized) noexcept
      : _cached_size_{0},
        ddrbanks_{},
        xmlfile_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        dlopenfilename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        devicename_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        devicedirectory_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        binarydirectory_(
            &::google::protobuf::internal::fixed_address_empty_string,
            ::_pbi::ConstantInitialized()),
        verbose_{false} {}

template <typename>
PROTOBUF_CONSTEXPR xclLoadBitstream_call::xclLoadBitstream_call(::_pbi::ConstantInitialized)
    : _impl_(::_pbi::ConstantInitialized()) {}
struct xclLoadBitstream_callDefaultTypeInternal {
  PROTOBUF_CONSTEXPR xclLoadBitstream_callDefaultTypeInternal() : _instance(::_pbi::ConstantInitialized{}) {}
  ~xclLoadBitstream_callDefaultTypeInternal() {}
  union {
    xclLoadBitstream_call _instance;
  };
};

PROTOBUF_ATTRIBUTE_NO_DESTROY PROTOBUF_CONSTINIT
    PROTOBUF_ATTRIBUTE_INIT_PRIORITY1 xclLoadBitstream_callDefaultTypeInternal _xclLoadBitstream_call_default_instance_;
static ::_pb::Metadata file_level_metadata_rpc_5fmessages_2eproto[100];
static constexpr const ::_pb::EnumDescriptor**
    file_level_enum_descriptors_rpc_5fmessages_2eproto = nullptr;
static constexpr const ::_pb::ServiceDescriptor**
    file_level_service_descriptors_rpc_5fmessages_2eproto = nullptr;
const ::uint32_t TableStruct_rpc_5fmessages_2eproto::offsets[] PROTOBUF_SECTION_VARIABLE(
    protodesc_cold) = {
    PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::call_packet_info, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::call_packet_info, _impl_.xcl_api_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::response_packet_info, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::response_packet_info, _impl_.xcl_api_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_.name_),
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call_namevaluepair, _impl_.value_),
    0,
    1,
    ~0u,  // no _has_bits_
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_call, _impl_.environment_),
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetEnvironment_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call_ddrbank, _impl_.size_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.xmlfile_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.dlopenfilename_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.devicename_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.devicedirectory_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.binarydirectory_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.verbose_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_call, _impl_.ddrbanks_),
    0,
    1,
    2,
    3,
    4,
    5,
    ~0u,
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclLoadBitstream_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.ddraddress_),
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_call, _impl_.peertopeer_),
    0,
    1,
    2,
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_.ack_),
    PROTOBUF_FIELD_OFFSET(::xclAllocDeviceBuffer_response, _impl_.filename_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_call, _impl_.ddraddress_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclFreeDeviceBuffer_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclClose_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_.xcldevicehandle_),
    PROTOBUF_FIELD_OFFSET(::xclClose_call, _impl_.closeall_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclClose_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclClose_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclClose_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.xcldevicehandle_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.dest_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.src_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.seek_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_call, _impl_.space_),
    0,
    2,
    1,
    3,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferHost2Device_response, _impl_.size_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.xcldevicehandle_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.dest_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.src_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.skip_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_call, _impl_.space_),
    0,
    1,
    2,
    3,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBufferDevice2Host_response, _impl_.dest_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_),
    1,
    2,
    0,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrSpaceDeviceRam_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.name_),
    1,
    2,
    0,
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.kernel_info_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_call, _impl_.bar_id_),
    1,
    2,
    0,
    ~0u,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteAddrKernelCtrl_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.baseaddress_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.offset_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_call, _impl_.bar_id_),
    1,
    2,
    0,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclRegWrite_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_),
    1,
    2,
    0,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_.valid_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrSpaceDeviceRam_response, _impl_.data_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_call, _impl_.bar_id_),
    1,
    2,
    0,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_.valid_),
    PROTOBUF_FIELD_OFFSET(::xclReadAddrKernelCtrl_response, _impl_.data_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.baseaddress_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.offset_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.pf_id_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_call, _impl_.bar_id_),
    1,
    2,
    3,
    0,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_.valid_),
    PROTOBUF_FIELD_OFFSET(::xclRegRead_response, _impl_.data_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_.slotname_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_call, _impl_.accel_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.wr_byte_count_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.wr_trans_count_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.total_wr_latency_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.rd_byte_count_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.rd_trans_count_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_response, _impl_.total_rd_latency_),
    0,
    1,
    2,
    3,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.ack_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.slotname_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_call, _impl_.accel_),
    1,
    0,
    2,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonGetTraceCount_response, _impl_.no_of_samples_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.ack_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.slotname_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_call, _impl_.accel_),
    1,
    0,
    2,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.awlen_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.arlen_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.eventflags_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.timestamp_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.host_timestamp_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.rd_bytes_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response_events, _impl_.wr_bytes_),
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    ~0u,  // no _has_bits_
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_response, _impl_.output_data_),
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.data_),
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.strb_data_),
    PROTOBUF_FIELD_OFFSET(::ddr_mem_msg, _impl_.fullsize_),
    0,
    2,
    1,
    3,
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_call, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGetDeviceTimestamp_response, _impl_.device_timestamp_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_call, _impl_.slot_n_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_wr_cycles_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.idle_bus_cycles_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_rd_id_),
    PROTOBUF_FIELD_OFFSET(::xclReadBusStatus_response, _impl_.uncomplete_wr_id_),
    0,
    1,
    2,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_.ack_),
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_call, _impl_.force_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.display_msgs_),
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.log_msgs_),
    PROTOBUF_FIELD_OFFSET(::xclGetDebugMessages_response, _impl_.stop_msgs_),
    0,
    1,
    2,
    PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.dst_filename_),
    PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.offset_),
    PROTOBUF_FIELD_OFFSET(::xclImportBO_call, _impl_.size_),
    0,
    1,
    2,
    PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclImportBO_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.src_handle_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.dst_filename_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.src_offset_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_call, _impl_.dst_offset_),
    1,
    0,
    2,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCopyBO_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.write_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.type_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.state_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.route_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.flow_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.qsize_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.desc_size_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_call, _impl_.flags_),
    0,
    1,
    2,
    3,
    4,
    5,
    6,
    7,
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclCreateQueue_response, _impl_.q_handle_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.q_handle_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.src_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.req_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.nonblocking_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_call, _impl_.eot_),
    1,
    0,
    2,
    3,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclWriteQueue_response, _impl_.written_size_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.q_handle_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.dest_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.req_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.nonblocking_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_call, _impl_.eot_),
    1,
    0,
    2,
    3,
    4,
    5,
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_.dest_),
    PROTOBUF_FIELD_OFFSET(::xclReadQueue_response, _impl_.size_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_call, _impl_.req_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_.dest_),
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response_request, _impl_.size_),
    0,
    1,
    ~0u,  // no _has_bits_
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPollCompletion_response, _impl_.fullrequest_),
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_call, _impl_.q_handle_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclDestroyQueue_response, _impl_.success_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.arg_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.flow_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call_argflowpair, _impl_.tag_),
    1,
    2,
    0,
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_.route_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_call, _impl_.setup_),
    0,
    ~0u,
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSetupInstance_response, _impl_.success_),
    0,
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_call, _impl_.version_),
    0,
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::swemuDriverVersion_response, _impl_.success_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_call, _impl_.slotname_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_data_bytes_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_busy_cycles_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_stall_cycles_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_),
    0,
    1,
    2,
    3,
    4,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_.ack_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_call, _impl_.slotname_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.timestamp_),
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_),
    0,
    1,
    2,
    ~0u,  // no _has_bits_
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclPerfMonReadTrace_Streaming_response, _impl_.output_data_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_call, _impl_.size_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_.valid_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveReadReq_response, _impl_.data_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.addr_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_call, _impl_.data_),
    1,
    2,
    0,
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSlaveWriteReq_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_call, _impl_.interrupt_line_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclInterruptOccured_response, _impl_.valid_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_call, _impl_.graphname_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphInit_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_call, _impl_.iterations_),
    0,
    1,
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphRun_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_call, _impl_.gh_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphWait_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_call, _impl_.gh_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphEnd_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.portname_),
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.buffer_),
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_call, _impl_.size_),
    3,
    0,
    1,
    2,
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphUpdateRTP_response, _impl_.size_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.portname_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.buffer_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_call, _impl_.size_),
    3,
    0,
    1,
    2,
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclGraphReadRTP_response, _impl_.buffer_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.gmioname_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.dir_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.size_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.offset_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_call, _impl_.boh_),
    0,
    4,
    1,
    2,
    3,
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclSyncBOAIENB_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_call, _impl_.gmioname_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGMIOWait_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_call, _impl_.cycletimeout_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedWait_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_.gh_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_call, _impl_.cycletimeout_),
    1,
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphTimedEnd_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_call, _impl_.gh_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclGraphResume_response, _impl_.ack_),
    0,
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.xmlbuff_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.xmlbuffsize_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.sharedbin_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.sharedbinsize_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.emuldata_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.emuldatasize_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_call, _impl_.keepdir_),
    0,
    3,
    1,
    4,
    2,
    5,
    6,
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _impl_._has_bits_),
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _internal_metadata_),
    ~0u,  // no _extensions_
    ~0u,  // no _oneof_case_
    ~0u,  // no _weak_field_map_
    ~0u,  // no _inlined_string_donated_
    ~0u,  // no _split_
    ~0u,  // no sizeof(Split)
    PROTOBUF_FIELD_OFFSET(::xclLoadXclbinContent_response, _impl_.ack_),
    0,
};

static const ::_pbi::MigrationSchema
    schemas[] PROTOBUF_SECTION_VARIABLE(protodesc_cold) = {
        {0, 10, -1, sizeof(::call_packet_info)},
        {12, 22, -1, sizeof(::response_packet_info)},
        {24, 34, -1, sizeof(::xclSetEnvironment_call_namevaluepair)},
        {36, -1, -1, sizeof(::xclSetEnvironment_call)},
        {45, 54, -1, sizeof(::xclSetEnvironment_response)},
        {55, 64, -1, sizeof(::xclLoadBitstream_call_ddrbank)},
        {65, 80, -1, sizeof(::xclLoadBitstream_call)},
        {87, 96, -1, sizeof(::xclLoadBitstream_response)},
        {97, 108, -1, sizeof(::xclAllocDeviceBuffer_call)},
        {111, 121, -1, sizeof(::xclAllocDeviceBuffer_response)},
        {123, 132, -1, sizeof(::xclFreeDeviceBuffer_call)},
        {133, 142, -1, sizeof(::xclFreeDeviceBuffer_response)},
        {143, 153, -1, sizeof(::xclClose_call)},
        {155, 164, -1, sizeof(::xclClose_response)},
        {165, 179, -1, sizeof(::xclCopyBufferHost2Device_call)},
        {185, 194, -1, sizeof(::xclCopyBufferHost2Device_response)},
        {195, 209, -1, sizeof(::xclCopyBufferDevice2Host_call)},
        {215, 225, -1, sizeof(::xclCopyBufferDevice2Host_response)},
        {227, 240, -1, sizeof(::xclWriteAddrSpaceDeviceRam_call)},
        {245, 254, -1, sizeof(::xclWriteAddrSpaceDeviceRam_response)},
        {255, 266, -1, sizeof(::xclWriteAddrKernelCtrl_call_kernelInfo)},
        {269, 283, -1, sizeof(::xclWriteAddrKernelCtrl_call)},
        {289, 298, -1, sizeof(::xclWriteAddrKernelCtrl_response)},
        {299, 312, -1, sizeof(::xclRegWrite_call)},
        {317, 326, -1, sizeof(::xclRegWrite_response)},
        {327, 340, -1, sizeof(::xclReadAddrSpaceDeviceRam_call)},
        {345, 355, -1, sizeof(::xclReadAddrSpaceDeviceRam_response)},
        {357, 370, -1, sizeof(::xclReadAddrKernelCtrl_call)},
        {375, 385, -1, sizeof(::xclReadAddrKernelCtrl_response)},
        {387, 401, -1, sizeof(::xclRegRead_call)},
        {407, 417, -1, sizeof(::xclRegRead_response)},
        {419, 429, -1, sizeof(::xclPerfMonReadCounters_call)},
        {431, 445, -1, sizeof(::xclPerfMonReadCounters_response)},
        {451, 462, -1, sizeof(::xclPerfMonGetTraceCount_call)},
        {465, 474, -1, sizeof(::xclPerfMonGetTraceCount_response)},
        {475, 486, -1, sizeof(::xclPerfMonReadTrace_call)},
        {489, 504, -1, sizeof(::xclPerfMonReadTrace_response_events)},
        {511, -1, -1, sizeof(::xclPerfMonReadTrace_response)},
        {520, 532, -1, sizeof(::ddr_mem_msg)},
        {536, 545, -1, sizeof(::xclGetDeviceTimestamp_call)},
        {546, 555, -1, sizeof(::xclGetDeviceTimestamp_response)},
        {556, 565, -1, sizeof(::xclReadBusStatus_call)},
        {566, 579, -1, sizeof(::xclReadBusStatus_response)},
        {584, 594, -1, sizeof(::xclGetDebugMessages_call)},
        {596, 607, -1, sizeof(::xclGetDebugMessages_response)},
        {610, 621, -1, sizeof(::xclImportBO_call)},
        {624, 633, -1, sizeof(::xclImportBO_response)},
        {634, 647, -1, sizeof(::xclCopyBO_call)},
        {652, 661, -1, sizeof(::xclCopyBO_response)},
        {662, 678, -1, sizeof(::xclCreateQueue_call)},
        {686, 695, -1, sizeof(::xclCreateQueue_response)},
        {696, 710, -1, sizeof(::xclWriteQueue_call)},
        {716, 725, -1, sizeof(::xclWriteQueue_response)},
        {726, 740, -1, sizeof(::xclReadQueue_call)},
        {746, 756, -1, sizeof(::xclReadQueue_response)},
        {758, 767, -1, sizeof(::xclPollCompletion_call)},
        {768, 778, -1, sizeof(::xclPollCompletion_response_request)},
        {780, -1, -1, sizeof(::xclPollCompletion_response)},
        {789, 798, -1, sizeof(::xclDestroyQueue_call)},
        {799, 808, -1, sizeof(::xclDestroyQueue_response)},
        {809, 820, -1, sizeof(::xclSetupInstance_call_argflowpair)},
        {823, 833, -1, sizeof(::xclSetupInstance_call)},
        {835, 844, -1, sizeof(::xclSetupInstance_response)},
        {845, 854, -1, sizeof(::swemuDriverVersion_call)},
        {855, 864, -1, sizeof(::swemuDriverVersion_response)},
        {865, 874, -1, sizeof(::xclPerfMonReadCounters_Streaming_call)},
        {875, 888, -1, sizeof(::xclPerfMonReadCounters_Streaming_response)},
        {893, 903, -1, sizeof(::xclPerfMonReadTrace_Streaming_call)},
        {905, 916, -1, sizeof(::xclPerfMonReadTrace_Streaming_response_events)},
        {919, -1, -1, sizeof(::xclPerfMonReadTrace_Streaming_response)},
        {928, 938, -1, sizeof(::xclSlaveReadReq_call)},
        {940, 950, -1, sizeof(::xclSlaveReadReq_response)},
        {952, 963, -1, sizeof(::xclSlaveWriteReq_call)},
        {966, 975, -1, sizeof(::xclSlaveWriteReq_response)},
        {976, 985, -1, sizeof(::xclInterruptOccured_call)},
        {986, 995, -1, sizeof(::xclInterruptOccured_response)},
        {996, 1006, -1, sizeof(::xclGraphInit_call)},
        {1008, 1017, -1, sizeof(::xclGraphInit_response)},
        {1018, 1028, -1, sizeof(::xclGraphRun_call)},
        {1030, 1039, -1, sizeof(::xclGraphRun_response)},
        {1040, 1049, -1, sizeof(::xclGraphWait_call)},
        {1050, 1059, -1, sizeof(::xclGraphWait_response)},
        {1060, 1069, -1, sizeof(::xclGraphEnd_call)},
        {1070, 1079, -1, sizeof(::xclGraphEnd_response)},
        {1080, 1092, -1, sizeof(::xclGraphUpdateRTP_call)},
        {1096, 1105, -1, sizeof(::xclGraphUpdateRTP_response)},
        {1106, 1118, -1, sizeof(::xclGraphReadRTP_call)},
        {1122, 1132, -1, sizeof(::xclGraphReadRTP_response)},
        {1134, 1147, -1, sizeof(::xclSyncBOAIENB_call)},
        {1152, 1161, -1, sizeof(::xclSyncBOAIENB_response)},
        {1162, 1171, -1, sizeof(::xclGMIOWait_call)},
        {1172, 1181, -1, sizeof(::xclGMIOWait_response)},
        {1182, 1192, -1, sizeof(::xclGraphTimedWait_call)},
        {1194, 1203, -1, sizeof(::xclGraphTimedWait_response)},
        {1204, 1214, -1, sizeof(::xclGraphTimedEnd_call)},
        {1216, 1225, -1, sizeof(::xclGraphTimedEnd_response)},
        {1226, 1235, -1, sizeof(::xclGraphResume_call)},
        {1236, 1245, -1, sizeof(::xclGraphResume_response)},
        {1246, 1261, -1, sizeof(::xclLoadXclbinContent_call)},
        {1268, 1277, -1, sizeof(::xclLoadXclbinContent_response)},
};

static const ::_pb::Message* const file_default_instances[] = {
    &::_call_packet_info_default_instance_._instance,
    &::_response_packet_info_default_instance_._instance,
    &::_xclSetEnvironment_call_namevaluepair_default_instance_._instance,
    &::_xclSetEnvironment_call_default_instance_._instance,
    &::_xclSetEnvironment_response_default_instance_._instance,
    &::_xclLoadBitstream_call_ddrbank_default_instance_._instance,
    &::_xclLoadBitstream_call_default_instance_._instance,
    &::_xclLoadBitstream_response_default_instance_._instance,
    &::_xclAllocDeviceBuffer_call_default_instance_._instance,
    &::_xclAllocDeviceBuffer_response_default_instance_._instance,
    &::_xclFreeDeviceBuffer_call_default_instance_._instance,
    &::_xclFreeDeviceBuffer_response_default_instance_._instance,
    &::_xclClose_call_default_instance_._instance,
    &::_xclClose_response_default_instance_._instance,
    &::_xclCopyBufferHost2Device_call_default_instance_._instance,
    &::_xclCopyBufferHost2Device_response_default_instance_._instance,
    &::_xclCopyBufferDevice2Host_call_default_instance_._instance,
    &::_xclCopyBufferDevice2Host_response_default_instance_._instance,
    &::_xclWriteAddrSpaceDeviceRam_call_default_instance_._instance,
    &::_xclWriteAddrSpaceDeviceRam_response_default_instance_._instance,
    &::_xclWriteAddrKernelCtrl_call_kernelInfo_default_instance_._instance,
    &::_xclWriteAddrKernelCtrl_call_default_instance_._instance,
    &::_xclWriteAddrKernelCtrl_response_default_instance_._instance,
    &::_xclRegWrite_call_default_instance_._instance,
    &::_xclRegWrite_response_default_instance_._instance,
    &::_xclReadAddrSpaceDeviceRam_call_default_instance_._instance,
    &::_xclReadAddrSpaceDeviceRam_response_default_instance_._instance,
    &::_xclReadAddrKernelCtrl_call_default_instance_._instance,
    &::_xclReadAddrKernelCtrl_response_default_instance_._instance,
    &::_xclRegRead_call_default_instance_._instance,
    &::_xclRegRead_response_default_instance_._instance,
    &::_xclPerfMonReadCounters_call_default_instance_._instance,
    &::_xclPerfMonReadCounters_response_default_instance_._instance,
    &::_xclPerfMonGetTraceCount_call_default_instance_._instance,
    &::_xclPerfMonGetTraceCount_response_default_instance_._instance,
    &::_xclPerfMonReadTrace_call_default_instance_._instance,
    &::_xclPerfMonReadTrace_response_events_default_instance_._instance,
    &::_xclPerfMonReadTrace_response_default_instance_._instance,
    &::_ddr_mem_msg_default_instance_._instance,
    &::_xclGetDeviceTimestamp_call_default_instance_._instance,
    &::_xclGetDeviceTimestamp_response_default_instance_._instance,
    &::_xclReadBusStatus_call_default_instance_._instance,
    &::_xclReadBusStatus_response_default_instance_._instance,
    &::_xclGetDebugMessages_call_default_instance_._instance,
    &::_xclGetDebugMessages_response_default_instance_._instance,
    &::_xclImportBO_call_default_instance_._instance,
    &::_xclImportBO_response_default_instance_._instance,
    &::_xclCopyBO_call_default_instance_._instance,
    &::_xclCopyBO_response_default_instance_._instance,
    &::_xclCreateQueue_call_default_instance_._instance,
    &::_xclCreateQueue_response_default_instance_._instance,
    &::_xclWriteQueue_call_default_instance_._instance,
    &::_xclWriteQueue_response_default_instance_._instance,
    &::_xclReadQueue_call_default_instance_._instance,
    &::_xclReadQueue_response_default_instance_._instance,
    &::_xclPollCompletion_call_default_instance_._instance,
    &::_xclPollCompletion_response_request_default_instance_._instance,
    &::_xclPollCompletion_response_default_instance_._instance,
    &::_xclDestroyQueue_call_default_instance_._instance,
    &::_xclDestroyQueue_response_default_instance_._instance,
    &::_xclSetupInstance_call_argflowpair_default_instance_._instance,
    &::_xclSetupInstance_call_default_instance_._instance,
    &::_xclSetupInstance_response_default_instance_._instance,
    &::_swemuDriverVersion_call_default_instance_._instance,
    &::_swemuDriverVersion_response_default_instance_._instance,
    &::_xclPerfMonReadCounters_Streaming_call_default_instance_._instance,
    &::_xclPerfMonReadCounters_Streaming_response_default_instance_._instance,
    &::_xclPerfMonReadTrace_Streaming_call_default_instance_._instance,
    &::_xclPerfMonReadTrace_Streaming_response_events_default_instance_._instance,
    &::_xclPerfMonReadTrace_Streaming_response_default_instance_._instance,
    &::_xclSlaveReadReq_call_default_instance_._instance,
    &::_xclSlaveReadReq_response_default_instance_._instance,
    &::_xclSlaveWriteReq_call_default_instance_._instance,
    &::_xclSlaveWriteReq_response_default_instance_._instance,
    &::_xclInterruptOccured_call_default_instance_._instance,
    &::_xclInterruptOccured_response_default_instance_._instance,
    &::_xclGraphInit_call_default_instance_._instance,
    &::_xclGraphInit_response_default_instance_._instance,
    &::_xclGraphRun_call_default_instance_._instance,
    &::_xclGraphRun_response_default_instance_._instance,
    &::_xclGraphWait_call_default_instance_._instance,
    &::_xclGraphWait_response_default_instance_._instance,
    &::_xclGraphEnd_call_default_instance_._instance,
    &::_xclGraphEnd_response_default_instance_._instance,
    &::_xclGraphUpdateRTP_call_default_instance_._instance,
    &::_xclGraphUpdateRTP_response_default_instance_._instance,
    &::_xclGraphReadRTP_call_default_instance_._instance,
    &::_xclGraphReadRTP_response_default_instance_._instance,
    &::_xclSyncBOAIENB_call_default_instance_._instance,
    &::_xclSyncBOAIENB_response_default_instance_._instance,
    &::_xclGMIOWait_call_default_instance_._instance,
    &::_xclGMIOWait_response_default_instance_._instance,
    &::_xclGraphTimedWait_call_default_instance_._instance,
    &::_xclGraphTimedWait_response_default_instance_._instance,
    &::_xclGraphTimedEnd_call_default_instance_._instance,
    &::_xclGraphTimedEnd_response_default_instance_._instance,
    &::_xclGraphResume_call_default_instance_._instance,
    &::_xclGraphResume_response_default_instance_._instance,
    &::_xclLoadXclbinContent_call_default_instance_._instance,
    &::_xclLoadXclbinContent_response_default_instance_._instance,
};
const char descriptor_table_protodef_rpc_5fmessages_2eproto[] PROTOBUF_SECTION_VARIABLE(protodesc_cold) = {
    "\n\022rpc_messages.proto\"1\n\020call_packet_info"
    "\022\014\n\004size\030\001 \002(\006\022\017\n\007xcl_api\030\002 \001(\007\"5\n\024respo"
    "nse_packet_info\022\014\n\004size\030\001 \002(\006\022\017\n\007xcl_api"
    "\030\002 \001(\007\"\202\001\n\026xclSetEnvironment_call\022:\n\013env"
    "ironment\030\003 \003(\0132%.xclSetEnvironment_call."
    "namevaluepair\032,\n\rnamevaluepair\022\014\n\004name\030\001"
    " \001(\t\022\r\n\005value\030\002 \001(\t\")\n\032xclSetEnvironment"
    "_response\022\013\n\003ack\030\001 \001(\010\"\342\001\n\025xclLoadBitstr"
    "eam_call\022\017\n\007xmlfile\030\001 \002(\014\022\026\n\016dlopenfilen"
    "ame\030\002 \002(\014\022\022\n\ndevicename\030\003 \002(\014\022\027\n\017deviced"
    "irectory\030\004 \002(\014\022\027\n\017binarydirectory\030\005 \002(\014\022"
    "\017\n\007verbose\030\006 \002(\010\0220\n\010ddrbanks\030\010 \003(\0132\036.xcl"
    "LoadBitstream_call.ddrbank\032\027\n\007ddrbank\022\014\n"
    "\004size\030\007 \001(\004\"(\n\031xclLoadBitstream_response"
    "\022\013\n\003ack\030\001 \002(\010\"Q\n\031xclAllocDeviceBuffer_ca"
    "ll\022\022\n\nddraddress\030\001 \002(\004\022\014\n\004size\030\002 \002(\004\022\022\n\n"
    "peertopeer\030\003 \001(\010\">\n\035xclAllocDeviceBuffer"
    "_response\022\013\n\003ack\030\001 \002(\010\022\020\n\010filename\030\002 \001(\t"
    "\".\n\030xclFreeDeviceBuffer_call\022\022\n\nddraddre"
    "ss\030\001 \002(\004\"+\n\034xclFreeDeviceBuffer_response"
    "\022\013\n\003ack\030\001 \002(\010\":\n\rxclClose_call\022\027\n\017xclDev"
    "iceHandle\030\002 \001(\014\022\020\n\010closeall\030\003 \001(\010\"\"\n\021xcl"
    "Close_response\022\r\n\005valid\030\001 \002(\010\"~\n\035xclCopy"
    "BufferHost2Device_call\022\027\n\017xclDeviceHandl"
    "e\030\002 \002(\014\022\014\n\004dest\030\003 \002(\004\022\013\n\003src\030\004 \002(\014\022\014\n\004si"
    "ze\030\005 \002(\004\022\014\n\004seek\030\006 \002(\004\022\r\n\005space\030\007 \001(\r\"1\n"
    "!xclCopyBufferHost2Device_response\022\014\n\004si"
    "ze\030\001 \002(\004\"~\n\035xclCopyBufferDevice2Host_cal"
    "l\022\027\n\017xclDeviceHandle\030\002 \002(\014\022\014\n\004dest\030\003 \002(\014"
    "\022\013\n\003src\030\004 \002(\004\022\014\n\004size\030\005 \002(\004\022\014\n\004skip\030\006 \002("
    "\004\022\r\n\005space\030\007 \001(\r\"\?\n!xclCopyBufferDevice2"
    "Host_response\022\014\n\004size\030\001 \002(\004\022\014\n\004dest\030\002 \002("
    "\014\"j\n\037xclWriteAddrSpaceDeviceRam_call\022\014\n\004"
    "addr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \002(\014\022\r"
    "\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006 \001(\r\"4\n#xclWrit"
    "eAddrSpaceDeviceRam_response\022\r\n\005valid\030\001 "
    "\002(\010\"\334\001\n\033xclWriteAddrKernelCtrl_call\022\014\n\004a"
    "ddr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \002(\014\022<\n"
    "\013kernel_info\030\005 \003(\0132\'.xclWriteAddrKernelC"
    "trl_call.kernelInfo\022\r\n\005PF_ID\030\006 \001(\r\022\016\n\006BA"
    "R_ID\030\007 \001(\r\0326\n\nkernelInfo\022\014\n\004addr\030\001 \001(\004\022\014"
    "\n\004size\030\002 \001(\r\022\014\n\004name\030\003 \001(\t\"0\n\037xclWriteAd"
    "drKernelCtrl_response\022\r\n\005valid\030\001 \002(\010\"d\n\020"
    "xclRegWrite_call\022\023\n\013baseaddress\030\001 \002(\r\022\016\n"
    "\006offset\030\002 \002(\r\022\014\n\004data\030\003 \002(\014\022\r\n\005PF_ID\030\004 \001"
    "(\r\022\016\n\006BAR_ID\030\005 \001(\r\"%\n\024xclRegWrite_respon"
    "se\022\r\n\005valid\030\001 \002(\010\"i\n\036xclReadAddrSpaceDev"
    "iceRam_call\022\014\n\004addr\030\002 \002(\004\022\014\n\004size\030\003 \002(\r\022"
    "\014\n\004data\030\004 \001(\014\022\r\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006"
    " \001(\r\"A\n\"xclReadAddrSpaceDeviceRam_respon"
    "se\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002 \001(\014\"e\n\032xclRe"
    "adAddrKernelCtrl_call\022\014\n\004addr\030\002 \002(\004\022\014\n\004s"
    "ize\030\003 \002(\r\022\014\n\004data\030\004 \001(\014\022\r\n\005PF_ID\030\005 \001(\r\022\016"
    "\n\006BAR_ID\030\006 \001(\r\"=\n\036xclReadAddrKernelCtrl_"
    "response\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002 \001(\014\"q\n"
    "\017xclRegRead_call\022\023\n\013baseaddress\030\001 \002(\r\022\016\n"
    "\006offset\030\002 \002(\r\022\014\n\004size\030\003 \002(\r\022\014\n\004data\030\004 \001("
    "\014\022\r\n\005PF_ID\030\005 \001(\r\022\016\n\006BAR_ID\030\006 \001(\r\"2\n\023xclR"
    "egRead_response\022\r\n\005valid\030\001 \002(\010\022\014\n\004data\030\002"
    " \001(\014\">\n\033xclPerfMonReadCounters_call\022\020\n\010s"
    "lotname\030\001 \002(\t\022\r\n\005accel\030\002 \001(\010\"\263\001\n\037xclPerf"
    "MonReadCounters_response\022\025\n\rwr_byte_coun"
    "t\030\001 \001(\r\022\026\n\016wr_trans_count\030\002 \001(\r\022\030\n\020total"
    "_wr_latency\030\003 \001(\r\022\025\n\rrd_byte_count\030\004 \001(\r"
    "\022\026\n\016rd_trans_count\030\005 \001(\r\022\030\n\020total_rd_lat"
    "ency\030\006 \001(\r\"L\n\034xclPerfMonGetTraceCount_ca"
    "ll\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname\030\002 \001(\t\022\r\n\005acc"
    "el\030\003 \001(\010\"9\n xclPerfMonGetTraceCount_resp"
    "onse\022\025\n\rno_of_samples\030\001 \001(\r\"H\n\030xclPerfMo"
    "nReadTrace_call\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname"
    "\030\002 \001(\t\022\r\n\005accel\030\003 \001(\010\"\345\001\n\034xclPerfMonRead"
    "Trace_response\0229\n\013output_data\030\010 \003(\0132$.xc"
    "lPerfMonReadTrace_response.events\032\211\001\n\006ev"
    "ents\022\r\n\005awlen\030\001 \001(\r\022\r\n\005arlen\030\002 \001(\r\022\022\n\nev"
    "entflags\030\003 \001(\r\022\021\n\ttimestamp\030\004 \001(\r\022\026\n\016hos"
    "t_timestamp\030\005 \001(\004\022\020\n\010rd_bytes\030\006 \001(\r\022\020\n\010w"
    "r_bytes\030\007 \001(\r\"N\n\013ddr_mem_msg\022\014\n\004data\030\001 \002"
    "(\014\022\014\n\004size\030\002 \001(\004\022\021\n\tstrb_data\030\003 \001(\014\022\020\n\010f"
    "ullsize\030\004 \001(\004\")\n\032xclGetDeviceTimestamp_c"
    "all\022\013\n\003ack\030\001 \001(\010\":\n\036xclGetDeviceTimestam"
    "p_response\022\030\n\020device_timestamp\030\001 \002(\004\"\'\n\025"
    "xclReadBusStatus_call\022\016\n\006slot_n\030\001 \001(\005\"\244\001"
    "\n\031xclReadBusStatus_response\022\034\n\024uncomplet"
    "e_rd_cycles\030\001 \001(\004\022\034\n\024uncomplete_wr_cycle"
    "s\030\002 \001(\004\022\027\n\017idle_bus_cycles\030\003 \001(\004\022\030\n\020unco"
    "mplete_rd_id\030\004 \001(\004\022\030\n\020uncomplete_wr_id\030\005"
    " \001(\004\"6\n\030xclGetDebugMessages_call\022\013\n\003ack\030"
    "\001 \001(\010\022\r\n\005force\030\002 \001(\010\"Y\n\034xclGetDebugMessa"
    "ges_response\022\024\n\014display_msgs\030\001 \001(\014\022\020\n\010lo"
    "g_msgs\030\002 \001(\014\022\021\n\tstop_msgs\030\003 \001(\014\"F\n\020xclIm"
    "portBO_call\022\024\n\014dst_filename\030\001 \001(\t\022\016\n\006off"
    "set\030\002 \001(\004\022\014\n\004size\030\003 \001(\004\"#\n\024xclImportBO_r"
    "esponse\022\013\n\003ack\030\001 \001(\010\"p\n\016xclCopyBO_call\022\022"
    "\n\nsrc_handle\030\001 \001(\004\022\024\n\014dst_filename\030\002 \001(\t"
    "\022\014\n\004size\030\003 \001(\004\022\022\n\nsrc_offset\030\004 \001(\004\022\022\n\nds"
    "t_offset\030\005 \001(\004\"!\n\022xclCopyBO_response\022\013\n\003"
    "ack\030\001 \001(\010\"\217\001\n\023xclCreateQueue_call\022\r\n\005wri"
    "te\030\001 \001(\010\022\014\n\004type\030\002 \001(\r\022\r\n\005state\030\003 \001(\r\022\r\n"
    "\005route\030\004 \001(\r\022\014\n\004flow\030\005 \001(\004\022\r\n\005qsize\030\006 \001("
    "\r\022\021\n\tdesc_size\030\007 \001(\r\022\r\n\005flags\030\010 \001(\004\"+\n\027x"
    "clCreateQueue_response\022\020\n\010q_handle\030\001 \001(\004"
    "\"p\n\022xclWriteQueue_call\022\020\n\010q_handle\030\001 \001(\004"
    "\022\013\n\003src\030\002 \001(\014\022\014\n\004size\030\003 \001(\004\022\013\n\003req\030\004 \001(\004"
    "\022\023\n\013nonblocking\030\005 \001(\010\022\013\n\003eot\030\006 \001(\010\".\n\026xc"
    "lWriteQueue_response\022\024\n\014written_size\030\001 \001"
    "(\004\"p\n\021xclReadQueue_call\022\020\n\010q_handle\030\001 \001("
    "\004\022\014\n\004dest\030\002 \001(\014\022\014\n\004size\030\003 \001(\004\022\013\n\003req\030\004 \001"
    "(\004\022\023\n\013nonblocking\030\005 \001(\010\022\013\n\003eot\030\006 \001(\010\"3\n\025"
    "xclReadQueue_response\022\014\n\004dest\030\001 \001(\014\022\014\n\004s"
    "ize\030\002 \001(\004\"%\n\026xclPollCompletion_call\022\013\n\003r"
    "eq\030\001 \001(\004\"}\n\032xclPollCompletion_response\0228"
    "\n\013fullrequest\030\003 \003(\0132#.xclPollCompletion_"
    "response.request\032%\n\007request\022\014\n\004dest\030\001 \001("
    "\014\022\014\n\004size\030\002 \001(\004\"(\n\024xclDestroyQueue_call\022"
    "\020\n\010q_handle\030\001 \001(\004\"+\n\030xclDestroyQueue_res"
    "ponse\022\017\n\007success\030\001 \001(\010\"\220\001\n\025xclSetupInsta"
    "nce_call\022\r\n\005route\030\001 \001(\004\0221\n\005setup\030\005 \003(\0132\""
    ".xclSetupInstance_call.argflowpair\0325\n\013ar"
    "gflowpair\022\013\n\003arg\030\002 \001(\004\022\014\n\004flow\030\003 \001(\004\022\013\n\003"
    "tag\030\004 \001(\t\",\n\031xclSetupInstance_response\022\017"
    "\n\007success\030\001 \001(\010\"*\n\027swemuDriverVersion_ca"
    "ll\022\017\n\007version\030\001 \002(\t\".\n\033swemuDriverVersio"
    "n_response\022\017\n\007success\030\001 \001(\010\"9\n%xclPerfMo"
    "nReadCounters_Streaming_call\022\020\n\010slotname"
    "\030\001 \002(\t\"\250\001\n)xclPerfMonReadCounters_Stream"
    "ing_response\022\025\n\rstr_num_tranx\030\007 \001(\004\022\026\n\016s"
    "tr_data_bytes\030\010 \001(\004\022\027\n\017str_busy_cycles\030\t"
    " \001(\004\022\030\n\020str_stall_cycles\030\n \001(\004\022\031\n\021str_st"
    "arve_cycles\030\013 \001(\004\"C\n\"xclPerfMonReadTrace"
    "_Streaming_call\022\013\n\003ack\030\001 \001(\010\022\020\n\010slotname"
    "\030\002 \001(\t\"\266\001\n&xclPerfMonReadTrace_Streaming"
    "_response\022C\n\013output_data\030\010 \003(\0132..xclPerf"
    "MonReadTrace_Streaming_response.events\032G"
    "\n\006events\022\022\n\neventflags\030\003 \001(\r\022\021\n\ttimestam"
    "p\030\004 \001(\r\022\026\n\016host_timestamp\030\005 \001(\004\"2\n\024xclSl"
    "aveReadReq_call\022\014\n\004addr\030\001 \002(\004\022\014\n\004size\030\002 "
    "\002(\r\"7\n\030xclSlaveReadReq_response\022\r\n\005valid"
    "\030\001 \002(\010\022\014\n\004data\030\002 \002(\014\"A\n\025xclSlaveWriteReq"
    "_call\022\014\n\004addr\030\001 \002(\004\022\014\n\004size\030\002 \002(\r\022\014\n\004dat"
    "a\030\003 \002(\014\"*\n\031xclSlaveWriteReq_response\022\r\n\005"
    "valid\030\001 \002(\010\"2\n\030xclInterruptOccured_call\022"
    "\026\n\016interrupt_line\030\001 \002(\r\"-\n\034xclInterruptO"
    "ccured_response\022\r\n\005valid\030\001 \002(\010\"2\n\021xclGra"
    "phInit_call\022\n\n\002gh\030\002 \002(\r\022\021\n\tgraphname\030\003 \002"
    "(\014\"$\n\025xclGraphInit_response\022\013\n\003ack\030\001 \002(\010"
    "\"2\n\020xclGraphRun_call\022\n\n\002gh\030\002 \002(\r\022\022\n\niter"
    "ations\030\003 \002(\r\"#\n\024xclGraphRun_response\022\013\n\003"
    "ack\030\001 \002(\010\"\037\n\021xclGraphWait_call\022\n\n\002gh\030\002 \002"
    "(\r\"$\n\025xclGraphWait_response\022\013\n\003ack\030\001 \002(\010"
    "\"\036\n\020xclGraphEnd_call\022\n\n\002gh\030\002 \002(\r\"#\n\024xclG"
    "raphEnd_response\022\013\n\003ack\030\001 \002(\r\"T\n\026xclGrap"
    "hUpdateRTP_call\022\n\n\002gh\030\002 \002(\r\022\020\n\010portname\030"
    "\003 \002(\014\022\016\n\006buffer\030\004 \002(\014\022\014\n\004size\030\005 \002(\004\"*\n\032x"
    "clGraphUpdateRTP_response\022\014\n\004size\030\001 \002(\004\""
    "R\n\024xclGraphReadRTP_call\022\n\n\002gh\030\002 \002(\r\022\020\n\010p"
    "ortname\030\003 \002(\014\022\016\n\006buffer\030\004 \002(\014\022\014\n\004size\030\005 "
    "\002(\004\"8\n\030xclGraphReadRTP_response\022\014\n\004size\030"
    "\001 \002(\004\022\016\n\006buffer\030\002 \002(\014\"_\n\023xclSyncBOAIENB_"
    "call\022\020\n\010gmioname\030\002 \002(\014\022\013\n\003dir\030\003 \002(\r\022\014\n\004s"
    "ize\030\004 \002(\004\022\016\n\006offset\030\005 \002(\004\022\013\n\003boh\030\006 \002(\004\"&"
    "\n\027xclSyncBOAIENB_response\022\013\n\003ack\030\001 \002(\010\"$"
    "\n\020xclGMIOWait_call\022\020\n\010gmioname\030\002 \002(\014\"#\n\024"
    "xclGMIOWait_response\022\013\n\003ack\030\001 \002(\010\":\n\026xcl"
    "GraphTimedWait_call\022\n\n\002gh\030\002 \002(\r\022\024\n\014cycle"
    "timeout\030\003 \002(\004\")\n\032xclGraphTimedWait_respo"
    "nse\022\013\n\003ack\030\001 \002(\010\"9\n\025xclGraphTimedEnd_cal"
    "l\022\n\n\002gh\030\002 \002(\r\022\024\n\014cycletimeout\030\003 \002(\004\"(\n\031x"
    "clGraphTimedEnd_response\022\013\n\003ack\030\001 \002(\010\"!\n"
    "\023xclGraphResume_call\022\n\n\002gh\030\002 \002(\r\"&\n\027xclG"
    "raphResume_response\022\013\n\003ack\030\001 \002(\010\"\244\001\n\031xcl"
    "LoadXclbinContent_call\022\017\n\007xmlbuff\030\002 \002(\014\022"
    "\023\n\013xmlbuffsize\030\003 \002(\004\022\021\n\tsharedbin\030\004 \002(\014\022"
    "\025\n\rsharedbinsize\030\005 \002(\004\022\020\n\010emuldata\030\006 \002(\014"
    "\022\024\n\014emuldatasize\030\007 \002(\004\022\017\n\007keepdir\030\010 \002(\010\""
    ",\n\035xclLoadXclbinContent_response\022\013\n\003ack\030"
    "\001 \002(\010"
};
static ::absl::once_flag descriptor_table_rpc_5fmessages_2eproto_once;
const ::_pbi::DescriptorTable descriptor_table_rpc_5fmessages_2eproto = {
    false,
    false,
    7005,
    descriptor_table_protodef_rpc_5fmessages_2eproto,
    "rpc_messages.proto",
    &descriptor_table_rpc_5fmessages_2eproto_once,
    nullptr,
    0,
    100,
    schemas,
    file_default_instances,
    TableStruct_rpc_5fmessages_2eproto::offsets,
    file_level_metadata_rpc_5fmessages_2eproto,
    file_level_enum_descriptors_rpc_5fmessages_2eproto,
    file_level_service_descriptors_rpc_5fmessages_2eproto,
};

// This function exists to be marked as weak.
// It can significantly speed up compilation by breaking up LLVM's SCC
// in the .pb.cc translation units. Large translation units see a
// reduction of more than 35% of walltime for optimized builds. Without
// the weak attribute all the messages in the file, including all the
// vtables and everything they use become part of the same SCC through
// a cycle like:
// GetMetadata -> descriptor table -> default instances ->
//   vtables -> GetMetadata
// By adding a weak function here we break the connection from the
// individual vtables back into the descriptor table.
PROTOBUF_ATTRIBUTE_WEAK const ::_pbi::DescriptorTable* descriptor_table_rpc_5fmessages_2eproto_getter() {
  return &descriptor_table_rpc_5fmessages_2eproto;
}
// Force running AddDescriptors() at dynamic initialization time.
PROTOBUF_ATTRIBUTE_INIT_PRIORITY2
static ::_pbi::AddDescriptorsRunner dynamic_init_dummy_rpc_5fmessages_2eproto(&descriptor_table_rpc_5fmessages_2eproto);
// ===================================================================

class call_packet_info::_Internal {
 public:
  using HasBits = decltype(std::declval<call_packet_info>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xcl_api(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

call_packet_info::call_packet_info(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:call_packet_info)
}
call_packet_info::call_packet_info(
    ::google::protobuf::Arena* arena, const call_packet_info& from)
    : call_packet_info(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE call_packet_info::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void call_packet_info::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, xcl_api_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::xcl_api_));
}
call_packet_info::~call_packet_info() {
  // @@protoc_insertion_point(destructor:call_packet_info)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void call_packet_info::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void call_packet_info::Clear() {
// @@protoc_insertion_point(message_clear_start:call_packet_info)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.xcl_api_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* call_packet_info::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> call_packet_info::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_call_packet_info_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional fixed32 xcl_api = 2;
    {::_pbi::TcParser::FastF32S1,
     {21, 1, 0, PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.xcl_api_)}},
    // required fixed64 size = 1;
    {::_pbi::TcParser::FastF64S1,
     {9, 0, 0, PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required fixed64 size = 1;
    {PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kFixed64)},
    // optional fixed32 xcl_api = 2;
    {PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.xcl_api_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kFixed32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* call_packet_info::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:call_packet_info)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required fixed64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed64ToArray(
        1, this->_internal_size(), target);
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed32ToArray(
        2, this->_internal_xcl_api(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:call_packet_info)
  return target;
}

::size_t call_packet_info::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:call_packet_info)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required fixed64 size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 9;
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    total_size += 5;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData call_packet_info::_class_data_ = {
    call_packet_info::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* call_packet_info::GetClassData() const {
  return &_class_data_;
}

void call_packet_info::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<call_packet_info*>(&to_msg);
  auto& from = static_cast<const call_packet_info&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:call_packet_info)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.xcl_api_ = from._impl_.xcl_api_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void call_packet_info::CopyFrom(const call_packet_info& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:call_packet_info)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool call_packet_info::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* call_packet_info::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void call_packet_info::InternalSwap(call_packet_info* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.xcl_api_)
      + sizeof(call_packet_info::_impl_.xcl_api_)
      - PROTOBUF_FIELD_OFFSET(call_packet_info, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata call_packet_info::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[0]);
}
// ===================================================================

class response_packet_info::_Internal {
 public:
  using HasBits = decltype(std::declval<response_packet_info>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xcl_api(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

response_packet_info::response_packet_info(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:response_packet_info)
}
response_packet_info::response_packet_info(
    ::google::protobuf::Arena* arena, const response_packet_info& from)
    : response_packet_info(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE response_packet_info::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void response_packet_info::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, xcl_api_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::xcl_api_));
}
response_packet_info::~response_packet_info() {
  // @@protoc_insertion_point(destructor:response_packet_info)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void response_packet_info::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void response_packet_info::Clear() {
// @@protoc_insertion_point(message_clear_start:response_packet_info)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.xcl_api_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.xcl_api_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* response_packet_info::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> response_packet_info::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_response_packet_info_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional fixed32 xcl_api = 2;
    {::_pbi::TcParser::FastF32S1,
     {21, 1, 0, PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.xcl_api_)}},
    // required fixed64 size = 1;
    {::_pbi::TcParser::FastF64S1,
     {9, 0, 0, PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required fixed64 size = 1;
    {PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kFixed64)},
    // optional fixed32 xcl_api = 2;
    {PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.xcl_api_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kFixed32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* response_packet_info::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:response_packet_info)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required fixed64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed64ToArray(
        1, this->_internal_size(), target);
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteFixed32ToArray(
        2, this->_internal_xcl_api(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:response_packet_info)
  return target;
}

::size_t response_packet_info::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:response_packet_info)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required fixed64 size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 9;
  }

  // optional fixed32 xcl_api = 2;
  if (cached_has_bits & 0x00000002u) {
    total_size += 5;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData response_packet_info::_class_data_ = {
    response_packet_info::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* response_packet_info::GetClassData() const {
  return &_class_data_;
}

void response_packet_info::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<response_packet_info*>(&to_msg);
  auto& from = static_cast<const response_packet_info&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:response_packet_info)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.xcl_api_ = from._impl_.xcl_api_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void response_packet_info::CopyFrom(const response_packet_info& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:response_packet_info)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool response_packet_info::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* response_packet_info::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void response_packet_info::InternalSwap(response_packet_info* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.xcl_api_)
      + sizeof(response_packet_info::_impl_.xcl_api_)
      - PROTOBUF_FIELD_OFFSET(response_packet_info, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata response_packet_info::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[1]);
}
// ===================================================================

class xclSetEnvironment_call_namevaluepair::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetEnvironment_call_namevaluepair>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_._has_bits_);
  static void set_has_name(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_value(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_call.namevaluepair)
}
inline PROTOBUF_NDEBUG_INLINE xclSetEnvironment_call_namevaluepair::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        name_(arena, from.name_),
        value_(arena, from.value_) {}

xclSetEnvironment_call_namevaluepair::xclSetEnvironment_call_namevaluepair(
    ::google::protobuf::Arena* arena,
    const xclSetEnvironment_call_namevaluepair& from)
    : ::google::protobuf::Message(arena) {
  xclSetEnvironment_call_namevaluepair* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclSetEnvironment_call.namevaluepair)
}
inline PROTOBUF_NDEBUG_INLINE xclSetEnvironment_call_namevaluepair::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        name_(arena),
        value_(arena) {}

inline void xclSetEnvironment_call_namevaluepair::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclSetEnvironment_call_namevaluepair::~xclSetEnvironment_call_namevaluepair() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_call.namevaluepair)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetEnvironment_call_namevaluepair::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.name_.Destroy();
  _impl_.value_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetEnvironment_call_namevaluepair::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_call.namevaluepair)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.name_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.value_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetEnvironment_call_namevaluepair::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 54, 2> xclSetEnvironment_call_namevaluepair::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSetEnvironment_call_namevaluepair_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional string value = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 1, 0, PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_.value_)}},
    // optional string name = 1;
    {::_pbi::TcParser::FastSS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_.name_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional string name = 1;
    {PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_.name_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional string value = 2;
    {PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call_namevaluepair, _impl_.value_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\44\4\5\0\0\0\0\0"
    "xclSetEnvironment_call.namevaluepair"
    "name"
    "value"
  }},
};

::uint8_t* xclSetEnvironment_call_namevaluepair::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_call.namevaluepair)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional string name = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_name();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclSetEnvironment_call.namevaluepair.name");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional string value = 2;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_value();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclSetEnvironment_call.namevaluepair.value");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_call.namevaluepair)
  return target;
}

::size_t xclSetEnvironment_call_namevaluepair::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_call.namevaluepair)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional string name = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_name());
    }

    // optional string value = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_value());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetEnvironment_call_namevaluepair::_class_data_ = {
    xclSetEnvironment_call_namevaluepair::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetEnvironment_call_namevaluepair::GetClassData() const {
  return &_class_data_;
}

void xclSetEnvironment_call_namevaluepair::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_call_namevaluepair*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_call_namevaluepair&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_call.namevaluepair)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_name(from._internal_name());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_value(from._internal_value());
    }
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_call_namevaluepair::CopyFrom(const xclSetEnvironment_call_namevaluepair& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_call.namevaluepair)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetEnvironment_call_namevaluepair::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetEnvironment_call_namevaluepair::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetEnvironment_call_namevaluepair::InternalSwap(xclSetEnvironment_call_namevaluepair* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.value_, &other->_impl_.value_, arena);
}

::google::protobuf::Metadata xclSetEnvironment_call_namevaluepair::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[2]);
}
// ===================================================================

class xclSetEnvironment_call::_Internal {
 public:
};

xclSetEnvironment_call::xclSetEnvironment_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSetEnvironment_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : environment_{visibility, arena, from.environment_},
        _cached_size_{0} {}

xclSetEnvironment_call::xclSetEnvironment_call(
    ::google::protobuf::Arena* arena,
    const xclSetEnvironment_call& from)
    : ::google::protobuf::Message(arena) {
  xclSetEnvironment_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclSetEnvironment_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSetEnvironment_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : environment_{visibility, arena},
        _cached_size_{0} {}

inline void xclSetEnvironment_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclSetEnvironment_call::~xclSetEnvironment_call() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetEnvironment_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetEnvironment_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.environment_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetEnvironment_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2> xclSetEnvironment_call::_table_ = {
  {
    0,  // no _has_bits_
    0, // no _extensions_
    3, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967291,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclSetEnvironment_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
    {::_pbi::TcParser::FastMtR1,
     {26, 63, 0, PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call, _impl_.environment_)}},
  }}, {{
    65535, 65535
  }}, {{
    // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
    {PROTOBUF_FIELD_OFFSET(xclSetEnvironment_call, _impl_.environment_), 0, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclSetEnvironment_call_namevaluepair>()},
  }}, {{
  }},
};

::uint8_t* xclSetEnvironment_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_environment_size()); i < n; i++) {
    const auto& repfield = this->_internal_environment().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(3, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_call)
  return target;
}

::size_t xclSetEnvironment_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclSetEnvironment_call.namevaluepair environment = 3;
  total_size += 1UL * this->_internal_environment_size();
  for (const auto& msg : this->_internal_environment()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetEnvironment_call::_class_data_ = {
    xclSetEnvironment_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetEnvironment_call::GetClassData() const {
  return &_class_data_;
}

void xclSetEnvironment_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_call*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_environment()->MergeFrom(
      from._internal_environment());
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_call::CopyFrom(const xclSetEnvironment_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetEnvironment_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetEnvironment_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetEnvironment_call::InternalSwap(xclSetEnvironment_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.environment_.InternalSwap(&other->_impl_.environment_);
}

::google::protobuf::Metadata xclSetEnvironment_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[3]);
}
// ===================================================================

class xclSetEnvironment_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetEnvironment_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSetEnvironment_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetEnvironment_response::xclSetEnvironment_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetEnvironment_response)
}
xclSetEnvironment_response::xclSetEnvironment_response(
    ::google::protobuf::Arena* arena, const xclSetEnvironment_response& from)
    : xclSetEnvironment_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclSetEnvironment_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclSetEnvironment_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclSetEnvironment_response::~xclSetEnvironment_response() {
  // @@protoc_insertion_point(destructor:xclSetEnvironment_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetEnvironment_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetEnvironment_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetEnvironment_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetEnvironment_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclSetEnvironment_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSetEnvironment_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSetEnvironment_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclSetEnvironment_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSetEnvironment_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclSetEnvironment_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSetEnvironment_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetEnvironment_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetEnvironment_response)
  return target;
}

::size_t xclSetEnvironment_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetEnvironment_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetEnvironment_response::_class_data_ = {
    xclSetEnvironment_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetEnvironment_response::GetClassData() const {
  return &_class_data_;
}

void xclSetEnvironment_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetEnvironment_response*>(&to_msg);
  auto& from = static_cast<const xclSetEnvironment_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetEnvironment_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetEnvironment_response::CopyFrom(const xclSetEnvironment_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetEnvironment_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetEnvironment_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetEnvironment_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetEnvironment_response::InternalSwap(xclSetEnvironment_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclSetEnvironment_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[4]);
}
// ===================================================================

class xclLoadBitstream_call_ddrbank::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_call_ddrbank>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call_ddrbank, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_call.ddrbank)
}
xclLoadBitstream_call_ddrbank::xclLoadBitstream_call_ddrbank(
    ::google::protobuf::Arena* arena, const xclLoadBitstream_call_ddrbank& from)
    : xclLoadBitstream_call_ddrbank(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclLoadBitstream_call_ddrbank::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclLoadBitstream_call_ddrbank::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclLoadBitstream_call_ddrbank::~xclLoadBitstream_call_ddrbank() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_call.ddrbank)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclLoadBitstream_call_ddrbank::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclLoadBitstream_call_ddrbank::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_call.ddrbank)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclLoadBitstream_call_ddrbank::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclLoadBitstream_call_ddrbank::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call_ddrbank, _impl_._has_bits_),
    0, // no _extensions_
    7, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967231,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclLoadBitstream_call_ddrbank_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 size = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclLoadBitstream_call_ddrbank, _impl_.size_), 0>(),
     {56, 0, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call_ddrbank, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 size = 7;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call_ddrbank, _impl_.size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclLoadBitstream_call_ddrbank::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_call.ddrbank)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 size = 7;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        7, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_call.ddrbank)
  return target;
}

::size_t xclLoadBitstream_call_ddrbank::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_call.ddrbank)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 size = 7;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclLoadBitstream_call_ddrbank::_class_data_ = {
    xclLoadBitstream_call_ddrbank::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclLoadBitstream_call_ddrbank::GetClassData() const {
  return &_class_data_;
}

void xclLoadBitstream_call_ddrbank::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_call_ddrbank*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_call_ddrbank&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_call.ddrbank)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_call_ddrbank::CopyFrom(const xclLoadBitstream_call_ddrbank& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_call.ddrbank)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclLoadBitstream_call_ddrbank::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclLoadBitstream_call_ddrbank::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclLoadBitstream_call_ddrbank::InternalSwap(xclLoadBitstream_call_ddrbank* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclLoadBitstream_call_ddrbank::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[5]);
}
// ===================================================================

class xclLoadBitstream_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_._has_bits_);
  static void set_has_xmlfile(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dlopenfilename(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_devicename(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_devicedirectory(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_binarydirectory(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_verbose(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000003f) ^ 0x0000003f) != 0;
  }
};

xclLoadBitstream_call::xclLoadBitstream_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_call)
}
inline PROTOBUF_NDEBUG_INLINE xclLoadBitstream_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        ddrbanks_{visibility, arena, from.ddrbanks_},
        xmlfile_(arena, from.xmlfile_),
        dlopenfilename_(arena, from.dlopenfilename_),
        devicename_(arena, from.devicename_),
        devicedirectory_(arena, from.devicedirectory_),
        binarydirectory_(arena, from.binarydirectory_) {}

xclLoadBitstream_call::xclLoadBitstream_call(
    ::google::protobuf::Arena* arena,
    const xclLoadBitstream_call& from)
    : ::google::protobuf::Message(arena) {
  xclLoadBitstream_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.verbose_ = from._impl_.verbose_;

  // @@protoc_insertion_point(copy_constructor:xclLoadBitstream_call)
}
inline PROTOBUF_NDEBUG_INLINE xclLoadBitstream_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        ddrbanks_{visibility, arena},
        xmlfile_(arena),
        dlopenfilename_(arena),
        devicename_(arena),
        devicedirectory_(arena),
        binarydirectory_(arena) {}

inline void xclLoadBitstream_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.verbose_ = {};
}
xclLoadBitstream_call::~xclLoadBitstream_call() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclLoadBitstream_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.xmlfile_.Destroy();
  _impl_.dlopenfilename_.Destroy();
  _impl_.devicename_.Destroy();
  _impl_.devicedirectory_.Destroy();
  _impl_.binarydirectory_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclLoadBitstream_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ddrbanks_.Clear();
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xmlfile_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.dlopenfilename_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.devicename_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000008u) {
      _impl_.devicedirectory_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000010u) {
      _impl_.binarydirectory_.ClearNonDefaultToEmpty();
    }
  }
  _impl_.verbose_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclLoadBitstream_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 7, 1, 0, 2> xclLoadBitstream_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_._has_bits_),
    0, // no _extensions_
    8, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967104,  // skipmap
    offsetof(decltype(_table_), field_entries),
    7,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclLoadBitstream_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
    {::_pbi::TcParser::FastMtR1,
     {66, 63, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.ddrbanks_)}},
    // required bytes xmlfile = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.xmlfile_)}},
    // required bytes dlopenfilename = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 1, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.dlopenfilename_)}},
    // required bytes devicename = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 2, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.devicename_)}},
    // required bytes devicedirectory = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 3, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.devicedirectory_)}},
    // required bytes binarydirectory = 5;
    {::_pbi::TcParser::FastBS1,
     {42, 4, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.binarydirectory_)}},
    // required bool verbose = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclLoadBitstream_call, _impl_.verbose_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.verbose_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes xmlfile = 1;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.xmlfile_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes dlopenfilename = 2;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.dlopenfilename_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes devicename = 3;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.devicename_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes devicedirectory = 4;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.devicedirectory_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes binarydirectory = 5;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.binarydirectory_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bool verbose = 6;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.verbose_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_call, _impl_.ddrbanks_), -1, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclLoadBitstream_call_ddrbank>()},
  }}, {{
  }},
};

::uint8_t* xclLoadBitstream_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xmlfile = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_xmlfile();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  // required bytes dlopenfilename = 2;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_dlopenfilename();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // required bytes devicename = 3;
  if (cached_has_bits & 0x00000004u) {
    const std::string& _s = this->_internal_devicename();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // required bytes devicedirectory = 4;
  if (cached_has_bits & 0x00000008u) {
    const std::string& _s = this->_internal_devicedirectory();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // required bytes binarydirectory = 5;
  if (cached_has_bits & 0x00000010u) {
    const std::string& _s = this->_internal_binarydirectory();
    target = stream->WriteBytesMaybeAliased(5, _s, target);
  }

  // required bool verbose = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        6, this->_internal_verbose(), target);
  }

  // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_ddrbanks_size()); i < n; i++) {
    const auto& repfield = this->_internal_ddrbanks().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_call)
  return target;
}

::size_t xclLoadBitstream_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclLoadBitstream_call.ddrbank ddrbanks = 8;
  total_size += 1UL * this->_internal_ddrbanks_size();
  for (const auto& msg : this->_internal_ddrbanks()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // required bytes xmlfile = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_xmlfile());
    }

    // required bytes dlopenfilename = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dlopenfilename());
    }

    // required bytes devicename = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_devicename());
    }

    // required bytes devicedirectory = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_devicedirectory());
    }

    // required bytes binarydirectory = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_binarydirectory());
    }

    // required bool verbose = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclLoadBitstream_call::_class_data_ = {
    xclLoadBitstream_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclLoadBitstream_call::GetClassData() const {
  return &_class_data_;
}

void xclLoadBitstream_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_call*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_ddrbanks()->MergeFrom(
      from._internal_ddrbanks());
  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xmlfile(from._internal_xmlfile());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_dlopenfilename(from._internal_dlopenfilename());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_devicename(from._internal_devicename());
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_internal_set_devicedirectory(from._internal_devicedirectory());
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_internal_set_binarydirectory(from._internal_binarydirectory());
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.verbose_ = from._impl_.verbose_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_call::CopyFrom(const xclLoadBitstream_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclLoadBitstream_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclLoadBitstream_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclLoadBitstream_call::InternalSwap(xclLoadBitstream_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.ddrbanks_.InternalSwap(&other->_impl_.ddrbanks_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.xmlfile_, &other->_impl_.xmlfile_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dlopenfilename_, &other->_impl_.dlopenfilename_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.devicename_, &other->_impl_.devicename_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.devicedirectory_, &other->_impl_.devicedirectory_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.binarydirectory_, &other->_impl_.binarydirectory_, arena);
        swap(_impl_.verbose_, other->_impl_.verbose_);
}

::google::protobuf::Metadata xclLoadBitstream_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[6]);
}
// ===================================================================

class xclLoadBitstream_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadBitstream_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclLoadBitstream_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclLoadBitstream_response::xclLoadBitstream_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclLoadBitstream_response)
}
xclLoadBitstream_response::xclLoadBitstream_response(
    ::google::protobuf::Arena* arena, const xclLoadBitstream_response& from)
    : xclLoadBitstream_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclLoadBitstream_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclLoadBitstream_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclLoadBitstream_response::~xclLoadBitstream_response() {
  // @@protoc_insertion_point(destructor:xclLoadBitstream_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclLoadBitstream_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclLoadBitstream_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadBitstream_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclLoadBitstream_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclLoadBitstream_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclLoadBitstream_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclLoadBitstream_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclLoadBitstream_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclLoadBitstream_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclLoadBitstream_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclLoadBitstream_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadBitstream_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadBitstream_response)
  return target;
}

::size_t xclLoadBitstream_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadBitstream_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclLoadBitstream_response::_class_data_ = {
    xclLoadBitstream_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclLoadBitstream_response::GetClassData() const {
  return &_class_data_;
}

void xclLoadBitstream_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclLoadBitstream_response*>(&to_msg);
  auto& from = static_cast<const xclLoadBitstream_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadBitstream_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadBitstream_response::CopyFrom(const xclLoadBitstream_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadBitstream_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclLoadBitstream_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclLoadBitstream_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclLoadBitstream_response::InternalSwap(xclLoadBitstream_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclLoadBitstream_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[7]);
}
// ===================================================================

class xclAllocDeviceBuffer_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclAllocDeviceBuffer_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_._has_bits_);
  static void set_has_ddraddress(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_peertopeer(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclAllocDeviceBuffer_call)
}
xclAllocDeviceBuffer_call::xclAllocDeviceBuffer_call(
    ::google::protobuf::Arena* arena, const xclAllocDeviceBuffer_call& from)
    : xclAllocDeviceBuffer_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclAllocDeviceBuffer_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclAllocDeviceBuffer_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ddraddress_),
           0,
           offsetof(Impl_, peertopeer_) -
               offsetof(Impl_, ddraddress_) +
               sizeof(Impl_::peertopeer_));
}
xclAllocDeviceBuffer_call::~xclAllocDeviceBuffer_call() {
  // @@protoc_insertion_point(destructor:xclAllocDeviceBuffer_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclAllocDeviceBuffer_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclAllocDeviceBuffer_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclAllocDeviceBuffer_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    ::memset(&_impl_.ddraddress_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.peertopeer_) -
        reinterpret_cast<char*>(&_impl_.ddraddress_)) + sizeof(_impl_.peertopeer_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclAllocDeviceBuffer_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 0, 2> xclAllocDeviceBuffer_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclAllocDeviceBuffer_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 ddraddress = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclAllocDeviceBuffer_call, _impl_.ddraddress_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.ddraddress_)}},
    // required uint64 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclAllocDeviceBuffer_call, _impl_.size_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.size_)}},
    // optional bool peertopeer = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclAllocDeviceBuffer_call, _impl_.peertopeer_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.peertopeer_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 ddraddress = 1;
    {PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.ddraddress_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bool peertopeer = 3;
    {PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.peertopeer_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclAllocDeviceBuffer_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclAllocDeviceBuffer_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 ddraddress = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_ddraddress(), target);
  }

  // required uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_size(), target);
  }

  // optional bool peertopeer = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        3, this->_internal_peertopeer(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclAllocDeviceBuffer_call)
  return target;
}

::size_t xclAllocDeviceBuffer_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclAllocDeviceBuffer_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required uint64 ddraddress = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_ddraddress());
    }

    // required uint64 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  // optional bool peertopeer = 3;
  if (cached_has_bits & 0x00000004u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclAllocDeviceBuffer_call::_class_data_ = {
    xclAllocDeviceBuffer_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclAllocDeviceBuffer_call::GetClassData() const {
  return &_class_data_;
}

void xclAllocDeviceBuffer_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclAllocDeviceBuffer_call*>(&to_msg);
  auto& from = static_cast<const xclAllocDeviceBuffer_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclAllocDeviceBuffer_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.ddraddress_ = from._impl_.ddraddress_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.peertopeer_ = from._impl_.peertopeer_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclAllocDeviceBuffer_call::CopyFrom(const xclAllocDeviceBuffer_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclAllocDeviceBuffer_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclAllocDeviceBuffer_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclAllocDeviceBuffer_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclAllocDeviceBuffer_call::InternalSwap(xclAllocDeviceBuffer_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.peertopeer_)
      + sizeof(xclAllocDeviceBuffer_call::_impl_.peertopeer_)
      - PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_call, _impl_.ddraddress_)>(
          reinterpret_cast<char*>(&_impl_.ddraddress_),
          reinterpret_cast<char*>(&other->_impl_.ddraddress_));
}

::google::protobuf::Metadata xclAllocDeviceBuffer_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[8]);
}
// ===================================================================

class xclAllocDeviceBuffer_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclAllocDeviceBuffer_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclAllocDeviceBuffer_response)
}
inline PROTOBUF_NDEBUG_INLINE xclAllocDeviceBuffer_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        filename_(arena, from.filename_) {}

xclAllocDeviceBuffer_response::xclAllocDeviceBuffer_response(
    ::google::protobuf::Arena* arena,
    const xclAllocDeviceBuffer_response& from)
    : ::google::protobuf::Message(arena) {
  xclAllocDeviceBuffer_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.ack_ = from._impl_.ack_;

  // @@protoc_insertion_point(copy_constructor:xclAllocDeviceBuffer_response)
}
inline PROTOBUF_NDEBUG_INLINE xclAllocDeviceBuffer_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        filename_(arena) {}

inline void xclAllocDeviceBuffer_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclAllocDeviceBuffer_response::~xclAllocDeviceBuffer_response() {
  // @@protoc_insertion_point(destructor:xclAllocDeviceBuffer_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclAllocDeviceBuffer_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.filename_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclAllocDeviceBuffer_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclAllocDeviceBuffer_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.filename_.ClearNonDefaultToEmpty();
  }
  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclAllocDeviceBuffer_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 46, 2> xclAllocDeviceBuffer_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclAllocDeviceBuffer_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional string filename = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_.filename_)}},
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclAllocDeviceBuffer_response, _impl_.ack_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_.ack_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional string filename = 2;
    {PROTOBUF_FIELD_OFFSET(xclAllocDeviceBuffer_response, _impl_.filename_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\35\0\10\0\0\0\0\0"
    "xclAllocDeviceBuffer_response"
    "filename"
  }},
};

::uint8_t* xclAllocDeviceBuffer_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclAllocDeviceBuffer_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  // optional string filename = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_filename();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclAllocDeviceBuffer_response.filename");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclAllocDeviceBuffer_response)
  return target;
}

::size_t xclAllocDeviceBuffer_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclAllocDeviceBuffer_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional string filename = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                    this->_internal_filename());
  }

  // required bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclAllocDeviceBuffer_response::_class_data_ = {
    xclAllocDeviceBuffer_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclAllocDeviceBuffer_response::GetClassData() const {
  return &_class_data_;
}

void xclAllocDeviceBuffer_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclAllocDeviceBuffer_response*>(&to_msg);
  auto& from = static_cast<const xclAllocDeviceBuffer_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclAllocDeviceBuffer_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_filename(from._internal_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclAllocDeviceBuffer_response::CopyFrom(const xclAllocDeviceBuffer_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclAllocDeviceBuffer_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclAllocDeviceBuffer_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclAllocDeviceBuffer_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclAllocDeviceBuffer_response::InternalSwap(xclAllocDeviceBuffer_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.filename_, &other->_impl_.filename_, arena);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclAllocDeviceBuffer_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[9]);
}
// ===================================================================

class xclFreeDeviceBuffer_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclFreeDeviceBuffer_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_call, _impl_._has_bits_);
  static void set_has_ddraddress(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclFreeDeviceBuffer_call)
}
xclFreeDeviceBuffer_call::xclFreeDeviceBuffer_call(
    ::google::protobuf::Arena* arena, const xclFreeDeviceBuffer_call& from)
    : xclFreeDeviceBuffer_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclFreeDeviceBuffer_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclFreeDeviceBuffer_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ddraddress_ = {};
}
xclFreeDeviceBuffer_call::~xclFreeDeviceBuffer_call() {
  // @@protoc_insertion_point(destructor:xclFreeDeviceBuffer_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclFreeDeviceBuffer_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclFreeDeviceBuffer_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclFreeDeviceBuffer_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ddraddress_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclFreeDeviceBuffer_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclFreeDeviceBuffer_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclFreeDeviceBuffer_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint64 ddraddress = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclFreeDeviceBuffer_call, _impl_.ddraddress_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_call, _impl_.ddraddress_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 ddraddress = 1;
    {PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_call, _impl_.ddraddress_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclFreeDeviceBuffer_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclFreeDeviceBuffer_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 ddraddress = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_ddraddress(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclFreeDeviceBuffer_call)
  return target;
}

::size_t xclFreeDeviceBuffer_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclFreeDeviceBuffer_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint64 ddraddress = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_ddraddress());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclFreeDeviceBuffer_call::_class_data_ = {
    xclFreeDeviceBuffer_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclFreeDeviceBuffer_call::GetClassData() const {
  return &_class_data_;
}

void xclFreeDeviceBuffer_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclFreeDeviceBuffer_call*>(&to_msg);
  auto& from = static_cast<const xclFreeDeviceBuffer_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclFreeDeviceBuffer_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ddraddress(from._internal_ddraddress());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclFreeDeviceBuffer_call::CopyFrom(const xclFreeDeviceBuffer_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclFreeDeviceBuffer_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclFreeDeviceBuffer_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclFreeDeviceBuffer_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclFreeDeviceBuffer_call::InternalSwap(xclFreeDeviceBuffer_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ddraddress_, other->_impl_.ddraddress_);
}

::google::protobuf::Metadata xclFreeDeviceBuffer_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[10]);
}
// ===================================================================

class xclFreeDeviceBuffer_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclFreeDeviceBuffer_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclFreeDeviceBuffer_response)
}
xclFreeDeviceBuffer_response::xclFreeDeviceBuffer_response(
    ::google::protobuf::Arena* arena, const xclFreeDeviceBuffer_response& from)
    : xclFreeDeviceBuffer_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclFreeDeviceBuffer_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclFreeDeviceBuffer_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclFreeDeviceBuffer_response::~xclFreeDeviceBuffer_response() {
  // @@protoc_insertion_point(destructor:xclFreeDeviceBuffer_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclFreeDeviceBuffer_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclFreeDeviceBuffer_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclFreeDeviceBuffer_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclFreeDeviceBuffer_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclFreeDeviceBuffer_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclFreeDeviceBuffer_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclFreeDeviceBuffer_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclFreeDeviceBuffer_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclFreeDeviceBuffer_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclFreeDeviceBuffer_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclFreeDeviceBuffer_response)
  return target;
}

::size_t xclFreeDeviceBuffer_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclFreeDeviceBuffer_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclFreeDeviceBuffer_response::_class_data_ = {
    xclFreeDeviceBuffer_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclFreeDeviceBuffer_response::GetClassData() const {
  return &_class_data_;
}

void xclFreeDeviceBuffer_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclFreeDeviceBuffer_response*>(&to_msg);
  auto& from = static_cast<const xclFreeDeviceBuffer_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclFreeDeviceBuffer_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclFreeDeviceBuffer_response::CopyFrom(const xclFreeDeviceBuffer_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclFreeDeviceBuffer_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclFreeDeviceBuffer_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclFreeDeviceBuffer_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclFreeDeviceBuffer_response::InternalSwap(xclFreeDeviceBuffer_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclFreeDeviceBuffer_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[11]);
}
// ===================================================================

class xclClose_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclClose_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_closeall(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclClose_call::xclClose_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclClose_call)
}
inline PROTOBUF_NDEBUG_INLINE xclClose_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        xcldevicehandle_(arena, from.xcldevicehandle_) {}

xclClose_call::xclClose_call(
    ::google::protobuf::Arena* arena,
    const xclClose_call& from)
    : ::google::protobuf::Message(arena) {
  xclClose_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.closeall_ = from._impl_.closeall_;

  // @@protoc_insertion_point(copy_constructor:xclClose_call)
}
inline PROTOBUF_NDEBUG_INLINE xclClose_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        xcldevicehandle_(arena) {}

inline void xclClose_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.closeall_ = {};
}
xclClose_call::~xclClose_call() {
  // @@protoc_insertion_point(destructor:xclClose_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclClose_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclClose_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclClose_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
  }
  _impl_.closeall_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclClose_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclClose_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclClose_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bytes xclDeviceHandle = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_.xcldevicehandle_)}},
    // optional bool closeall = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclClose_call, _impl_.closeall_), 1>(),
     {24, 1, 0, PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_.closeall_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bytes xclDeviceHandle = 2;
    {PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_.xcldevicehandle_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional bool closeall = 3;
    {PROTOBUF_FIELD_OFFSET(xclClose_call, _impl_.closeall_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclClose_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclClose_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_xcldevicehandle();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // optional bool closeall = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        3, this->_internal_closeall(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclClose_call)
  return target;
}

::size_t xclClose_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclClose_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes xclDeviceHandle = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_xcldevicehandle());
    }

    // optional bool closeall = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclClose_call::_class_data_ = {
    xclClose_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclClose_call::GetClassData() const {
  return &_class_data_;
}

void xclClose_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclClose_call*>(&to_msg);
  auto& from = static_cast<const xclClose_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclClose_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.closeall_ = from._impl_.closeall_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclClose_call::CopyFrom(const xclClose_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclClose_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclClose_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclClose_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclClose_call::InternalSwap(xclClose_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.xcldevicehandle_, &other->_impl_.xcldevicehandle_, arena);
        swap(_impl_.closeall_, other->_impl_.closeall_);
}

::google::protobuf::Metadata xclClose_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[12]);
}
// ===================================================================

class xclClose_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclClose_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclClose_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclClose_response::xclClose_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclClose_response)
}
xclClose_response::xclClose_response(
    ::google::protobuf::Arena* arena, const xclClose_response& from)
    : xclClose_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclClose_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclClose_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclClose_response::~xclClose_response() {
  // @@protoc_insertion_point(destructor:xclClose_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclClose_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclClose_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclClose_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclClose_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclClose_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclClose_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclClose_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclClose_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclClose_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclClose_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclClose_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclClose_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclClose_response)
  return target;
}

::size_t xclClose_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclClose_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclClose_response::_class_data_ = {
    xclClose_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclClose_response::GetClassData() const {
  return &_class_data_;
}

void xclClose_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclClose_response*>(&to_msg);
  auto& from = static_cast<const xclClose_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclClose_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclClose_response::CopyFrom(const xclClose_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclClose_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclClose_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclClose_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclClose_response::InternalSwap(xclClose_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclClose_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[13]);
}
// ===================================================================

class xclCopyBufferHost2Device_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferHost2Device_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_seek(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_space(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferHost2Device_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferHost2Device_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        xcldevicehandle_(arena, from.xcldevicehandle_),
        src_(arena, from.src_) {}

xclCopyBufferHost2Device_call::xclCopyBufferHost2Device_call(
    ::google::protobuf::Arena* arena,
    const xclCopyBufferHost2Device_call& from)
    : ::google::protobuf::Message(arena) {
  xclCopyBufferHost2Device_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, dest_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, dest_),
           offsetof(Impl_, space_) -
               offsetof(Impl_, dest_) +
               sizeof(Impl_::space_));

  // @@protoc_insertion_point(copy_constructor:xclCopyBufferHost2Device_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferHost2Device_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        xcldevicehandle_(arena),
        src_(arena) {}

inline void xclCopyBufferHost2Device_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, dest_),
           0,
           offsetof(Impl_, space_) -
               offsetof(Impl_, dest_) +
               sizeof(Impl_::space_));
}
xclCopyBufferHost2Device_call::~xclCopyBufferHost2Device_call() {
  // @@protoc_insertion_point(destructor:xclCopyBufferHost2Device_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBufferHost2Device_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
  _impl_.src_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBufferHost2Device_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferHost2Device_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.src_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000003cu) {
    ::memset(&_impl_.dest_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.space_) -
        reinterpret_cast<char*>(&_impl_.dest_)) + sizeof(_impl_.space_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBufferHost2Device_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclCopyBufferHost2Device_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_._has_bits_),
    0, // no _extensions_
    7, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967169,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBufferHost2Device_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required bytes xclDeviceHandle = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.xcldevicehandle_)}},
    // required uint64 dest = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferHost2Device_call, _impl_.dest_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.dest_)}},
    // required bytes src = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 1, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.src_)}},
    // required uint64 size = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferHost2Device_call, _impl_.size_), 3>(),
     {40, 3, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.size_)}},
    // required uint64 seek = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferHost2Device_call, _impl_.seek_), 4>(),
     {48, 4, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.seek_)}},
    // optional uint32 space = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCopyBufferHost2Device_call, _impl_.space_), 5>(),
     {56, 5, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.space_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes xclDeviceHandle = 2;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.xcldevicehandle_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 dest = 3;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.dest_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bytes src = 4;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.src_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 size = 5;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.size_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 seek = 6;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.seek_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint32 space = 7;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.space_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCopyBufferHost2Device_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferHost2Device_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_xcldevicehandle();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // required uint64 dest = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_dest(), target);
  }

  // required bytes src = 4;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_src();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_size(), target);
  }

  // required uint64 seek = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        6, this->_internal_seek(), target);
  }

  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        7, this->_internal_space(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferHost2Device_call)
  return target;
}

::size_t xclCopyBufferHost2Device_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferHost2Device_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // required bytes xclDeviceHandle = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_xcldevicehandle());
    }

    // required bytes src = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_src());
    }

    // required uint64 dest = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_dest());
    }

    // required uint64 size = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // required uint64 seek = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_seek());
    }

  }
  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_space());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBufferHost2Device_call::_class_data_ = {
    xclCopyBufferHost2Device_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBufferHost2Device_call::GetClassData() const {
  return &_class_data_;
}

void xclCopyBufferHost2Device_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferHost2Device_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferHost2Device_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferHost2Device_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_src(from._internal_src());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.dest_ = from._impl_.dest_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.seek_ = from._impl_.seek_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.space_ = from._impl_.space_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferHost2Device_call::CopyFrom(const xclCopyBufferHost2Device_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferHost2Device_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBufferHost2Device_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclCopyBufferHost2Device_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBufferHost2Device_call::InternalSwap(xclCopyBufferHost2Device_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.xcldevicehandle_, &other->_impl_.xcldevicehandle_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.src_, &other->_impl_.src_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.space_)
      + sizeof(xclCopyBufferHost2Device_call::_impl_.space_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_call, _impl_.dest_)>(
          reinterpret_cast<char*>(&_impl_.dest_),
          reinterpret_cast<char*>(&other->_impl_.dest_));
}

::google::protobuf::Metadata xclCopyBufferHost2Device_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[14]);
}
// ===================================================================

class xclCopyBufferHost2Device_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferHost2Device_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_response, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferHost2Device_response)
}
xclCopyBufferHost2Device_response::xclCopyBufferHost2Device_response(
    ::google::protobuf::Arena* arena, const xclCopyBufferHost2Device_response& from)
    : xclCopyBufferHost2Device_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferHost2Device_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclCopyBufferHost2Device_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclCopyBufferHost2Device_response::~xclCopyBufferHost2Device_response() {
  // @@protoc_insertion_point(destructor:xclCopyBufferHost2Device_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBufferHost2Device_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBufferHost2Device_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferHost2Device_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBufferHost2Device_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclCopyBufferHost2Device_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBufferHost2Device_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint64 size = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferHost2Device_response, _impl_.size_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_response, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 size = 1;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferHost2Device_response, _impl_.size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCopyBufferHost2Device_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferHost2Device_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferHost2Device_response)
  return target;
}

::size_t xclCopyBufferHost2Device_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferHost2Device_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint64 size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBufferHost2Device_response::_class_data_ = {
    xclCopyBufferHost2Device_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBufferHost2Device_response::GetClassData() const {
  return &_class_data_;
}

void xclCopyBufferHost2Device_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferHost2Device_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferHost2Device_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferHost2Device_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferHost2Device_response::CopyFrom(const xclCopyBufferHost2Device_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferHost2Device_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBufferHost2Device_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclCopyBufferHost2Device_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBufferHost2Device_response::InternalSwap(xclCopyBufferHost2Device_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclCopyBufferHost2Device_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[15]);
}
// ===================================================================

class xclCopyBufferDevice2Host_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferDevice2Host_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_._has_bits_);
  static void set_has_xcldevicehandle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_skip(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_space(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferDevice2Host_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferDevice2Host_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        xcldevicehandle_(arena, from.xcldevicehandle_),
        dest_(arena, from.dest_) {}

xclCopyBufferDevice2Host_call::xclCopyBufferDevice2Host_call(
    ::google::protobuf::Arena* arena,
    const xclCopyBufferDevice2Host_call& from)
    : ::google::protobuf::Message(arena) {
  xclCopyBufferDevice2Host_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, src_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, src_),
           offsetof(Impl_, space_) -
               offsetof(Impl_, src_) +
               sizeof(Impl_::space_));

  // @@protoc_insertion_point(copy_constructor:xclCopyBufferDevice2Host_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferDevice2Host_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        xcldevicehandle_(arena),
        dest_(arena) {}

inline void xclCopyBufferDevice2Host_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, src_),
           0,
           offsetof(Impl_, space_) -
               offsetof(Impl_, src_) +
               sizeof(Impl_::space_));
}
xclCopyBufferDevice2Host_call::~xclCopyBufferDevice2Host_call() {
  // @@protoc_insertion_point(destructor:xclCopyBufferDevice2Host_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBufferDevice2Host_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.xcldevicehandle_.Destroy();
  _impl_.dest_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBufferDevice2Host_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferDevice2Host_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xcldevicehandle_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.dest_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000003cu) {
    ::memset(&_impl_.src_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.space_) -
        reinterpret_cast<char*>(&_impl_.src_)) + sizeof(_impl_.space_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBufferDevice2Host_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclCopyBufferDevice2Host_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_._has_bits_),
    0, // no _extensions_
    7, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967169,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBufferDevice2Host_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required bytes xclDeviceHandle = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.xcldevicehandle_)}},
    // required bytes dest = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 1, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.dest_)}},
    // required uint64 src = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferDevice2Host_call, _impl_.src_), 2>(),
     {32, 2, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.src_)}},
    // required uint64 size = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferDevice2Host_call, _impl_.size_), 3>(),
     {40, 3, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.size_)}},
    // required uint64 skip = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferDevice2Host_call, _impl_.skip_), 4>(),
     {48, 4, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.skip_)}},
    // optional uint32 space = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCopyBufferDevice2Host_call, _impl_.space_), 5>(),
     {56, 5, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.space_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes xclDeviceHandle = 2;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.xcldevicehandle_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes dest = 3;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.dest_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 src = 4;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.src_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 size = 5;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.size_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 skip = 6;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.skip_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint32 space = 7;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.space_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCopyBufferDevice2Host_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferDevice2Host_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xclDeviceHandle = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_xcldevicehandle();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // required bytes dest = 3;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_dest();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // required uint64 src = 4;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_src(), target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_size(), target);
  }

  // required uint64 skip = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        6, this->_internal_skip(), target);
  }

  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        7, this->_internal_space(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferDevice2Host_call)
  return target;
}

::size_t xclCopyBufferDevice2Host_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferDevice2Host_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // required bytes xclDeviceHandle = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_xcldevicehandle());
    }

    // required bytes dest = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dest());
    }

    // required uint64 src = 4;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_src());
    }

    // required uint64 size = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // required uint64 skip = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_skip());
    }

  }
  // optional uint32 space = 7;
  if (cached_has_bits & 0x00000020u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_space());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBufferDevice2Host_call::_class_data_ = {
    xclCopyBufferDevice2Host_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBufferDevice2Host_call::GetClassData() const {
  return &_class_data_;
}

void xclCopyBufferDevice2Host_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferDevice2Host_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferDevice2Host_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferDevice2Host_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xcldevicehandle(from._internal_xcldevicehandle());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.src_ = from._impl_.src_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.skip_ = from._impl_.skip_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.space_ = from._impl_.space_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferDevice2Host_call::CopyFrom(const xclCopyBufferDevice2Host_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferDevice2Host_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBufferDevice2Host_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclCopyBufferDevice2Host_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBufferDevice2Host_call::InternalSwap(xclCopyBufferDevice2Host_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.xcldevicehandle_, &other->_impl_.xcldevicehandle_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dest_, &other->_impl_.dest_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.space_)
      + sizeof(xclCopyBufferDevice2Host_call::_impl_.space_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_call, _impl_.src_)>(
          reinterpret_cast<char*>(&_impl_.src_),
          reinterpret_cast<char*>(&other->_impl_.src_));
}

::google::protobuf::Metadata xclCopyBufferDevice2Host_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[16]);
}
// ===================================================================

class xclCopyBufferDevice2Host_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBufferDevice2Host_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBufferDevice2Host_response)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferDevice2Host_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dest_(arena, from.dest_) {}

xclCopyBufferDevice2Host_response::xclCopyBufferDevice2Host_response(
    ::google::protobuf::Arena* arena,
    const xclCopyBufferDevice2Host_response& from)
    : ::google::protobuf::Message(arena) {
  xclCopyBufferDevice2Host_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.size_ = from._impl_.size_;

  // @@protoc_insertion_point(copy_constructor:xclCopyBufferDevice2Host_response)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBufferDevice2Host_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dest_(arena) {}

inline void xclCopyBufferDevice2Host_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclCopyBufferDevice2Host_response::~xclCopyBufferDevice2Host_response() {
  // @@protoc_insertion_point(destructor:xclCopyBufferDevice2Host_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBufferDevice2Host_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dest_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBufferDevice2Host_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBufferDevice2Host_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBufferDevice2Host_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclCopyBufferDevice2Host_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBufferDevice2Host_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes dest = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_.dest_)}},
    // required uint64 size = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBufferDevice2Host_response, _impl_.size_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 size = 1;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bytes dest = 2;
    {PROTOBUF_FIELD_OFFSET(xclCopyBufferDevice2Host_response, _impl_.dest_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCopyBufferDevice2Host_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBufferDevice2Host_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_size(), target);
  }

  // required bytes dest = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dest();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBufferDevice2Host_response)
  return target;
}

::size_t xclCopyBufferDevice2Host_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBufferDevice2Host_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required bytes dest = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dest());
    }

    // required uint64 size = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBufferDevice2Host_response::_class_data_ = {
    xclCopyBufferDevice2Host_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBufferDevice2Host_response::GetClassData() const {
  return &_class_data_;
}

void xclCopyBufferDevice2Host_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBufferDevice2Host_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBufferDevice2Host_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBufferDevice2Host_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBufferDevice2Host_response::CopyFrom(const xclCopyBufferDevice2Host_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBufferDevice2Host_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBufferDevice2Host_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclCopyBufferDevice2Host_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBufferDevice2Host_response::InternalSwap(xclCopyBufferDevice2Host_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dest_, &other->_impl_.dest_, arena);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclCopyBufferDevice2Host_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[17]);
}
// ===================================================================

class xclWriteAddrSpaceDeviceRam_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrSpaceDeviceRam_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrSpaceDeviceRam_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclWriteAddrSpaceDeviceRam_call::xclWriteAddrSpaceDeviceRam_call(
    ::google::protobuf::Arena* arena,
    const xclWriteAddrSpaceDeviceRam_call& from)
    : ::google::protobuf::Message(arena) {
  xclWriteAddrSpaceDeviceRam_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclWriteAddrSpaceDeviceRam_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclWriteAddrSpaceDeviceRam_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));
}
xclWriteAddrSpaceDeviceRam_call::~xclWriteAddrSpaceDeviceRam_call() {
  // @@protoc_insertion_point(destructor:xclWriteAddrSpaceDeviceRam_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteAddrSpaceDeviceRam_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteAddrSpaceDeviceRam_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrSpaceDeviceRam_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteAddrSpaceDeviceRam_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclWriteAddrSpaceDeviceRam_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967233,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteAddrSpaceDeviceRam_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 addr = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteAddrSpaceDeviceRam_call, _impl_.addr_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.addr_)}},
    // required uint32 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrSpaceDeviceRam_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.size_)}},
    // required bytes data = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.data_)}},
    // optional uint32 PF_ID = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrSpaceDeviceRam_call, _impl_.pf_id_), 3>(),
     {40, 3, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_), 4>(),
     {48, 4, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 2;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes data = 4;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint32 PF_ID = 5;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 6;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclWriteAddrSpaceDeviceRam_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrSpaceDeviceRam_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_size(), target);
  }

  // required bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrSpaceDeviceRam_call)
  return target;
}

::size_t xclWriteAddrSpaceDeviceRam_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrSpaceDeviceRam_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // required bytes data = 4;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_data());
    }

    // required uint64 addr = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteAddrSpaceDeviceRam_call::_class_data_ = {
    xclWriteAddrSpaceDeviceRam_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteAddrSpaceDeviceRam_call::GetClassData() const {
  return &_class_data_;
}

void xclWriteAddrSpaceDeviceRam_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrSpaceDeviceRam_call*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrSpaceDeviceRam_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrSpaceDeviceRam_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrSpaceDeviceRam_call::CopyFrom(const xclWriteAddrSpaceDeviceRam_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrSpaceDeviceRam_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteAddrSpaceDeviceRam_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclWriteAddrSpaceDeviceRam_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteAddrSpaceDeviceRam_call::InternalSwap(xclWriteAddrSpaceDeviceRam_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.bar_id_)
      + sizeof(xclWriteAddrSpaceDeviceRam_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclWriteAddrSpaceDeviceRam_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[18]);
}
// ===================================================================

class xclWriteAddrSpaceDeviceRam_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrSpaceDeviceRam_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrSpaceDeviceRam_response)
}
xclWriteAddrSpaceDeviceRam_response::xclWriteAddrSpaceDeviceRam_response(
    ::google::protobuf::Arena* arena, const xclWriteAddrSpaceDeviceRam_response& from)
    : xclWriteAddrSpaceDeviceRam_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrSpaceDeviceRam_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclWriteAddrSpaceDeviceRam_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclWriteAddrSpaceDeviceRam_response::~xclWriteAddrSpaceDeviceRam_response() {
  // @@protoc_insertion_point(destructor:xclWriteAddrSpaceDeviceRam_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteAddrSpaceDeviceRam_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteAddrSpaceDeviceRam_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrSpaceDeviceRam_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteAddrSpaceDeviceRam_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclWriteAddrSpaceDeviceRam_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteAddrSpaceDeviceRam_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclWriteAddrSpaceDeviceRam_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrSpaceDeviceRam_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclWriteAddrSpaceDeviceRam_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrSpaceDeviceRam_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrSpaceDeviceRam_response)
  return target;
}

::size_t xclWriteAddrSpaceDeviceRam_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrSpaceDeviceRam_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteAddrSpaceDeviceRam_response::_class_data_ = {
    xclWriteAddrSpaceDeviceRam_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteAddrSpaceDeviceRam_response::GetClassData() const {
  return &_class_data_;
}

void xclWriteAddrSpaceDeviceRam_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrSpaceDeviceRam_response*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrSpaceDeviceRam_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrSpaceDeviceRam_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrSpaceDeviceRam_response::CopyFrom(const xclWriteAddrSpaceDeviceRam_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrSpaceDeviceRam_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteAddrSpaceDeviceRam_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclWriteAddrSpaceDeviceRam_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteAddrSpaceDeviceRam_response::InternalSwap(xclWriteAddrSpaceDeviceRam_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclWriteAddrSpaceDeviceRam_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[19]);
}
// ===================================================================

class xclWriteAddrKernelCtrl_call_kernelInfo::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_call_kernelInfo>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_name(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_call.kernelInfo)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrKernelCtrl_call_kernelInfo::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        name_(arena, from.name_) {}

xclWriteAddrKernelCtrl_call_kernelInfo::xclWriteAddrKernelCtrl_call_kernelInfo(
    ::google::protobuf::Arena* arena,
    const xclWriteAddrKernelCtrl_call_kernelInfo& from)
    : ::google::protobuf::Message(arena) {
  xclWriteAddrKernelCtrl_call_kernelInfo* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, size_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::size_));

  // @@protoc_insertion_point(copy_constructor:xclWriteAddrKernelCtrl_call.kernelInfo)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrKernelCtrl_call_kernelInfo::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        name_(arena) {}

inline void xclWriteAddrKernelCtrl_call_kernelInfo::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, size_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::size_));
}
xclWriteAddrKernelCtrl_call_kernelInfo::~xclWriteAddrKernelCtrl_call_kernelInfo() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_call.kernelInfo)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteAddrKernelCtrl_call_kernelInfo::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.name_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteAddrKernelCtrl_call_kernelInfo::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.name_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_call_kernelInfo::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 51, 2> xclWriteAddrKernelCtrl_call_kernelInfo::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteAddrKernelCtrl_call_kernelInfo_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 addr = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_)}},
    // optional uint32 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_), 2>(),
     {16, 2, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_)}},
    // optional string name = 3;
    {::_pbi::TcParser::FastSS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.name_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 addr = 1;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint32 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional string name = 3;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.name_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\46\0\0\4\0\0\0\0"
    "xclWriteAddrKernelCtrl_call.kernelInfo"
    "name"
  }},
};

::uint8_t* xclWriteAddrKernelCtrl_call_kernelInfo::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 addr = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_addr(), target);
  }

  // optional uint32 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_size(), target);
  }

  // optional string name = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_name();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclWriteAddrKernelCtrl_call.kernelInfo.name");
    target = stream->WriteStringMaybeAliased(3, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_call.kernelInfo)
  return target;
}

::size_t xclWriteAddrKernelCtrl_call_kernelInfo::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string name = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_name());
    }

    // optional uint64 addr = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // optional uint32 size = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteAddrKernelCtrl_call_kernelInfo::_class_data_ = {
    xclWriteAddrKernelCtrl_call_kernelInfo::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteAddrKernelCtrl_call_kernelInfo::GetClassData() const {
  return &_class_data_;
}

void xclWriteAddrKernelCtrl_call_kernelInfo::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_call_kernelInfo*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_call_kernelInfo&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_name(from._internal_name());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_call_kernelInfo::CopyFrom(const xclWriteAddrKernelCtrl_call_kernelInfo& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_call.kernelInfo)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteAddrKernelCtrl_call_kernelInfo::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclWriteAddrKernelCtrl_call_kernelInfo::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteAddrKernelCtrl_call_kernelInfo::InternalSwap(xclWriteAddrKernelCtrl_call_kernelInfo* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.name_, &other->_impl_.name_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.size_)
      + sizeof(xclWriteAddrKernelCtrl_call_kernelInfo::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call_kernelInfo, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclWriteAddrKernelCtrl_call_kernelInfo::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[20]);
}
// ===================================================================

class xclWriteAddrKernelCtrl_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrKernelCtrl_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        kernel_info_{visibility, arena, from.kernel_info_},
        data_(arena, from.data_) {}

xclWriteAddrKernelCtrl_call::xclWriteAddrKernelCtrl_call(
    ::google::protobuf::Arena* arena,
    const xclWriteAddrKernelCtrl_call& from)
    : ::google::protobuf::Message(arena) {
  xclWriteAddrKernelCtrl_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclWriteAddrKernelCtrl_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrKernelCtrl_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        kernel_info_{visibility, arena},
        data_(arena) {}

inline void xclWriteAddrKernelCtrl_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));
}
xclWriteAddrKernelCtrl_call::~xclWriteAddrKernelCtrl_call() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteAddrKernelCtrl_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteAddrKernelCtrl_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.kernel_info_.Clear();
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 1, 0, 2> xclWriteAddrKernelCtrl_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_._has_bits_),
    0, // no _extensions_
    7, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967169,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclWriteAddrKernelCtrl_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 addr = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteAddrKernelCtrl_call, _impl_.addr_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.addr_)}},
    // required uint32 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrKernelCtrl_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.size_)}},
    // required bytes data = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.data_)}},
    // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
    {::_pbi::TcParser::FastMtR1,
     {42, 63, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.kernel_info_)}},
    // optional uint32 PF_ID = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrKernelCtrl_call, _impl_.pf_id_), 3>(),
     {48, 3, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclWriteAddrKernelCtrl_call, _impl_.bar_id_), 4>(),
     {56, 4, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.bar_id_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 2;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes data = 4;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.kernel_info_), -1, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
    // optional uint32 PF_ID = 6;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 7;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclWriteAddrKernelCtrl_call_kernelInfo>()},
  }}, {{
  }},
};

::uint8_t* xclWriteAddrKernelCtrl_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_size(), target);
  }

  // required bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_kernel_info_size()); i < n; i++) {
    const auto& repfield = this->_internal_kernel_info().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(5, repfield, repfield.GetCachedSize(), target, stream);
  }

  // optional uint32 PF_ID = 6;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 7;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        7, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_call)
  return target;
}

::size_t xclWriteAddrKernelCtrl_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclWriteAddrKernelCtrl_call.kernelInfo kernel_info = 5;
  total_size += 1UL * this->_internal_kernel_info_size();
  for (const auto& msg : this->_internal_kernel_info()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // required bytes data = 4;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_data());
    }

    // required uint64 addr = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 6;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 7;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteAddrKernelCtrl_call::_class_data_ = {
    xclWriteAddrKernelCtrl_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteAddrKernelCtrl_call::GetClassData() const {
  return &_class_data_;
}

void xclWriteAddrKernelCtrl_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_call*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_kernel_info()->MergeFrom(
      from._internal_kernel_info());
  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_call::CopyFrom(const xclWriteAddrKernelCtrl_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteAddrKernelCtrl_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclWriteAddrKernelCtrl_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteAddrKernelCtrl_call::InternalSwap(xclWriteAddrKernelCtrl_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.kernel_info_.InternalSwap(&other->_impl_.kernel_info_);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.bar_id_)
      + sizeof(xclWriteAddrKernelCtrl_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclWriteAddrKernelCtrl_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[21]);
}
// ===================================================================

class xclWriteAddrKernelCtrl_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteAddrKernelCtrl_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteAddrKernelCtrl_response)
}
xclWriteAddrKernelCtrl_response::xclWriteAddrKernelCtrl_response(
    ::google::protobuf::Arena* arena, const xclWriteAddrKernelCtrl_response& from)
    : xclWriteAddrKernelCtrl_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclWriteAddrKernelCtrl_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclWriteAddrKernelCtrl_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclWriteAddrKernelCtrl_response::~xclWriteAddrKernelCtrl_response() {
  // @@protoc_insertion_point(destructor:xclWriteAddrKernelCtrl_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteAddrKernelCtrl_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteAddrKernelCtrl_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteAddrKernelCtrl_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteAddrKernelCtrl_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclWriteAddrKernelCtrl_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteAddrKernelCtrl_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclWriteAddrKernelCtrl_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclWriteAddrKernelCtrl_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclWriteAddrKernelCtrl_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteAddrKernelCtrl_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteAddrKernelCtrl_response)
  return target;
}

::size_t xclWriteAddrKernelCtrl_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteAddrKernelCtrl_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteAddrKernelCtrl_response::_class_data_ = {
    xclWriteAddrKernelCtrl_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteAddrKernelCtrl_response::GetClassData() const {
  return &_class_data_;
}

void xclWriteAddrKernelCtrl_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteAddrKernelCtrl_response*>(&to_msg);
  auto& from = static_cast<const xclWriteAddrKernelCtrl_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteAddrKernelCtrl_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteAddrKernelCtrl_response::CopyFrom(const xclWriteAddrKernelCtrl_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteAddrKernelCtrl_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteAddrKernelCtrl_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclWriteAddrKernelCtrl_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteAddrKernelCtrl_response::InternalSwap(xclWriteAddrKernelCtrl_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclWriteAddrKernelCtrl_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[22]);
}
// ===================================================================

class xclRegWrite_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegWrite_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_._has_bits_);
  static void set_has_baseaddress(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclRegWrite_call::xclRegWrite_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclRegWrite_call)
}
inline PROTOBUF_NDEBUG_INLINE xclRegWrite_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclRegWrite_call::xclRegWrite_call(
    ::google::protobuf::Arena* arena,
    const xclRegWrite_call& from)
    : ::google::protobuf::Message(arena) {
  xclRegWrite_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, baseaddress_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, baseaddress_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, baseaddress_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclRegWrite_call)
}
inline PROTOBUF_NDEBUG_INLINE xclRegWrite_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclRegWrite_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, baseaddress_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, baseaddress_) +
               sizeof(Impl_::bar_id_));
}
xclRegWrite_call::~xclRegWrite_call() {
  // @@protoc_insertion_point(destructor:xclRegWrite_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclRegWrite_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclRegWrite_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegWrite_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.baseaddress_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclRegWrite_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclRegWrite_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_._has_bits_),
    0, // no _extensions_
    5, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967264,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclRegWrite_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // required uint32 baseaddress = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegWrite_call, _impl_.baseaddress_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.baseaddress_)}},
    // required uint32 offset = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegWrite_call, _impl_.offset_), 2>(),
     {16, 2, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.offset_)}},
    // required bytes data = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.data_)}},
    // optional uint32 PF_ID = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegWrite_call, _impl_.pf_id_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegWrite_call, _impl_.bar_id_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.bar_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 baseaddress = 1;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.baseaddress_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint32 offset = 2;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.offset_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes data = 3;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint32 PF_ID = 4;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 5;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclRegWrite_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegWrite_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 baseaddress = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_baseaddress(), target);
  }

  // required uint32 offset = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_offset(), target);
  }

  // required bytes data = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // optional uint32 PF_ID = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegWrite_call)
  return target;
}

::size_t xclRegWrite_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegWrite_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // required bytes data = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_data());
    }

    // required uint32 baseaddress = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_baseaddress());
    }

    // required uint32 offset = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_offset());
    }

  }
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclRegWrite_call::_class_data_ = {
    xclRegWrite_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclRegWrite_call::GetClassData() const {
  return &_class_data_;
}

void xclRegWrite_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclRegWrite_call*>(&to_msg);
  auto& from = static_cast<const xclRegWrite_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegWrite_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.baseaddress_ = from._impl_.baseaddress_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegWrite_call::CopyFrom(const xclRegWrite_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegWrite_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclRegWrite_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclRegWrite_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclRegWrite_call::InternalSwap(xclRegWrite_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.bar_id_)
      + sizeof(xclRegWrite_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclRegWrite_call, _impl_.baseaddress_)>(
          reinterpret_cast<char*>(&_impl_.baseaddress_),
          reinterpret_cast<char*>(&other->_impl_.baseaddress_));
}

::google::protobuf::Metadata xclRegWrite_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[23]);
}
// ===================================================================

class xclRegWrite_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegWrite_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclRegWrite_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclRegWrite_response::xclRegWrite_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclRegWrite_response)
}
xclRegWrite_response::xclRegWrite_response(
    ::google::protobuf::Arena* arena, const xclRegWrite_response& from)
    : xclRegWrite_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclRegWrite_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclRegWrite_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclRegWrite_response::~xclRegWrite_response() {
  // @@protoc_insertion_point(destructor:xclRegWrite_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclRegWrite_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclRegWrite_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegWrite_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclRegWrite_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclRegWrite_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclRegWrite_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclRegWrite_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclRegWrite_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclRegWrite_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclRegWrite_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclRegWrite_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegWrite_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegWrite_response)
  return target;
}

::size_t xclRegWrite_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegWrite_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclRegWrite_response::_class_data_ = {
    xclRegWrite_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclRegWrite_response::GetClassData() const {
  return &_class_data_;
}

void xclRegWrite_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclRegWrite_response*>(&to_msg);
  auto& from = static_cast<const xclRegWrite_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegWrite_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegWrite_response::CopyFrom(const xclRegWrite_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegWrite_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclRegWrite_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclRegWrite_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclRegWrite_response::InternalSwap(xclRegWrite_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclRegWrite_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[24]);
}
// ===================================================================

class xclReadAddrSpaceDeviceRam_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrSpaceDeviceRam_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000006) ^ 0x00000006) != 0;
  }
};

xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrSpaceDeviceRam_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclReadAddrSpaceDeviceRam_call::xclReadAddrSpaceDeviceRam_call(
    ::google::protobuf::Arena* arena,
    const xclReadAddrSpaceDeviceRam_call& from)
    : ::google::protobuf::Message(arena) {
  xclReadAddrSpaceDeviceRam_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclReadAddrSpaceDeviceRam_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrSpaceDeviceRam_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclReadAddrSpaceDeviceRam_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));
}
xclReadAddrSpaceDeviceRam_call::~xclReadAddrSpaceDeviceRam_call() {
  // @@protoc_insertion_point(destructor:xclReadAddrSpaceDeviceRam_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadAddrSpaceDeviceRam_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadAddrSpaceDeviceRam_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrSpaceDeviceRam_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadAddrSpaceDeviceRam_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclReadAddrSpaceDeviceRam_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967233,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadAddrSpaceDeviceRam_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 addr = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadAddrSpaceDeviceRam_call, _impl_.addr_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.addr_)}},
    // required uint32 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrSpaceDeviceRam_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.size_)}},
    // optional bytes data = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.data_)}},
    // optional uint32 PF_ID = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrSpaceDeviceRam_call, _impl_.pf_id_), 3>(),
     {40, 3, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_), 4>(),
     {48, 4, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional bytes data = 4;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint32 PF_ID = 5;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 6;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadAddrSpaceDeviceRam_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrSpaceDeviceRam_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrSpaceDeviceRam_call)
  return target;
}

::size_t xclReadAddrSpaceDeviceRam_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrSpaceDeviceRam_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  if (cached_has_bits & 0x00000006u) {
    // required uint64 addr = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadAddrSpaceDeviceRam_call::_class_data_ = {
    xclReadAddrSpaceDeviceRam_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadAddrSpaceDeviceRam_call::GetClassData() const {
  return &_class_data_;
}

void xclReadAddrSpaceDeviceRam_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrSpaceDeviceRam_call*>(&to_msg);
  auto& from = static_cast<const xclReadAddrSpaceDeviceRam_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrSpaceDeviceRam_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrSpaceDeviceRam_call::CopyFrom(const xclReadAddrSpaceDeviceRam_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrSpaceDeviceRam_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadAddrSpaceDeviceRam_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclReadAddrSpaceDeviceRam_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadAddrSpaceDeviceRam_call::InternalSwap(xclReadAddrSpaceDeviceRam_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.bar_id_)
      + sizeof(xclReadAddrSpaceDeviceRam_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclReadAddrSpaceDeviceRam_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[25]);
}
// ===================================================================

class xclReadAddrSpaceDeviceRam_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrSpaceDeviceRam_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrSpaceDeviceRam_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrSpaceDeviceRam_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclReadAddrSpaceDeviceRam_response::xclReadAddrSpaceDeviceRam_response(
    ::google::protobuf::Arena* arena,
    const xclReadAddrSpaceDeviceRam_response& from)
    : ::google::protobuf::Message(arena) {
  xclReadAddrSpaceDeviceRam_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.valid_ = from._impl_.valid_;

  // @@protoc_insertion_point(copy_constructor:xclReadAddrSpaceDeviceRam_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrSpaceDeviceRam_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclReadAddrSpaceDeviceRam_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclReadAddrSpaceDeviceRam_response::~xclReadAddrSpaceDeviceRam_response() {
  // @@protoc_insertion_point(destructor:xclReadAddrSpaceDeviceRam_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadAddrSpaceDeviceRam_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadAddrSpaceDeviceRam_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrSpaceDeviceRam_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadAddrSpaceDeviceRam_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclReadAddrSpaceDeviceRam_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadAddrSpaceDeviceRam_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bytes data = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_.data_)}},
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclReadAddrSpaceDeviceRam_response, _impl_.valid_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_.valid_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bytes data = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrSpaceDeviceRam_response, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadAddrSpaceDeviceRam_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrSpaceDeviceRam_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrSpaceDeviceRam_response)
  return target;
}

::size_t xclReadAddrSpaceDeviceRam_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrSpaceDeviceRam_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadAddrSpaceDeviceRam_response::_class_data_ = {
    xclReadAddrSpaceDeviceRam_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadAddrSpaceDeviceRam_response::GetClassData() const {
  return &_class_data_;
}

void xclReadAddrSpaceDeviceRam_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrSpaceDeviceRam_response*>(&to_msg);
  auto& from = static_cast<const xclReadAddrSpaceDeviceRam_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrSpaceDeviceRam_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrSpaceDeviceRam_response::CopyFrom(const xclReadAddrSpaceDeviceRam_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrSpaceDeviceRam_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadAddrSpaceDeviceRam_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclReadAddrSpaceDeviceRam_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadAddrSpaceDeviceRam_response::InternalSwap(xclReadAddrSpaceDeviceRam_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclReadAddrSpaceDeviceRam_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[26]);
}
// ===================================================================

class xclReadAddrKernelCtrl_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrKernelCtrl_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000006) ^ 0x00000006) != 0;
  }
};

xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrKernelCtrl_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrKernelCtrl_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclReadAddrKernelCtrl_call::xclReadAddrKernelCtrl_call(
    ::google::protobuf::Arena* arena,
    const xclReadAddrKernelCtrl_call& from)
    : ::google::protobuf::Message(arena) {
  xclReadAddrKernelCtrl_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclReadAddrKernelCtrl_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrKernelCtrl_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclReadAddrKernelCtrl_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::bar_id_));
}
xclReadAddrKernelCtrl_call::~xclReadAddrKernelCtrl_call() {
  // @@protoc_insertion_point(destructor:xclReadAddrKernelCtrl_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadAddrKernelCtrl_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadAddrKernelCtrl_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrKernelCtrl_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadAddrKernelCtrl_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclReadAddrKernelCtrl_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967233,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadAddrKernelCtrl_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 addr = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadAddrKernelCtrl_call, _impl_.addr_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.addr_)}},
    // required uint32 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrKernelCtrl_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.size_)}},
    // optional bytes data = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.data_)}},
    // optional uint32 PF_ID = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrKernelCtrl_call, _impl_.pf_id_), 3>(),
     {40, 3, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadAddrKernelCtrl_call, _impl_.bar_id_), 4>(),
     {48, 4, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.bar_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional bytes data = 4;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint32 PF_ID = 5;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 6;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadAddrKernelCtrl_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrKernelCtrl_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_addr(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrKernelCtrl_call)
  return target;
}

::size_t xclReadAddrKernelCtrl_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrKernelCtrl_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  if (cached_has_bits & 0x00000006u) {
    // required uint64 addr = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  if (cached_has_bits & 0x00000018u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadAddrKernelCtrl_call::_class_data_ = {
    xclReadAddrKernelCtrl_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadAddrKernelCtrl_call::GetClassData() const {
  return &_class_data_;
}

void xclReadAddrKernelCtrl_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrKernelCtrl_call*>(&to_msg);
  auto& from = static_cast<const xclReadAddrKernelCtrl_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrKernelCtrl_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrKernelCtrl_call::CopyFrom(const xclReadAddrKernelCtrl_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrKernelCtrl_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadAddrKernelCtrl_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclReadAddrKernelCtrl_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadAddrKernelCtrl_call::InternalSwap(xclReadAddrKernelCtrl_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.bar_id_)
      + sizeof(xclReadAddrKernelCtrl_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclReadAddrKernelCtrl_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[27]);
}
// ===================================================================

class xclReadAddrKernelCtrl_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadAddrKernelCtrl_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadAddrKernelCtrl_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrKernelCtrl_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclReadAddrKernelCtrl_response::xclReadAddrKernelCtrl_response(
    ::google::protobuf::Arena* arena,
    const xclReadAddrKernelCtrl_response& from)
    : ::google::protobuf::Message(arena) {
  xclReadAddrKernelCtrl_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.valid_ = from._impl_.valid_;

  // @@protoc_insertion_point(copy_constructor:xclReadAddrKernelCtrl_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadAddrKernelCtrl_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclReadAddrKernelCtrl_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclReadAddrKernelCtrl_response::~xclReadAddrKernelCtrl_response() {
  // @@protoc_insertion_point(destructor:xclReadAddrKernelCtrl_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadAddrKernelCtrl_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadAddrKernelCtrl_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadAddrKernelCtrl_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadAddrKernelCtrl_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclReadAddrKernelCtrl_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadAddrKernelCtrl_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bytes data = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_.data_)}},
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclReadAddrKernelCtrl_response, _impl_.valid_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_.valid_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bytes data = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadAddrKernelCtrl_response, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadAddrKernelCtrl_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadAddrKernelCtrl_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadAddrKernelCtrl_response)
  return target;
}

::size_t xclReadAddrKernelCtrl_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadAddrKernelCtrl_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadAddrKernelCtrl_response::_class_data_ = {
    xclReadAddrKernelCtrl_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadAddrKernelCtrl_response::GetClassData() const {
  return &_class_data_;
}

void xclReadAddrKernelCtrl_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadAddrKernelCtrl_response*>(&to_msg);
  auto& from = static_cast<const xclReadAddrKernelCtrl_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadAddrKernelCtrl_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadAddrKernelCtrl_response::CopyFrom(const xclReadAddrKernelCtrl_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadAddrKernelCtrl_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadAddrKernelCtrl_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclReadAddrKernelCtrl_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadAddrKernelCtrl_response::InternalSwap(xclReadAddrKernelCtrl_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclReadAddrKernelCtrl_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[28]);
}
// ===================================================================

class xclRegRead_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegRead_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_._has_bits_);
  static void set_has_baseaddress(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_pf_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_bar_id(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000e) ^ 0x0000000e) != 0;
  }
};

xclRegRead_call::xclRegRead_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclRegRead_call)
}
inline PROTOBUF_NDEBUG_INLINE xclRegRead_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclRegRead_call::xclRegRead_call(
    ::google::protobuf::Arena* arena,
    const xclRegRead_call& from)
    : ::google::protobuf::Message(arena) {
  xclRegRead_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, baseaddress_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, baseaddress_),
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, baseaddress_) +
               sizeof(Impl_::bar_id_));

  // @@protoc_insertion_point(copy_constructor:xclRegRead_call)
}
inline PROTOBUF_NDEBUG_INLINE xclRegRead_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclRegRead_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, baseaddress_),
           0,
           offsetof(Impl_, bar_id_) -
               offsetof(Impl_, baseaddress_) +
               sizeof(Impl_::bar_id_));
}
xclRegRead_call::~xclRegRead_call() {
  // @@protoc_insertion_point(destructor:xclRegRead_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclRegRead_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclRegRead_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegRead_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.baseaddress_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.bar_id_) -
        reinterpret_cast<char*>(&_impl_.baseaddress_)) + sizeof(_impl_.bar_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclRegRead_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclRegRead_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967232,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclRegRead_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // required uint32 baseaddress = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegRead_call, _impl_.baseaddress_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.baseaddress_)}},
    // required uint32 offset = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegRead_call, _impl_.offset_), 2>(),
     {16, 2, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.offset_)}},
    // required uint32 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegRead_call, _impl_.size_), 3>(),
     {24, 3, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.size_)}},
    // optional bytes data = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.data_)}},
    // optional uint32 PF_ID = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegRead_call, _impl_.pf_id_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.pf_id_)}},
    // optional uint32 BAR_ID = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclRegRead_call, _impl_.bar_id_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.bar_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 baseaddress = 1;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.baseaddress_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint32 offset = 2;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.offset_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint32 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.size_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional bytes data = 4;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint32 PF_ID = 5;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.pf_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 BAR_ID = 6;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.bar_id_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclRegRead_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegRead_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 baseaddress = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_baseaddress(), target);
  }

  // required uint32 offset = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_offset(), target);
  }

  // required uint32 size = 3;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_size(), target);
  }

  // optional bytes data = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // optional uint32 PF_ID = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_pf_id(), target);
  }

  // optional uint32 BAR_ID = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_bar_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegRead_call)
  return target;
}

::size_t xclRegRead_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegRead_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 4;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  if (cached_has_bits & 0x0000000eu) {
    // required uint32 baseaddress = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_baseaddress());
    }

    // required uint32 offset = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_offset());
    }

    // required uint32 size = 3;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  if (cached_has_bits & 0x00000030u) {
    // optional uint32 PF_ID = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_pf_id());
    }

    // optional uint32 BAR_ID = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_bar_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclRegRead_call::_class_data_ = {
    xclRegRead_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclRegRead_call::GetClassData() const {
  return &_class_data_;
}

void xclRegRead_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclRegRead_call*>(&to_msg);
  auto& from = static_cast<const xclRegRead_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegRead_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.baseaddress_ = from._impl_.baseaddress_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.pf_id_ = from._impl_.pf_id_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.bar_id_ = from._impl_.bar_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegRead_call::CopyFrom(const xclRegRead_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegRead_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclRegRead_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclRegRead_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclRegRead_call::InternalSwap(xclRegRead_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.bar_id_)
      + sizeof(xclRegRead_call::_impl_.bar_id_)
      - PROTOBUF_FIELD_OFFSET(xclRegRead_call, _impl_.baseaddress_)>(
          reinterpret_cast<char*>(&_impl_.baseaddress_),
          reinterpret_cast<char*>(&other->_impl_.baseaddress_));
}

::google::protobuf::Metadata xclRegRead_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[29]);
}
// ===================================================================

class xclRegRead_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclRegRead_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000002) ^ 0x00000002) != 0;
  }
};

xclRegRead_response::xclRegRead_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclRegRead_response)
}
inline PROTOBUF_NDEBUG_INLINE xclRegRead_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclRegRead_response::xclRegRead_response(
    ::google::protobuf::Arena* arena,
    const xclRegRead_response& from)
    : ::google::protobuf::Message(arena) {
  xclRegRead_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.valid_ = from._impl_.valid_;

  // @@protoc_insertion_point(copy_constructor:xclRegRead_response)
}
inline PROTOBUF_NDEBUG_INLINE xclRegRead_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclRegRead_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclRegRead_response::~xclRegRead_response() {
  // @@protoc_insertion_point(destructor:xclRegRead_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclRegRead_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclRegRead_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclRegRead_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclRegRead_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclRegRead_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclRegRead_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bytes data = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_.data_)}},
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclRegRead_response, _impl_.valid_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_.valid_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bytes data = 2;
    {PROTOBUF_FIELD_OFFSET(xclRegRead_response, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclRegRead_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclRegRead_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  // optional bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclRegRead_response)
  return target;
}

::size_t xclRegRead_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclRegRead_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bytes data = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclRegRead_response::_class_data_ = {
    xclRegRead_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclRegRead_response::GetClassData() const {
  return &_class_data_;
}

void xclRegRead_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclRegRead_response*>(&to_msg);
  auto& from = static_cast<const xclRegRead_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclRegRead_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclRegRead_response::CopyFrom(const xclRegRead_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclRegRead_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclRegRead_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclRegRead_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclRegRead_response::InternalSwap(xclRegRead_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclRegRead_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[30]);
}
// ===================================================================

class xclPerfMonReadCounters_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_._has_bits_);
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        slotname_(arena, from.slotname_) {}

xclPerfMonReadCounters_call::xclPerfMonReadCounters_call(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadCounters_call& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadCounters_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.accel_ = from._impl_.accel_;

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        slotname_(arena) {}

inline void xclPerfMonReadCounters_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.accel_ = {};
}
xclPerfMonReadCounters_call::~xclPerfMonReadCounters_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadCounters_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.slotname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadCounters_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_.accel_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 44, 2> xclPerfMonReadCounters_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadCounters_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool accel = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonReadCounters_call, _impl_.accel_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_.accel_)}},
    // required string slotname = 1;
    {::_pbi::TcParser::FastSS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_.slotname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required string slotname = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_.slotname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional bool accel = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_call, _impl_.accel_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
    "\33\10\0\0\0\0\0\0"
    "xclPerfMonReadCounters_call"
    "slotname"
  }},
};

::uint8_t* xclPerfMonReadCounters_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string slotname = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_slotname();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclPerfMonReadCounters_call.slotname");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional bool accel = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        2, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_call)
  return target;
}

::size_t xclPerfMonReadCounters_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required string slotname = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                    this->_internal_slotname());
  }

  // optional bool accel = 2;
  if (cached_has_bits & 0x00000002u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadCounters_call::_class_data_ = {
    xclPerfMonReadCounters_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadCounters_call::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadCounters_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_call::CopyFrom(const xclPerfMonReadCounters_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadCounters_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclPerfMonReadCounters_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadCounters_call::InternalSwap(xclPerfMonReadCounters_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.slotname_, &other->_impl_.slotname_, arena);
        swap(_impl_.accel_, other->_impl_.accel_);
}

::google::protobuf::Metadata xclPerfMonReadCounters_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[31]);
}
// ===================================================================

class xclPerfMonReadCounters_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_._has_bits_);
  static void set_has_wr_byte_count(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_wr_trans_count(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_total_wr_latency(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_rd_byte_count(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_rd_trans_count(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_total_rd_latency(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_response)
}
xclPerfMonReadCounters_response::xclPerfMonReadCounters_response(
    ::google::protobuf::Arena* arena, const xclPerfMonReadCounters_response& from)
    : xclPerfMonReadCounters_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPerfMonReadCounters_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, wr_byte_count_),
           0,
           offsetof(Impl_, total_rd_latency_) -
               offsetof(Impl_, wr_byte_count_) +
               sizeof(Impl_::total_rd_latency_));
}
xclPerfMonReadCounters_response::~xclPerfMonReadCounters_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadCounters_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadCounters_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    ::memset(&_impl_.wr_byte_count_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.total_rd_latency_) -
        reinterpret_cast<char*>(&_impl_.wr_byte_count_)) + sizeof(_impl_.total_rd_latency_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclPerfMonReadCounters_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967232,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadCounters_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint32 wr_byte_count = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.wr_byte_count_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_byte_count_)}},
    // optional uint32 wr_trans_count = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.wr_trans_count_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_trans_count_)}},
    // optional uint32 total_wr_latency = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.total_wr_latency_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_wr_latency_)}},
    // optional uint32 rd_byte_count = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.rd_byte_count_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.rd_byte_count_)}},
    // optional uint32 rd_trans_count = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.rd_trans_count_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.rd_trans_count_)}},
    // optional uint32 total_rd_latency = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadCounters_response, _impl_.total_rd_latency_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_rd_latency_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint32 wr_byte_count = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_byte_count_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 wr_trans_count = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_trans_count_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 total_wr_latency = 3;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_wr_latency_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 rd_byte_count = 4;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.rd_byte_count_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 rd_trans_count = 5;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.rd_trans_count_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 total_rd_latency = 6;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_rd_latency_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPerfMonReadCounters_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 wr_byte_count = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_wr_byte_count(), target);
  }

  // optional uint32 wr_trans_count = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_wr_trans_count(), target);
  }

  // optional uint32 total_wr_latency = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_total_wr_latency(), target);
  }

  // optional uint32 rd_byte_count = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this->_internal_rd_byte_count(), target);
  }

  // optional uint32 rd_trans_count = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        5, this->_internal_rd_trans_count(), target);
  }

  // optional uint32 total_rd_latency = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_total_rd_latency(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_response)
  return target;
}

::size_t xclPerfMonReadCounters_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional uint32 wr_byte_count = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_wr_byte_count());
    }

    // optional uint32 wr_trans_count = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_wr_trans_count());
    }

    // optional uint32 total_wr_latency = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_total_wr_latency());
    }

    // optional uint32 rd_byte_count = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_rd_byte_count());
    }

    // optional uint32 rd_trans_count = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_rd_trans_count());
    }

    // optional uint32 total_rd_latency = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_total_rd_latency());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadCounters_response::_class_data_ = {
    xclPerfMonReadCounters_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadCounters_response::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadCounters_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.wr_byte_count_ = from._impl_.wr_byte_count_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.wr_trans_count_ = from._impl_.wr_trans_count_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.total_wr_latency_ = from._impl_.total_wr_latency_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.rd_byte_count_ = from._impl_.rd_byte_count_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.rd_trans_count_ = from._impl_.rd_trans_count_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.total_rd_latency_ = from._impl_.total_rd_latency_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_response::CopyFrom(const xclPerfMonReadCounters_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadCounters_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadCounters_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadCounters_response::InternalSwap(xclPerfMonReadCounters_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.total_rd_latency_)
      + sizeof(xclPerfMonReadCounters_response::_impl_.total_rd_latency_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_response, _impl_.wr_byte_count_)>(
          reinterpret_cast<char*>(&_impl_.wr_byte_count_),
          reinterpret_cast<char*>(&other->_impl_.wr_byte_count_));
}

::google::protobuf::Metadata xclPerfMonReadCounters_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[32]);
}
// ===================================================================

class xclPerfMonGetTraceCount_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonGetTraceCount_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonGetTraceCount_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonGetTraceCount_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        slotname_(arena, from.slotname_) {}

xclPerfMonGetTraceCount_call::xclPerfMonGetTraceCount_call(
    ::google::protobuf::Arena* arena,
    const xclPerfMonGetTraceCount_call& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonGetTraceCount_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ack_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, ack_),
           offsetof(Impl_, accel_) -
               offsetof(Impl_, ack_) +
               sizeof(Impl_::accel_));

  // @@protoc_insertion_point(copy_constructor:xclPerfMonGetTraceCount_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonGetTraceCount_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        slotname_(arena) {}

inline void xclPerfMonGetTraceCount_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ack_),
           0,
           offsetof(Impl_, accel_) -
               offsetof(Impl_, ack_) +
               sizeof(Impl_::accel_));
}
xclPerfMonGetTraceCount_call::~xclPerfMonGetTraceCount_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonGetTraceCount_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonGetTraceCount_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.slotname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonGetTraceCount_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonGetTraceCount_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  ::memset(&_impl_.ack_, 0, static_cast<::size_t>(
      reinterpret_cast<char*>(&_impl_.accel_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonGetTraceCount_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 45, 2> xclPerfMonGetTraceCount_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonGetTraceCount_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonGetTraceCount_call, _impl_.ack_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.ack_)}},
    // optional string slotname = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.slotname_)}},
    // optional bool accel = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonGetTraceCount_call, _impl_.accel_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.accel_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.ack_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional string slotname = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.slotname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional bool accel = 3;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.accel_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
    "\34\0\10\0\0\0\0\0"
    "xclPerfMonGetTraceCount_call"
    "slotname"
  }},
};

::uint8_t* xclPerfMonGetTraceCount_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonGetTraceCount_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_slotname();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclPerfMonGetTraceCount_call.slotname");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional bool accel = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        3, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonGetTraceCount_call)
  return target;
}

::size_t xclPerfMonGetTraceCount_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonGetTraceCount_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

    // optional bool accel = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonGetTraceCount_call::_class_data_ = {
    xclPerfMonGetTraceCount_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonGetTraceCount_call::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonGetTraceCount_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonGetTraceCount_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonGetTraceCount_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonGetTraceCount_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonGetTraceCount_call::CopyFrom(const xclPerfMonGetTraceCount_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonGetTraceCount_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonGetTraceCount_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonGetTraceCount_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonGetTraceCount_call::InternalSwap(xclPerfMonGetTraceCount_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.slotname_, &other->_impl_.slotname_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.accel_)
      + sizeof(xclPerfMonGetTraceCount_call::_impl_.accel_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::google::protobuf::Metadata xclPerfMonGetTraceCount_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[33]);
}
// ===================================================================

class xclPerfMonGetTraceCount_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonGetTraceCount_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_response, _impl_._has_bits_);
  static void set_has_no_of_samples(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonGetTraceCount_response)
}
xclPerfMonGetTraceCount_response::xclPerfMonGetTraceCount_response(
    ::google::protobuf::Arena* arena, const xclPerfMonGetTraceCount_response& from)
    : xclPerfMonGetTraceCount_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonGetTraceCount_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPerfMonGetTraceCount_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.no_of_samples_ = {};
}
xclPerfMonGetTraceCount_response::~xclPerfMonGetTraceCount_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonGetTraceCount_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonGetTraceCount_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonGetTraceCount_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonGetTraceCount_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.no_of_samples_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonGetTraceCount_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclPerfMonGetTraceCount_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonGetTraceCount_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint32 no_of_samples = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonGetTraceCount_response, _impl_.no_of_samples_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_response, _impl_.no_of_samples_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint32 no_of_samples = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonGetTraceCount_response, _impl_.no_of_samples_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPerfMonGetTraceCount_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonGetTraceCount_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 no_of_samples = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_no_of_samples(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonGetTraceCount_response)
  return target;
}

::size_t xclPerfMonGetTraceCount_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonGetTraceCount_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint32 no_of_samples = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_no_of_samples());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonGetTraceCount_response::_class_data_ = {
    xclPerfMonGetTraceCount_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonGetTraceCount_response::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonGetTraceCount_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonGetTraceCount_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonGetTraceCount_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonGetTraceCount_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_no_of_samples(from._internal_no_of_samples());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonGetTraceCount_response::CopyFrom(const xclPerfMonGetTraceCount_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonGetTraceCount_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonGetTraceCount_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonGetTraceCount_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonGetTraceCount_response::InternalSwap(xclPerfMonGetTraceCount_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.no_of_samples_, other->_impl_.no_of_samples_);
}

::google::protobuf::Metadata xclPerfMonGetTraceCount_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[34]);
}
// ===================================================================

class xclPerfMonReadTrace_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_accel(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        slotname_(arena, from.slotname_) {}

xclPerfMonReadTrace_call::xclPerfMonReadTrace_call(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadTrace_call& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadTrace_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ack_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, ack_),
           offsetof(Impl_, accel_) -
               offsetof(Impl_, ack_) +
               sizeof(Impl_::accel_));

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        slotname_(arena) {}

inline void xclPerfMonReadTrace_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ack_),
           0,
           offsetof(Impl_, accel_) -
               offsetof(Impl_, ack_) +
               sizeof(Impl_::accel_));
}
xclPerfMonReadTrace_call::~xclPerfMonReadTrace_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.slotname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  ::memset(&_impl_.ack_, 0, static_cast<::size_t>(
      reinterpret_cast<char*>(&_impl_.accel_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.accel_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 41, 2> xclPerfMonReadTrace_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadTrace_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonReadTrace_call, _impl_.ack_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.ack_)}},
    // optional string slotname = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.slotname_)}},
    // optional bool accel = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonReadTrace_call, _impl_.accel_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.accel_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.ack_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional string slotname = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.slotname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional bool accel = 3;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.accel_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
    "\30\0\10\0\0\0\0\0"
    "xclPerfMonReadTrace_call"
    "slotname"
  }},
};

::uint8_t* xclPerfMonReadTrace_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_slotname();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclPerfMonReadTrace_call.slotname");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional bool accel = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        3, this->_internal_accel(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_call)
  return target;
}

::size_t xclPerfMonReadTrace_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

    // optional bool accel = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_call::_class_data_ = {
    xclPerfMonReadTrace_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_call::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.accel_ = from._impl_.accel_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_call::CopyFrom(const xclPerfMonReadTrace_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_call::InternalSwap(xclPerfMonReadTrace_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.slotname_, &other->_impl_.slotname_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.accel_)
      + sizeof(xclPerfMonReadTrace_call::_impl_.accel_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::google::protobuf::Metadata xclPerfMonReadTrace_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[35]);
}
// ===================================================================

class xclPerfMonReadTrace_response_events::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_response_events>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_._has_bits_);
  static void set_has_awlen(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_arlen(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_eventflags(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_host_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_rd_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_wr_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
};

xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_response.events)
}
xclPerfMonReadTrace_response_events::xclPerfMonReadTrace_response_events(
    ::google::protobuf::Arena* arena, const xclPerfMonReadTrace_response_events& from)
    : xclPerfMonReadTrace_response_events(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_response_events::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPerfMonReadTrace_response_events::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, awlen_),
           0,
           offsetof(Impl_, wr_bytes_) -
               offsetof(Impl_, awlen_) +
               sizeof(Impl_::wr_bytes_));
}
xclPerfMonReadTrace_response_events::~xclPerfMonReadTrace_response_events() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_response.events)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_response_events::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_response_events::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_response.events)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    ::memset(&_impl_.awlen_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.wr_bytes_) -
        reinterpret_cast<char*>(&_impl_.awlen_)) + sizeof(_impl_.wr_bytes_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_response_events::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 7, 0, 0, 2> xclPerfMonReadTrace_response_events::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_._has_bits_),
    0, // no _extensions_
    7, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967168,  // skipmap
    offsetof(decltype(_table_), field_entries),
    7,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadTrace_response_events_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint32 awlen = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.awlen_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.awlen_)}},
    // optional uint32 arlen = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.arlen_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.arlen_)}},
    // optional uint32 eventflags = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.eventflags_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.eventflags_)}},
    // optional uint32 timestamp = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.timestamp_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.timestamp_)}},
    // optional uint64 host_timestamp = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.host_timestamp_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.host_timestamp_)}},
    // optional uint32 rd_bytes = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.rd_bytes_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.rd_bytes_)}},
    // optional uint32 wr_bytes = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_response_events, _impl_.wr_bytes_), 6>(),
     {56, 6, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.wr_bytes_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint32 awlen = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.awlen_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 arlen = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.arlen_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 eventflags = 3;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.eventflags_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 timestamp = 4;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.timestamp_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint64 host_timestamp = 5;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.host_timestamp_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint32 rd_bytes = 6;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.rd_bytes_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 wr_bytes = 7;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.wr_bytes_), _Internal::kHasBitsOffset + 6, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPerfMonReadTrace_response_events::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_response.events)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 awlen = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_awlen(), target);
  }

  // optional uint32 arlen = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_arlen(), target);
  }

  // optional uint32 eventflags = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_eventflags(), target);
  }

  // optional uint32 timestamp = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this->_internal_timestamp(), target);
  }

  // optional uint64 host_timestamp = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_host_timestamp(), target);
  }

  // optional uint32 rd_bytes = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_rd_bytes(), target);
  }

  // optional uint32 wr_bytes = 7;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        7, this->_internal_wr_bytes(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_response.events)
  return target;
}

::size_t xclPerfMonReadTrace_response_events::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_response.events)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    // optional uint32 awlen = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_awlen());
    }

    // optional uint32 arlen = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_arlen());
    }

    // optional uint32 eventflags = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_eventflags());
    }

    // optional uint32 timestamp = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_timestamp());
    }

    // optional uint64 host_timestamp = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_host_timestamp());
    }

    // optional uint32 rd_bytes = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_rd_bytes());
    }

    // optional uint32 wr_bytes = 7;
    if (cached_has_bits & 0x00000040u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_wr_bytes());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_response_events::_class_data_ = {
    xclPerfMonReadTrace_response_events::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_response_events::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_response_events::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_response_events*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_response_events&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_response.events)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.awlen_ = from._impl_.awlen_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.arlen_ = from._impl_.arlen_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.eventflags_ = from._impl_.eventflags_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.timestamp_ = from._impl_.timestamp_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.host_timestamp_ = from._impl_.host_timestamp_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.rd_bytes_ = from._impl_.rd_bytes_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.wr_bytes_ = from._impl_.wr_bytes_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_response_events::CopyFrom(const xclPerfMonReadTrace_response_events& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_response.events)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_response_events::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_response_events::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_response_events::InternalSwap(xclPerfMonReadTrace_response_events* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.wr_bytes_)
      + sizeof(xclPerfMonReadTrace_response_events::_impl_.wr_bytes_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response_events, _impl_.awlen_)>(
          reinterpret_cast<char*>(&_impl_.awlen_),
          reinterpret_cast<char*>(&other->_impl_.awlen_));
}

::google::protobuf::Metadata xclPerfMonReadTrace_response_events::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[36]);
}
// ===================================================================

class xclPerfMonReadTrace_response::_Internal {
 public:
};

xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : output_data_{visibility, arena, from.output_data_},
        _cached_size_{0} {}

xclPerfMonReadTrace_response::xclPerfMonReadTrace_response(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadTrace_response& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadTrace_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : output_data_{visibility, arena},
        _cached_size_{0} {}

inline void xclPerfMonReadTrace_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclPerfMonReadTrace_response::~xclPerfMonReadTrace_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.output_data_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2> xclPerfMonReadTrace_response::_table_ = {
  {
    0,  // no _has_bits_
    0, // no _extensions_
    8, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967167,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclPerfMonReadTrace_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // repeated .xclPerfMonReadTrace_response.events output_data = 8;
    {::_pbi::TcParser::FastMtR1,
     {66, 63, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response, _impl_.output_data_)}},
  }}, {{
    65535, 65535
  }}, {{
    // repeated .xclPerfMonReadTrace_response.events output_data = 8;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_response, _impl_.output_data_), 0, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclPerfMonReadTrace_response_events>()},
  }}, {{
  }},
};

::uint8_t* xclPerfMonReadTrace_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  // repeated .xclPerfMonReadTrace_response.events output_data = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_output_data_size()); i < n; i++) {
    const auto& repfield = this->_internal_output_data().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_response)
  return target;
}

::size_t xclPerfMonReadTrace_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_response.events output_data = 8;
  total_size += 1UL * this->_internal_output_data_size();
  for (const auto& msg : this->_internal_output_data()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_response::_class_data_ = {
    xclPerfMonReadTrace_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_response::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_output_data()->MergeFrom(
      from._internal_output_data());
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_response::CopyFrom(const xclPerfMonReadTrace_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_response::InternalSwap(xclPerfMonReadTrace_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.output_data_.InternalSwap(&other->_impl_.output_data_);
}

::google::protobuf::Metadata xclPerfMonReadTrace_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[37]);
}
// ===================================================================

class ddr_mem_msg::_Internal {
 public:
  using HasBits = decltype(std::declval<ddr_mem_msg>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_._has_bits_);
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_strb_data(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_fullsize(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

ddr_mem_msg::ddr_mem_msg(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:ddr_mem_msg)
}
inline PROTOBUF_NDEBUG_INLINE ddr_mem_msg::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_),
        strb_data_(arena, from.strb_data_) {}

ddr_mem_msg::ddr_mem_msg(
    ::google::protobuf::Arena* arena,
    const ddr_mem_msg& from)
    : ::google::protobuf::Message(arena) {
  ddr_mem_msg* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, size_),
           offsetof(Impl_, fullsize_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::fullsize_));

  // @@protoc_insertion_point(copy_constructor:ddr_mem_msg)
}
inline PROTOBUF_NDEBUG_INLINE ddr_mem_msg::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena),
        strb_data_(arena) {}

inline void ddr_mem_msg::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, fullsize_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::fullsize_));
}
ddr_mem_msg::~ddr_mem_msg() {
  // @@protoc_insertion_point(destructor:ddr_mem_msg)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void ddr_mem_msg::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.strb_data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void ddr_mem_msg::Clear() {
// @@protoc_insertion_point(message_clear_start:ddr_mem_msg)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.data_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.strb_data_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.fullsize_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.fullsize_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* ddr_mem_msg::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 0, 0, 2> ddr_mem_msg::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_._has_bits_),
    0, // no _extensions_
    4, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967280,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_ddr_mem_msg_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 fullsize = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ddr_mem_msg, _impl_.fullsize_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.fullsize_)}},
    // required bytes data = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.data_)}},
    // optional uint64 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(ddr_mem_msg, _impl_.size_), 2>(),
     {16, 2, 0, PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.size_)}},
    // optional bytes strb_data = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 1, 0, PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.strb_data_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes data = 1;
    {PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 size = 2;
    {PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bytes strb_data = 3;
    {PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.strb_data_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 fullsize = 4;
    {PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.fullsize_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* ddr_mem_msg::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:ddr_mem_msg)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes data = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_size(), target);
  }

  // optional bytes strb_data = 3;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_strb_data();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // optional uint64 fullsize = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_fullsize(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:ddr_mem_msg)
  return target;
}

::size_t ddr_mem_msg::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:ddr_mem_msg)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bytes data = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_data());
  }

  if (cached_has_bits & 0x0000000eu) {
    // optional bytes strb_data = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_strb_data());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // optional uint64 fullsize = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_fullsize());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData ddr_mem_msg::_class_data_ = {
    ddr_mem_msg::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* ddr_mem_msg::GetClassData() const {
  return &_class_data_;
}

void ddr_mem_msg::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<ddr_mem_msg*>(&to_msg);
  auto& from = static_cast<const ddr_mem_msg&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:ddr_mem_msg)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_strb_data(from._internal_strb_data());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.fullsize_ = from._impl_.fullsize_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void ddr_mem_msg::CopyFrom(const ddr_mem_msg& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:ddr_mem_msg)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool ddr_mem_msg::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* ddr_mem_msg::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void ddr_mem_msg::InternalSwap(ddr_mem_msg* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.strb_data_, &other->_impl_.strb_data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.fullsize_)
      + sizeof(ddr_mem_msg::_impl_.fullsize_)
      - PROTOBUF_FIELD_OFFSET(ddr_mem_msg, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata ddr_mem_msg::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[38]);
}
// ===================================================================

class xclGetDeviceTimestamp_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDeviceTimestamp_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_call, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGetDeviceTimestamp_call)
}
xclGetDeviceTimestamp_call::xclGetDeviceTimestamp_call(
    ::google::protobuf::Arena* arena, const xclGetDeviceTimestamp_call& from)
    : xclGetDeviceTimestamp_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGetDeviceTimestamp_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGetDeviceTimestamp_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGetDeviceTimestamp_call::~xclGetDeviceTimestamp_call() {
  // @@protoc_insertion_point(destructor:xclGetDeviceTimestamp_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGetDeviceTimestamp_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGetDeviceTimestamp_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDeviceTimestamp_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGetDeviceTimestamp_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGetDeviceTimestamp_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGetDeviceTimestamp_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGetDeviceTimestamp_call, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_call, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_call, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGetDeviceTimestamp_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDeviceTimestamp_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDeviceTimestamp_call)
  return target;
}

::size_t xclGetDeviceTimestamp_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDeviceTimestamp_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGetDeviceTimestamp_call::_class_data_ = {
    xclGetDeviceTimestamp_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGetDeviceTimestamp_call::GetClassData() const {
  return &_class_data_;
}

void xclGetDeviceTimestamp_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGetDeviceTimestamp_call*>(&to_msg);
  auto& from = static_cast<const xclGetDeviceTimestamp_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDeviceTimestamp_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDeviceTimestamp_call::CopyFrom(const xclGetDeviceTimestamp_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDeviceTimestamp_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGetDeviceTimestamp_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclGetDeviceTimestamp_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGetDeviceTimestamp_call::InternalSwap(xclGetDeviceTimestamp_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGetDeviceTimestamp_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[39]);
}
// ===================================================================

class xclGetDeviceTimestamp_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDeviceTimestamp_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_response, _impl_._has_bits_);
  static void set_has_device_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGetDeviceTimestamp_response)
}
xclGetDeviceTimestamp_response::xclGetDeviceTimestamp_response(
    ::google::protobuf::Arena* arena, const xclGetDeviceTimestamp_response& from)
    : xclGetDeviceTimestamp_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGetDeviceTimestamp_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGetDeviceTimestamp_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.device_timestamp_ = {};
}
xclGetDeviceTimestamp_response::~xclGetDeviceTimestamp_response() {
  // @@protoc_insertion_point(destructor:xclGetDeviceTimestamp_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGetDeviceTimestamp_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGetDeviceTimestamp_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDeviceTimestamp_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.device_timestamp_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGetDeviceTimestamp_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGetDeviceTimestamp_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGetDeviceTimestamp_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint64 device_timestamp = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGetDeviceTimestamp_response, _impl_.device_timestamp_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_response, _impl_.device_timestamp_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 device_timestamp = 1;
    {PROTOBUF_FIELD_OFFSET(xclGetDeviceTimestamp_response, _impl_.device_timestamp_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGetDeviceTimestamp_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDeviceTimestamp_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 device_timestamp = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_device_timestamp(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDeviceTimestamp_response)
  return target;
}

::size_t xclGetDeviceTimestamp_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDeviceTimestamp_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint64 device_timestamp = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_device_timestamp());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGetDeviceTimestamp_response::_class_data_ = {
    xclGetDeviceTimestamp_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGetDeviceTimestamp_response::GetClassData() const {
  return &_class_data_;
}

void xclGetDeviceTimestamp_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGetDeviceTimestamp_response*>(&to_msg);
  auto& from = static_cast<const xclGetDeviceTimestamp_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDeviceTimestamp_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_device_timestamp(from._internal_device_timestamp());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDeviceTimestamp_response::CopyFrom(const xclGetDeviceTimestamp_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDeviceTimestamp_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGetDeviceTimestamp_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGetDeviceTimestamp_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGetDeviceTimestamp_response::InternalSwap(xclGetDeviceTimestamp_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.device_timestamp_, other->_impl_.device_timestamp_);
}

::google::protobuf::Metadata xclGetDeviceTimestamp_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[40]);
}
// ===================================================================

class xclReadBusStatus_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadBusStatus_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadBusStatus_call, _impl_._has_bits_);
  static void set_has_slot_n(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclReadBusStatus_call::xclReadBusStatus_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadBusStatus_call)
}
xclReadBusStatus_call::xclReadBusStatus_call(
    ::google::protobuf::Arena* arena, const xclReadBusStatus_call& from)
    : xclReadBusStatus_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclReadBusStatus_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclReadBusStatus_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.slot_n_ = {};
}
xclReadBusStatus_call::~xclReadBusStatus_call() {
  // @@protoc_insertion_point(destructor:xclReadBusStatus_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadBusStatus_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadBusStatus_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadBusStatus_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.slot_n_ = 0;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadBusStatus_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclReadBusStatus_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadBusStatus_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadBusStatus_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional int32 slot_n = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclReadBusStatus_call, _impl_.slot_n_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_call, _impl_.slot_n_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional int32 slot_n = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_call, _impl_.slot_n_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadBusStatus_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadBusStatus_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional int32 slot_n = 1;
  if (cached_has_bits & 0x00000001u) {
    target = ::google::protobuf::internal::WireFormatLite::
        WriteInt32ToArrayWithField<1>(
            stream, this->_internal_slot_n(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadBusStatus_call)
  return target;
}

::size_t xclReadBusStatus_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadBusStatus_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional int32 slot_n = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::Int32SizePlusOne(
        this->_internal_slot_n());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadBusStatus_call::_class_data_ = {
    xclReadBusStatus_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadBusStatus_call::GetClassData() const {
  return &_class_data_;
}

void xclReadBusStatus_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadBusStatus_call*>(&to_msg);
  auto& from = static_cast<const xclReadBusStatus_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadBusStatus_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_slot_n(from._internal_slot_n());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadBusStatus_call::CopyFrom(const xclReadBusStatus_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadBusStatus_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadBusStatus_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclReadBusStatus_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadBusStatus_call::InternalSwap(xclReadBusStatus_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.slot_n_, other->_impl_.slot_n_);
}

::google::protobuf::Metadata xclReadBusStatus_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[41]);
}
// ===================================================================

class xclReadBusStatus_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadBusStatus_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_._has_bits_);
  static void set_has_uncomplete_rd_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_uncomplete_wr_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_idle_bus_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_uncomplete_rd_id(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_uncomplete_wr_id(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclReadBusStatus_response::xclReadBusStatus_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadBusStatus_response)
}
xclReadBusStatus_response::xclReadBusStatus_response(
    ::google::protobuf::Arena* arena, const xclReadBusStatus_response& from)
    : xclReadBusStatus_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclReadBusStatus_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclReadBusStatus_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, uncomplete_rd_cycles_),
           0,
           offsetof(Impl_, uncomplete_wr_id_) -
               offsetof(Impl_, uncomplete_rd_cycles_) +
               sizeof(Impl_::uncomplete_wr_id_));
}
xclReadBusStatus_response::~xclReadBusStatus_response() {
  // @@protoc_insertion_point(destructor:xclReadBusStatus_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadBusStatus_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadBusStatus_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadBusStatus_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    ::memset(&_impl_.uncomplete_rd_cycles_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.uncomplete_wr_id_) -
        reinterpret_cast<char*>(&_impl_.uncomplete_rd_cycles_)) + sizeof(_impl_.uncomplete_wr_id_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadBusStatus_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclReadBusStatus_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_._has_bits_),
    0, // no _extensions_
    5, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967264,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadBusStatus_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 uncomplete_rd_cycles = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_)}},
    // optional uint64 uncomplete_wr_cycles = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadBusStatus_response, _impl_.uncomplete_wr_cycles_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_cycles_)}},
    // optional uint64 idle_bus_cycles = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadBusStatus_response, _impl_.idle_bus_cycles_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.idle_bus_cycles_)}},
    // optional uint64 uncomplete_rd_id = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadBusStatus_response, _impl_.uncomplete_rd_id_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_id_)}},
    // optional uint64 uncomplete_wr_id = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadBusStatus_response, _impl_.uncomplete_wr_id_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_id_)}},
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 uncomplete_rd_cycles = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 uncomplete_wr_cycles = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_cycles_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 idle_bus_cycles = 3;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.idle_bus_cycles_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 uncomplete_rd_id = 4;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_id_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 uncomplete_wr_id = 5;
    {PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_id_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadBusStatus_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadBusStatus_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 uncomplete_rd_cycles = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_uncomplete_rd_cycles(), target);
  }

  // optional uint64 uncomplete_wr_cycles = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_uncomplete_wr_cycles(), target);
  }

  // optional uint64 idle_bus_cycles = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_idle_bus_cycles(), target);
  }

  // optional uint64 uncomplete_rd_id = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_uncomplete_rd_id(), target);
  }

  // optional uint64 uncomplete_wr_id = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_uncomplete_wr_id(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadBusStatus_response)
  return target;
}

::size_t xclReadBusStatus_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadBusStatus_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional uint64 uncomplete_rd_cycles = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_uncomplete_rd_cycles());
    }

    // optional uint64 uncomplete_wr_cycles = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_uncomplete_wr_cycles());
    }

    // optional uint64 idle_bus_cycles = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_idle_bus_cycles());
    }

    // optional uint64 uncomplete_rd_id = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_uncomplete_rd_id());
    }

    // optional uint64 uncomplete_wr_id = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_uncomplete_wr_id());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadBusStatus_response::_class_data_ = {
    xclReadBusStatus_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadBusStatus_response::GetClassData() const {
  return &_class_data_;
}

void xclReadBusStatus_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadBusStatus_response*>(&to_msg);
  auto& from = static_cast<const xclReadBusStatus_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadBusStatus_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.uncomplete_rd_cycles_ = from._impl_.uncomplete_rd_cycles_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.uncomplete_wr_cycles_ = from._impl_.uncomplete_wr_cycles_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.idle_bus_cycles_ = from._impl_.idle_bus_cycles_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.uncomplete_rd_id_ = from._impl_.uncomplete_rd_id_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.uncomplete_wr_id_ = from._impl_.uncomplete_wr_id_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadBusStatus_response::CopyFrom(const xclReadBusStatus_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadBusStatus_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadBusStatus_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclReadBusStatus_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadBusStatus_response::InternalSwap(xclReadBusStatus_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_wr_id_)
      + sizeof(xclReadBusStatus_response::_impl_.uncomplete_wr_id_)
      - PROTOBUF_FIELD_OFFSET(xclReadBusStatus_response, _impl_.uncomplete_rd_cycles_)>(
          reinterpret_cast<char*>(&_impl_.uncomplete_rd_cycles_),
          reinterpret_cast<char*>(&other->_impl_.uncomplete_rd_cycles_));
}

::google::protobuf::Metadata xclReadBusStatus_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[42]);
}
// ===================================================================

class xclGetDebugMessages_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDebugMessages_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_force(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclGetDebugMessages_call::xclGetDebugMessages_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGetDebugMessages_call)
}
xclGetDebugMessages_call::xclGetDebugMessages_call(
    ::google::protobuf::Arena* arena, const xclGetDebugMessages_call& from)
    : xclGetDebugMessages_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGetDebugMessages_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGetDebugMessages_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, ack_),
           0,
           offsetof(Impl_, force_) -
               offsetof(Impl_, ack_) +
               sizeof(Impl_::force_));
}
xclGetDebugMessages_call::~xclGetDebugMessages_call() {
  // @@protoc_insertion_point(destructor:xclGetDebugMessages_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGetDebugMessages_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGetDebugMessages_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDebugMessages_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  ::memset(&_impl_.ack_, 0, static_cast<::size_t>(
      reinterpret_cast<char*>(&_impl_.force_) -
      reinterpret_cast<char*>(&_impl_.ack_)) + sizeof(_impl_.force_));
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGetDebugMessages_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGetDebugMessages_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGetDebugMessages_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool force = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGetDebugMessages_call, _impl_.force_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.force_)}},
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGetDebugMessages_call, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bool force = 2;
    {PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.force_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGetDebugMessages_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDebugMessages_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  // optional bool force = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        2, this->_internal_force(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDebugMessages_call)
  return target;
}

::size_t xclGetDebugMessages_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDebugMessages_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bool ack = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 2;
    }

    // optional bool force = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGetDebugMessages_call::_class_data_ = {
    xclGetDebugMessages_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGetDebugMessages_call::GetClassData() const {
  return &_class_data_;
}

void xclGetDebugMessages_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGetDebugMessages_call*>(&to_msg);
  auto& from = static_cast<const xclGetDebugMessages_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDebugMessages_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.force_ = from._impl_.force_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDebugMessages_call::CopyFrom(const xclGetDebugMessages_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDebugMessages_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGetDebugMessages_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclGetDebugMessages_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGetDebugMessages_call::InternalSwap(xclGetDebugMessages_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.force_)
      + sizeof(xclGetDebugMessages_call::_impl_.force_)
      - PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_call, _impl_.ack_)>(
          reinterpret_cast<char*>(&_impl_.ack_),
          reinterpret_cast<char*>(&other->_impl_.ack_));
}

::google::protobuf::Metadata xclGetDebugMessages_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[43]);
}
// ===================================================================

class xclGetDebugMessages_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGetDebugMessages_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_._has_bits_);
  static void set_has_display_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_log_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_stop_msgs(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclGetDebugMessages_response::xclGetDebugMessages_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGetDebugMessages_response)
}
inline PROTOBUF_NDEBUG_INLINE xclGetDebugMessages_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        display_msgs_(arena, from.display_msgs_),
        log_msgs_(arena, from.log_msgs_),
        stop_msgs_(arena, from.stop_msgs_) {}

xclGetDebugMessages_response::xclGetDebugMessages_response(
    ::google::protobuf::Arena* arena,
    const xclGetDebugMessages_response& from)
    : ::google::protobuf::Message(arena) {
  xclGetDebugMessages_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclGetDebugMessages_response)
}
inline PROTOBUF_NDEBUG_INLINE xclGetDebugMessages_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        display_msgs_(arena),
        log_msgs_(arena),
        stop_msgs_(arena) {}

inline void xclGetDebugMessages_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclGetDebugMessages_response::~xclGetDebugMessages_response() {
  // @@protoc_insertion_point(destructor:xclGetDebugMessages_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGetDebugMessages_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.display_msgs_.Destroy();
  _impl_.log_msgs_.Destroy();
  _impl_.stop_msgs_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGetDebugMessages_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGetDebugMessages_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.display_msgs_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.log_msgs_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.stop_msgs_.ClearNonDefaultToEmpty();
    }
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGetDebugMessages_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 0, 2> xclGetDebugMessages_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGetDebugMessages_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional bytes display_msgs = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.display_msgs_)}},
    // optional bytes log_msgs = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 1, 0, PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.log_msgs_)}},
    // optional bytes stop_msgs = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 2, 0, PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.stop_msgs_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bytes display_msgs = 1;
    {PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.display_msgs_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional bytes log_msgs = 2;
    {PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.log_msgs_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional bytes stop_msgs = 3;
    {PROTOBUF_FIELD_OFFSET(xclGetDebugMessages_response, _impl_.stop_msgs_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGetDebugMessages_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGetDebugMessages_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes display_msgs = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_display_msgs();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  // optional bytes log_msgs = 2;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_log_msgs();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // optional bytes stop_msgs = 3;
  if (cached_has_bits & 0x00000004u) {
    const std::string& _s = this->_internal_stop_msgs();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGetDebugMessages_response)
  return target;
}

::size_t xclGetDebugMessages_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGetDebugMessages_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional bytes display_msgs = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_display_msgs());
    }

    // optional bytes log_msgs = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_log_msgs());
    }

    // optional bytes stop_msgs = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_stop_msgs());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGetDebugMessages_response::_class_data_ = {
    xclGetDebugMessages_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGetDebugMessages_response::GetClassData() const {
  return &_class_data_;
}

void xclGetDebugMessages_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGetDebugMessages_response*>(&to_msg);
  auto& from = static_cast<const xclGetDebugMessages_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGetDebugMessages_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_display_msgs(from._internal_display_msgs());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_log_msgs(from._internal_log_msgs());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_stop_msgs(from._internal_stop_msgs());
    }
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGetDebugMessages_response::CopyFrom(const xclGetDebugMessages_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGetDebugMessages_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGetDebugMessages_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclGetDebugMessages_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGetDebugMessages_response::InternalSwap(xclGetDebugMessages_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.display_msgs_, &other->_impl_.display_msgs_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.log_msgs_, &other->_impl_.log_msgs_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.stop_msgs_, &other->_impl_.stop_msgs_, arena);
}

::google::protobuf::Metadata xclGetDebugMessages_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[44]);
}
// ===================================================================

class xclImportBO_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclImportBO_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_._has_bits_);
  static void set_has_dst_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclImportBO_call::xclImportBO_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclImportBO_call)
}
inline PROTOBUF_NDEBUG_INLINE xclImportBO_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dst_filename_(arena, from.dst_filename_) {}

xclImportBO_call::xclImportBO_call(
    ::google::protobuf::Arena* arena,
    const xclImportBO_call& from)
    : ::google::protobuf::Message(arena) {
  xclImportBO_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, offset_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, offset_),
           offsetof(Impl_, size_) -
               offsetof(Impl_, offset_) +
               sizeof(Impl_::size_));

  // @@protoc_insertion_point(copy_constructor:xclImportBO_call)
}
inline PROTOBUF_NDEBUG_INLINE xclImportBO_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dst_filename_(arena) {}

inline void xclImportBO_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, offset_),
           0,
           offsetof(Impl_, size_) -
               offsetof(Impl_, offset_) +
               sizeof(Impl_::size_));
}
xclImportBO_call::~xclImportBO_call() {
  // @@protoc_insertion_point(destructor:xclImportBO_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclImportBO_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dst_filename_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclImportBO_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclImportBO_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dst_filename_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.offset_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.offset_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclImportBO_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 37, 2> xclImportBO_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclImportBO_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional string dst_filename = 1;
    {::_pbi::TcParser::FastSS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.dst_filename_)}},
    // optional uint64 offset = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclImportBO_call, _impl_.offset_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.offset_)}},
    // optional uint64 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclImportBO_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional string dst_filename = 1;
    {PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.dst_filename_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional uint64 offset = 2;
    {PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.offset_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
    "\20\14\0\0\0\0\0\0"
    "xclImportBO_call"
    "dst_filename"
  }},
};

::uint8_t* xclImportBO_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclImportBO_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional string dst_filename = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dst_filename();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclImportBO_call.dst_filename");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  // optional uint64 offset = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_offset(), target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclImportBO_call)
  return target;
}

::size_t xclImportBO_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclImportBO_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string dst_filename = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_dst_filename());
    }

    // optional uint64 offset = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_offset());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclImportBO_call::_class_data_ = {
    xclImportBO_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclImportBO_call::GetClassData() const {
  return &_class_data_;
}

void xclImportBO_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclImportBO_call*>(&to_msg);
  auto& from = static_cast<const xclImportBO_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclImportBO_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dst_filename(from._internal_dst_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclImportBO_call::CopyFrom(const xclImportBO_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclImportBO_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclImportBO_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclImportBO_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclImportBO_call::InternalSwap(xclImportBO_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dst_filename_, &other->_impl_.dst_filename_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.size_)
      + sizeof(xclImportBO_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclImportBO_call, _impl_.offset_)>(
          reinterpret_cast<char*>(&_impl_.offset_),
          reinterpret_cast<char*>(&other->_impl_.offset_));
}

::google::protobuf::Metadata xclImportBO_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[45]);
}
// ===================================================================

class xclImportBO_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclImportBO_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclImportBO_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclImportBO_response::xclImportBO_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclImportBO_response)
}
xclImportBO_response::xclImportBO_response(
    ::google::protobuf::Arena* arena, const xclImportBO_response& from)
    : xclImportBO_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclImportBO_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclImportBO_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclImportBO_response::~xclImportBO_response() {
  // @@protoc_insertion_point(destructor:xclImportBO_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclImportBO_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclImportBO_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclImportBO_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclImportBO_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclImportBO_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclImportBO_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclImportBO_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclImportBO_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclImportBO_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclImportBO_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclImportBO_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclImportBO_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclImportBO_response)
  return target;
}

::size_t xclImportBO_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclImportBO_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclImportBO_response::_class_data_ = {
    xclImportBO_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclImportBO_response::GetClassData() const {
  return &_class_data_;
}

void xclImportBO_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclImportBO_response*>(&to_msg);
  auto& from = static_cast<const xclImportBO_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclImportBO_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclImportBO_response::CopyFrom(const xclImportBO_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclImportBO_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclImportBO_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclImportBO_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclImportBO_response::InternalSwap(xclImportBO_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclImportBO_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[46]);
}
// ===================================================================

class xclCopyBO_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBO_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_._has_bits_);
  static void set_has_src_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dst_filename(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_src_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_dst_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclCopyBO_call::xclCopyBO_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBO_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBO_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dst_filename_(arena, from.dst_filename_) {}

xclCopyBO_call::xclCopyBO_call(
    ::google::protobuf::Arena* arena,
    const xclCopyBO_call& from)
    : ::google::protobuf::Message(arena) {
  xclCopyBO_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, src_handle_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, src_handle_),
           offsetof(Impl_, dst_offset_) -
               offsetof(Impl_, src_handle_) +
               sizeof(Impl_::dst_offset_));

  // @@protoc_insertion_point(copy_constructor:xclCopyBO_call)
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBO_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dst_filename_(arena) {}

inline void xclCopyBO_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, src_handle_),
           0,
           offsetof(Impl_, dst_offset_) -
               offsetof(Impl_, src_handle_) +
               sizeof(Impl_::dst_offset_));
}
xclCopyBO_call::~xclCopyBO_call() {
  // @@protoc_insertion_point(destructor:xclCopyBO_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBO_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dst_filename_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBO_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBO_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dst_filename_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.src_handle_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.dst_offset_) -
        reinterpret_cast<char*>(&_impl_.src_handle_)) + sizeof(_impl_.dst_offset_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBO_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 35, 2> xclCopyBO_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_._has_bits_),
    0, // no _extensions_
    5, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967264,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBO_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 src_handle = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBO_call, _impl_.src_handle_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_handle_)}},
    // optional string dst_filename = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_filename_)}},
    // optional uint64 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBO_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.size_)}},
    // optional uint64 src_offset = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBO_call, _impl_.src_offset_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_offset_)}},
    // optional uint64 dst_offset = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCopyBO_call, _impl_.dst_offset_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_offset_)}},
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 src_handle = 1;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_handle_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional string dst_filename = 2;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_filename_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
    // optional uint64 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 src_offset = 4;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_offset_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 dst_offset = 5;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_offset_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
    "\16\0\14\0\0\0\0\0"
    "xclCopyBO_call"
    "dst_filename"
  }},
};

::uint8_t* xclCopyBO_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBO_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 src_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_src_handle(), target);
  }

  // optional string dst_filename = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dst_filename();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclCopyBO_call.dst_filename");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_size(), target);
  }

  // optional uint64 src_offset = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_src_offset(), target);
  }

  // optional uint64 dst_offset = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_dst_offset(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBO_call)
  return target;
}

::size_t xclCopyBO_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBO_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional string dst_filename = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_dst_filename());
    }

    // optional uint64 src_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_src_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // optional uint64 src_offset = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_src_offset());
    }

    // optional uint64 dst_offset = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_dst_offset());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBO_call::_class_data_ = {
    xclCopyBO_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBO_call::GetClassData() const {
  return &_class_data_;
}

void xclCopyBO_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBO_call*>(&to_msg);
  auto& from = static_cast<const xclCopyBO_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBO_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dst_filename(from._internal_dst_filename());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.src_handle_ = from._impl_.src_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.src_offset_ = from._impl_.src_offset_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.dst_offset_ = from._impl_.dst_offset_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBO_call::CopyFrom(const xclCopyBO_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBO_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBO_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclCopyBO_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBO_call::InternalSwap(xclCopyBO_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dst_filename_, &other->_impl_.dst_filename_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.dst_offset_)
      + sizeof(xclCopyBO_call::_impl_.dst_offset_)
      - PROTOBUF_FIELD_OFFSET(xclCopyBO_call, _impl_.src_handle_)>(
          reinterpret_cast<char*>(&_impl_.src_handle_),
          reinterpret_cast<char*>(&other->_impl_.src_handle_));
}

::google::protobuf::Metadata xclCopyBO_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[47]);
}
// ===================================================================

class xclCopyBO_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCopyBO_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCopyBO_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclCopyBO_response::xclCopyBO_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCopyBO_response)
}
xclCopyBO_response::xclCopyBO_response(
    ::google::protobuf::Arena* arena, const xclCopyBO_response& from)
    : xclCopyBO_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclCopyBO_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclCopyBO_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclCopyBO_response::~xclCopyBO_response() {
  // @@protoc_insertion_point(destructor:xclCopyBO_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCopyBO_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCopyBO_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCopyBO_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCopyBO_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclCopyBO_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCopyBO_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCopyBO_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclCopyBO_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclCopyBO_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclCopyBO_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCopyBO_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCopyBO_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCopyBO_response)
  return target;
}

::size_t xclCopyBO_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCopyBO_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCopyBO_response::_class_data_ = {
    xclCopyBO_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCopyBO_response::GetClassData() const {
  return &_class_data_;
}

void xclCopyBO_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCopyBO_response*>(&to_msg);
  auto& from = static_cast<const xclCopyBO_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCopyBO_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCopyBO_response::CopyFrom(const xclCopyBO_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCopyBO_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCopyBO_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclCopyBO_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCopyBO_response::InternalSwap(xclCopyBO_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclCopyBO_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[48]);
}
// ===================================================================

class xclCreateQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCreateQueue_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_._has_bits_);
  static void set_has_write(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_type(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_state(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_route(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_flow(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_qsize(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_desc_size(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
  static void set_has_flags(HasBits* has_bits) {
    (*has_bits)[0] |= 128u;
  }
};

xclCreateQueue_call::xclCreateQueue_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCreateQueue_call)
}
xclCreateQueue_call::xclCreateQueue_call(
    ::google::protobuf::Arena* arena, const xclCreateQueue_call& from)
    : xclCreateQueue_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclCreateQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclCreateQueue_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, write_),
           0,
           offsetof(Impl_, flags_) -
               offsetof(Impl_, write_) +
               sizeof(Impl_::flags_));
}
xclCreateQueue_call::~xclCreateQueue_call() {
  // @@protoc_insertion_point(destructor:xclCreateQueue_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCreateQueue_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCreateQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCreateQueue_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    ::memset(&_impl_.write_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.flags_) -
        reinterpret_cast<char*>(&_impl_.write_)) + sizeof(_impl_.flags_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCreateQueue_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 8, 0, 0, 2> xclCreateQueue_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_._has_bits_),
    0, // no _extensions_
    8, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967040,  // skipmap
    offsetof(decltype(_table_), field_entries),
    8,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCreateQueue_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 flags = 8;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCreateQueue_call, _impl_.flags_), 7>(),
     {64, 7, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flags_)}},
    // optional bool write = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclCreateQueue_call, _impl_.write_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.write_)}},
    // optional uint32 type = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCreateQueue_call, _impl_.type_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.type_)}},
    // optional uint32 state = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCreateQueue_call, _impl_.state_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.state_)}},
    // optional uint32 route = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCreateQueue_call, _impl_.route_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.route_)}},
    // optional uint64 flow = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCreateQueue_call, _impl_.flow_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flow_)}},
    // optional uint32 qsize = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCreateQueue_call, _impl_.qsize_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.qsize_)}},
    // optional uint32 desc_size = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclCreateQueue_call, _impl_.desc_size_), 6>(),
     {56, 6, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.desc_size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool write = 1;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.write_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional uint32 type = 2;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.type_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 state = 3;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.state_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 route = 4;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.route_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint64 flow = 5;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flow_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint32 qsize = 6;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.qsize_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 desc_size = 7;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.desc_size_), _Internal::kHasBitsOffset + 6, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint64 flags = 8;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flags_), _Internal::kHasBitsOffset + 7, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCreateQueue_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCreateQueue_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool write = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_write(), target);
  }

  // optional uint32 type = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_type(), target);
  }

  // optional uint32 state = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_state(), target);
  }

  // optional uint32 route = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this->_internal_route(), target);
  }

  // optional uint64 flow = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_flow(), target);
  }

  // optional uint32 qsize = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        6, this->_internal_qsize(), target);
  }

  // optional uint32 desc_size = 7;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        7, this->_internal_desc_size(), target);
  }

  // optional uint64 flags = 8;
  if (cached_has_bits & 0x00000080u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        8, this->_internal_flags(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCreateQueue_call)
  return target;
}

::size_t xclCreateQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCreateQueue_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    // optional bool write = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 2;
    }

    // optional uint32 type = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_type());
    }

    // optional uint32 state = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_state());
    }

    // optional uint32 route = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_route());
    }

    // optional uint64 flow = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_flow());
    }

    // optional uint32 qsize = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_qsize());
    }

    // optional uint32 desc_size = 7;
    if (cached_has_bits & 0x00000040u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_desc_size());
    }

    // optional uint64 flags = 8;
    if (cached_has_bits & 0x00000080u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_flags());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCreateQueue_call::_class_data_ = {
    xclCreateQueue_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCreateQueue_call::GetClassData() const {
  return &_class_data_;
}

void xclCreateQueue_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCreateQueue_call*>(&to_msg);
  auto& from = static_cast<const xclCreateQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCreateQueue_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x000000ffu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.write_ = from._impl_.write_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.type_ = from._impl_.type_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.state_ = from._impl_.state_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.route_ = from._impl_.route_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.flow_ = from._impl_.flow_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.qsize_ = from._impl_.qsize_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.desc_size_ = from._impl_.desc_size_;
    }
    if (cached_has_bits & 0x00000080u) {
      _this->_impl_.flags_ = from._impl_.flags_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCreateQueue_call::CopyFrom(const xclCreateQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCreateQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCreateQueue_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclCreateQueue_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCreateQueue_call::InternalSwap(xclCreateQueue_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.flags_)
      + sizeof(xclCreateQueue_call::_impl_.flags_)
      - PROTOBUF_FIELD_OFFSET(xclCreateQueue_call, _impl_.write_)>(
          reinterpret_cast<char*>(&_impl_.write_),
          reinterpret_cast<char*>(&other->_impl_.write_));
}

::google::protobuf::Metadata xclCreateQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[49]);
}
// ===================================================================

class xclCreateQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclCreateQueue_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclCreateQueue_response, _impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclCreateQueue_response::xclCreateQueue_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclCreateQueue_response)
}
xclCreateQueue_response::xclCreateQueue_response(
    ::google::protobuf::Arena* arena, const xclCreateQueue_response& from)
    : xclCreateQueue_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclCreateQueue_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclCreateQueue_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.q_handle_ = {};
}
xclCreateQueue_response::~xclCreateQueue_response() {
  // @@protoc_insertion_point(destructor:xclCreateQueue_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclCreateQueue_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclCreateQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclCreateQueue_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.q_handle_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclCreateQueue_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclCreateQueue_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclCreateQueue_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclCreateQueue_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 q_handle = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclCreateQueue_response, _impl_.q_handle_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclCreateQueue_response, _impl_.q_handle_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 q_handle = 1;
    {PROTOBUF_FIELD_OFFSET(xclCreateQueue_response, _impl_.q_handle_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclCreateQueue_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclCreateQueue_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_q_handle(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclCreateQueue_response)
  return target;
}

::size_t xclCreateQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclCreateQueue_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 q_handle = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_q_handle());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclCreateQueue_response::_class_data_ = {
    xclCreateQueue_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclCreateQueue_response::GetClassData() const {
  return &_class_data_;
}

void xclCreateQueue_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclCreateQueue_response*>(&to_msg);
  auto& from = static_cast<const xclCreateQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclCreateQueue_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_q_handle(from._internal_q_handle());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclCreateQueue_response::CopyFrom(const xclCreateQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclCreateQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclCreateQueue_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclCreateQueue_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclCreateQueue_response::InternalSwap(xclCreateQueue_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.q_handle_, other->_impl_.q_handle_);
}

::google::protobuf::Metadata xclCreateQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[50]);
}
// ===================================================================

class xclWriteQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteQueue_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_src(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_nonblocking(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_eot(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclWriteQueue_call::xclWriteQueue_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteQueue_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        src_(arena, from.src_) {}

xclWriteQueue_call::xclWriteQueue_call(
    ::google::protobuf::Arena* arena,
    const xclWriteQueue_call& from)
    : ::google::protobuf::Message(arena) {
  xclWriteQueue_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, q_handle_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, q_handle_),
           offsetof(Impl_, eot_) -
               offsetof(Impl_, q_handle_) +
               sizeof(Impl_::eot_));

  // @@protoc_insertion_point(copy_constructor:xclWriteQueue_call)
}
inline PROTOBUF_NDEBUG_INLINE xclWriteQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        src_(arena) {}

inline void xclWriteQueue_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, q_handle_),
           0,
           offsetof(Impl_, eot_) -
               offsetof(Impl_, q_handle_) +
               sizeof(Impl_::eot_));
}
xclWriteQueue_call::~xclWriteQueue_call() {
  // @@protoc_insertion_point(destructor:xclWriteQueue_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteQueue_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.src_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteQueue_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.src_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.q_handle_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.eot_) -
        reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteQueue_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclWriteQueue_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967232,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteQueue_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 q_handle = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteQueue_call, _impl_.q_handle_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.q_handle_)}},
    // optional bytes src = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.src_)}},
    // optional uint64 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteQueue_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.size_)}},
    // optional uint64 req = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteQueue_call, _impl_.req_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.req_)}},
    // optional bool nonblocking = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclWriteQueue_call, _impl_.nonblocking_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.nonblocking_)}},
    // optional bool eot = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclWriteQueue_call, _impl_.eot_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.eot_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 q_handle = 1;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.q_handle_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bytes src = 2;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.src_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 req = 4;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.req_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bool nonblocking = 5;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.nonblocking_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bool eot = 6;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.eot_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclWriteQueue_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteQueue_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_q_handle(), target);
  }

  // optional bytes src = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_src();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_size(), target);
  }

  // optional uint64 req = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_req(), target);
  }

  // optional bool nonblocking = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        5, this->_internal_nonblocking(), target);
  }

  // optional bool eot = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        6, this->_internal_eot(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteQueue_call)
  return target;
}

::size_t xclWriteQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteQueue_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional bytes src = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_src());
    }

    // optional uint64 q_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_q_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // optional uint64 req = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_req());
    }

    // optional bool nonblocking = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += 2;
    }

    // optional bool eot = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteQueue_call::_class_data_ = {
    xclWriteQueue_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteQueue_call::GetClassData() const {
  return &_class_data_;
}

void xclWriteQueue_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteQueue_call*>(&to_msg);
  auto& from = static_cast<const xclWriteQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteQueue_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_src(from._internal_src());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.q_handle_ = from._impl_.q_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.req_ = from._impl_.req_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.nonblocking_ = from._impl_.nonblocking_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.eot_ = from._impl_.eot_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteQueue_call::CopyFrom(const xclWriteQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteQueue_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclWriteQueue_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteQueue_call::InternalSwap(xclWriteQueue_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.src_, &other->_impl_.src_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.eot_)
      + sizeof(xclWriteQueue_call::_impl_.eot_)
      - PROTOBUF_FIELD_OFFSET(xclWriteQueue_call, _impl_.q_handle_)>(
          reinterpret_cast<char*>(&_impl_.q_handle_),
          reinterpret_cast<char*>(&other->_impl_.q_handle_));
}

::google::protobuf::Metadata xclWriteQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[51]);
}
// ===================================================================

class xclWriteQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclWriteQueue_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclWriteQueue_response, _impl_._has_bits_);
  static void set_has_written_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclWriteQueue_response::xclWriteQueue_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclWriteQueue_response)
}
xclWriteQueue_response::xclWriteQueue_response(
    ::google::protobuf::Arena* arena, const xclWriteQueue_response& from)
    : xclWriteQueue_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclWriteQueue_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclWriteQueue_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.written_size_ = {};
}
xclWriteQueue_response::~xclWriteQueue_response() {
  // @@protoc_insertion_point(destructor:xclWriteQueue_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclWriteQueue_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclWriteQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclWriteQueue_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.written_size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclWriteQueue_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclWriteQueue_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclWriteQueue_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclWriteQueue_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 written_size = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclWriteQueue_response, _impl_.written_size_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclWriteQueue_response, _impl_.written_size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 written_size = 1;
    {PROTOBUF_FIELD_OFFSET(xclWriteQueue_response, _impl_.written_size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclWriteQueue_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclWriteQueue_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 written_size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_written_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclWriteQueue_response)
  return target;
}

::size_t xclWriteQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclWriteQueue_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 written_size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_written_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclWriteQueue_response::_class_data_ = {
    xclWriteQueue_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclWriteQueue_response::GetClassData() const {
  return &_class_data_;
}

void xclWriteQueue_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclWriteQueue_response*>(&to_msg);
  auto& from = static_cast<const xclWriteQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclWriteQueue_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_written_size(from._internal_written_size());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclWriteQueue_response::CopyFrom(const xclWriteQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclWriteQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclWriteQueue_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclWriteQueue_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclWriteQueue_response::InternalSwap(xclWriteQueue_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.written_size_, other->_impl_.written_size_);
}

::google::protobuf::Metadata xclWriteQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[52]);
}
// ===================================================================

class xclReadQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadQueue_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_nonblocking(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_eot(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
};

xclReadQueue_call::xclReadQueue_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadQueue_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dest_(arena, from.dest_) {}

xclReadQueue_call::xclReadQueue_call(
    ::google::protobuf::Arena* arena,
    const xclReadQueue_call& from)
    : ::google::protobuf::Message(arena) {
  xclReadQueue_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, q_handle_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, q_handle_),
           offsetof(Impl_, eot_) -
               offsetof(Impl_, q_handle_) +
               sizeof(Impl_::eot_));

  // @@protoc_insertion_point(copy_constructor:xclReadQueue_call)
}
inline PROTOBUF_NDEBUG_INLINE xclReadQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dest_(arena) {}

inline void xclReadQueue_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, q_handle_),
           0,
           offsetof(Impl_, eot_) -
               offsetof(Impl_, q_handle_) +
               sizeof(Impl_::eot_));
}
xclReadQueue_call::~xclReadQueue_call() {
  // @@protoc_insertion_point(destructor:xclReadQueue_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadQueue_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dest_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadQueue_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000003eu) {
    ::memset(&_impl_.q_handle_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.eot_) -
        reinterpret_cast<char*>(&_impl_.q_handle_)) + sizeof(_impl_.eot_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadQueue_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 6, 0, 0, 2> xclReadQueue_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967232,  // skipmap
    offsetof(decltype(_table_), field_entries),
    6,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadQueue_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 q_handle = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadQueue_call, _impl_.q_handle_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.q_handle_)}},
    // optional bytes dest = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.dest_)}},
    // optional uint64 size = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadQueue_call, _impl_.size_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.size_)}},
    // optional uint64 req = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadQueue_call, _impl_.req_), 3>(),
     {32, 3, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.req_)}},
    // optional bool nonblocking = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclReadQueue_call, _impl_.nonblocking_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.nonblocking_)}},
    // optional bool eot = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclReadQueue_call, _impl_.eot_), 5>(),
     {48, 5, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.eot_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 q_handle = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.q_handle_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bytes dest = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.dest_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 size = 3;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 req = 4;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.req_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional bool nonblocking = 5;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.nonblocking_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional bool eot = 6;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.eot_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadQueue_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadQueue_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_q_handle(), target);
  }

  // optional bytes dest = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dest();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // optional uint64 size = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_size(), target);
  }

  // optional uint64 req = 4;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_req(), target);
  }

  // optional bool nonblocking = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        5, this->_internal_nonblocking(), target);
  }

  // optional bool eot = 6;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        6, this->_internal_eot(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadQueue_call)
  return target;
}

::size_t xclReadQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadQueue_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    // optional bytes dest = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dest());
    }

    // optional uint64 q_handle = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_q_handle());
    }

    // optional uint64 size = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // optional uint64 req = 4;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_req());
    }

    // optional bool nonblocking = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += 2;
    }

    // optional bool eot = 6;
    if (cached_has_bits & 0x00000020u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadQueue_call::_class_data_ = {
    xclReadQueue_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadQueue_call::GetClassData() const {
  return &_class_data_;
}

void xclReadQueue_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadQueue_call*>(&to_msg);
  auto& from = static_cast<const xclReadQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadQueue_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000003fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.q_handle_ = from._impl_.q_handle_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.req_ = from._impl_.req_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.nonblocking_ = from._impl_.nonblocking_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.eot_ = from._impl_.eot_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadQueue_call::CopyFrom(const xclReadQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadQueue_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclReadQueue_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadQueue_call::InternalSwap(xclReadQueue_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dest_, &other->_impl_.dest_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.eot_)
      + sizeof(xclReadQueue_call::_impl_.eot_)
      - PROTOBUF_FIELD_OFFSET(xclReadQueue_call, _impl_.q_handle_)>(
          reinterpret_cast<char*>(&_impl_.q_handle_),
          reinterpret_cast<char*>(&other->_impl_.q_handle_));
}

::google::protobuf::Metadata xclReadQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[53]);
}
// ===================================================================

class xclReadQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclReadQueue_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_._has_bits_);
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclReadQueue_response::xclReadQueue_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclReadQueue_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadQueue_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dest_(arena, from.dest_) {}

xclReadQueue_response::xclReadQueue_response(
    ::google::protobuf::Arena* arena,
    const xclReadQueue_response& from)
    : ::google::protobuf::Message(arena) {
  xclReadQueue_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.size_ = from._impl_.size_;

  // @@protoc_insertion_point(copy_constructor:xclReadQueue_response)
}
inline PROTOBUF_NDEBUG_INLINE xclReadQueue_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dest_(arena) {}

inline void xclReadQueue_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclReadQueue_response::~xclReadQueue_response() {
  // @@protoc_insertion_point(destructor:xclReadQueue_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclReadQueue_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dest_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclReadQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclReadQueue_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclReadQueue_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclReadQueue_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclReadQueue_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclReadQueue_response, _impl_.size_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_.size_)}},
    // optional bytes dest = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_.dest_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bytes dest = 1;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_.dest_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclReadQueue_response, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclReadQueue_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclReadQueue_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes dest = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dest();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclReadQueue_response)
  return target;
}

::size_t xclReadQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclReadQueue_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes dest = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dest());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclReadQueue_response::_class_data_ = {
    xclReadQueue_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclReadQueue_response::GetClassData() const {
  return &_class_data_;
}

void xclReadQueue_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclReadQueue_response*>(&to_msg);
  auto& from = static_cast<const xclReadQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclReadQueue_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclReadQueue_response::CopyFrom(const xclReadQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclReadQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclReadQueue_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclReadQueue_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclReadQueue_response::InternalSwap(xclReadQueue_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dest_, &other->_impl_.dest_, arena);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclReadQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[54]);
}
// ===================================================================

class xclPollCompletion_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPollCompletion_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPollCompletion_call, _impl_._has_bits_);
  static void set_has_req(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPollCompletion_call::xclPollCompletion_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_call)
}
xclPollCompletion_call::xclPollCompletion_call(
    ::google::protobuf::Arena* arena, const xclPollCompletion_call& from)
    : xclPollCompletion_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPollCompletion_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPollCompletion_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.req_ = {};
}
xclPollCompletion_call::~xclPollCompletion_call() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPollCompletion_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPollCompletion_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.req_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPollCompletion_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclPollCompletion_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPollCompletion_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPollCompletion_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 req = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPollCompletion_call, _impl_.req_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclPollCompletion_call, _impl_.req_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 req = 1;
    {PROTOBUF_FIELD_OFFSET(xclPollCompletion_call, _impl_.req_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPollCompletion_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 req = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_req(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_call)
  return target;
}

::size_t xclPollCompletion_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 req = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_req());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPollCompletion_call::_class_data_ = {
    xclPollCompletion_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPollCompletion_call::GetClassData() const {
  return &_class_data_;
}

void xclPollCompletion_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_call*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_req(from._internal_req());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_call::CopyFrom(const xclPollCompletion_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPollCompletion_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPollCompletion_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPollCompletion_call::InternalSwap(xclPollCompletion_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.req_, other->_impl_.req_);
}

::google::protobuf::Metadata xclPollCompletion_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[55]);
}
// ===================================================================

class xclPollCompletion_response_request::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPollCompletion_response_request>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_._has_bits_);
  static void set_has_dest(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
};

xclPollCompletion_response_request::xclPollCompletion_response_request(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_response.request)
}
inline PROTOBUF_NDEBUG_INLINE xclPollCompletion_response_request::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        dest_(arena, from.dest_) {}

xclPollCompletion_response_request::xclPollCompletion_response_request(
    ::google::protobuf::Arena* arena,
    const xclPollCompletion_response_request& from)
    : ::google::protobuf::Message(arena) {
  xclPollCompletion_response_request* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.size_ = from._impl_.size_;

  // @@protoc_insertion_point(copy_constructor:xclPollCompletion_response.request)
}
inline PROTOBUF_NDEBUG_INLINE xclPollCompletion_response_request::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        dest_(arena) {}

inline void xclPollCompletion_response_request::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclPollCompletion_response_request::~xclPollCompletion_response_request() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_response.request)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPollCompletion_response_request::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.dest_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPollCompletion_response_request::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_response.request)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.dest_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPollCompletion_response_request::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclPollCompletion_response_request::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPollCompletion_response_request_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPollCompletion_response_request, _impl_.size_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_.size_)}},
    // optional bytes dest = 1;
    {::_pbi::TcParser::FastBS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_.dest_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bytes dest = 1;
    {PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_.dest_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // optional uint64 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclPollCompletion_response_request, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPollCompletion_response_request::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_response.request)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bytes dest = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_dest();
    target = stream->WriteBytesMaybeAliased(1, _s, target);
  }

  // optional uint64 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_response.request)
  return target;
}

::size_t xclPollCompletion_response_request::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_response.request)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional bytes dest = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_dest());
    }

    // optional uint64 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPollCompletion_response_request::_class_data_ = {
    xclPollCompletion_response_request::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPollCompletion_response_request::GetClassData() const {
  return &_class_data_;
}

void xclPollCompletion_response_request::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_response_request*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_response_request&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_response.request)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_dest(from._internal_dest());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_response_request::CopyFrom(const xclPollCompletion_response_request& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_response.request)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPollCompletion_response_request::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPollCompletion_response_request::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPollCompletion_response_request::InternalSwap(xclPollCompletion_response_request* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.dest_, &other->_impl_.dest_, arena);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclPollCompletion_response_request::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[56]);
}
// ===================================================================

class xclPollCompletion_response::_Internal {
 public:
};

xclPollCompletion_response::xclPollCompletion_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPollCompletion_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPollCompletion_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : fullrequest_{visibility, arena, from.fullrequest_},
        _cached_size_{0} {}

xclPollCompletion_response::xclPollCompletion_response(
    ::google::protobuf::Arena* arena,
    const xclPollCompletion_response& from)
    : ::google::protobuf::Message(arena) {
  xclPollCompletion_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclPollCompletion_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPollCompletion_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : fullrequest_{visibility, arena},
        _cached_size_{0} {}

inline void xclPollCompletion_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclPollCompletion_response::~xclPollCompletion_response() {
  // @@protoc_insertion_point(destructor:xclPollCompletion_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPollCompletion_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPollCompletion_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPollCompletion_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.fullrequest_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPollCompletion_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2> xclPollCompletion_response::_table_ = {
  {
    0,  // no _has_bits_
    0, // no _extensions_
    3, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967291,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclPollCompletion_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // repeated .xclPollCompletion_response.request fullrequest = 3;
    {::_pbi::TcParser::FastMtR1,
     {26, 63, 0, PROTOBUF_FIELD_OFFSET(xclPollCompletion_response, _impl_.fullrequest_)}},
  }}, {{
    65535, 65535
  }}, {{
    // repeated .xclPollCompletion_response.request fullrequest = 3;
    {PROTOBUF_FIELD_OFFSET(xclPollCompletion_response, _impl_.fullrequest_), 0, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclPollCompletion_response_request>()},
  }}, {{
  }},
};

::uint8_t* xclPollCompletion_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPollCompletion_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  // repeated .xclPollCompletion_response.request fullrequest = 3;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_fullrequest_size()); i < n; i++) {
    const auto& repfield = this->_internal_fullrequest().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(3, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPollCompletion_response)
  return target;
}

::size_t xclPollCompletion_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPollCompletion_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPollCompletion_response.request fullrequest = 3;
  total_size += 1UL * this->_internal_fullrequest_size();
  for (const auto& msg : this->_internal_fullrequest()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPollCompletion_response::_class_data_ = {
    xclPollCompletion_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPollCompletion_response::GetClassData() const {
  return &_class_data_;
}

void xclPollCompletion_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPollCompletion_response*>(&to_msg);
  auto& from = static_cast<const xclPollCompletion_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPollCompletion_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_fullrequest()->MergeFrom(
      from._internal_fullrequest());
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPollCompletion_response::CopyFrom(const xclPollCompletion_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPollCompletion_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPollCompletion_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPollCompletion_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPollCompletion_response::InternalSwap(xclPollCompletion_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.fullrequest_.InternalSwap(&other->_impl_.fullrequest_);
}

::google::protobuf::Metadata xclPollCompletion_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[57]);
}
// ===================================================================

class xclDestroyQueue_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclDestroyQueue_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclDestroyQueue_call, _impl_._has_bits_);
  static void set_has_q_handle(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclDestroyQueue_call::xclDestroyQueue_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclDestroyQueue_call)
}
xclDestroyQueue_call::xclDestroyQueue_call(
    ::google::protobuf::Arena* arena, const xclDestroyQueue_call& from)
    : xclDestroyQueue_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclDestroyQueue_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclDestroyQueue_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.q_handle_ = {};
}
xclDestroyQueue_call::~xclDestroyQueue_call() {
  // @@protoc_insertion_point(destructor:xclDestroyQueue_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclDestroyQueue_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclDestroyQueue_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclDestroyQueue_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.q_handle_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclDestroyQueue_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclDestroyQueue_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclDestroyQueue_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclDestroyQueue_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 q_handle = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclDestroyQueue_call, _impl_.q_handle_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclDestroyQueue_call, _impl_.q_handle_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 q_handle = 1;
    {PROTOBUF_FIELD_OFFSET(xclDestroyQueue_call, _impl_.q_handle_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclDestroyQueue_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclDestroyQueue_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 q_handle = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_q_handle(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclDestroyQueue_call)
  return target;
}

::size_t xclDestroyQueue_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclDestroyQueue_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional uint64 q_handle = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_q_handle());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclDestroyQueue_call::_class_data_ = {
    xclDestroyQueue_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclDestroyQueue_call::GetClassData() const {
  return &_class_data_;
}

void xclDestroyQueue_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclDestroyQueue_call*>(&to_msg);
  auto& from = static_cast<const xclDestroyQueue_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclDestroyQueue_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_q_handle(from._internal_q_handle());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclDestroyQueue_call::CopyFrom(const xclDestroyQueue_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclDestroyQueue_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclDestroyQueue_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclDestroyQueue_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclDestroyQueue_call::InternalSwap(xclDestroyQueue_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.q_handle_, other->_impl_.q_handle_);
}

::google::protobuf::Metadata xclDestroyQueue_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[58]);
}
// ===================================================================

class xclDestroyQueue_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclDestroyQueue_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclDestroyQueue_response, _impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclDestroyQueue_response::xclDestroyQueue_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclDestroyQueue_response)
}
xclDestroyQueue_response::xclDestroyQueue_response(
    ::google::protobuf::Arena* arena, const xclDestroyQueue_response& from)
    : xclDestroyQueue_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclDestroyQueue_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclDestroyQueue_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.success_ = {};
}
xclDestroyQueue_response::~xclDestroyQueue_response() {
  // @@protoc_insertion_point(destructor:xclDestroyQueue_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclDestroyQueue_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclDestroyQueue_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclDestroyQueue_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclDestroyQueue_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclDestroyQueue_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclDestroyQueue_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclDestroyQueue_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool success = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclDestroyQueue_response, _impl_.success_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclDestroyQueue_response, _impl_.success_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool success = 1;
    {PROTOBUF_FIELD_OFFSET(xclDestroyQueue_response, _impl_.success_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclDestroyQueue_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclDestroyQueue_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclDestroyQueue_response)
  return target;
}

::size_t xclDestroyQueue_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclDestroyQueue_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclDestroyQueue_response::_class_data_ = {
    xclDestroyQueue_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclDestroyQueue_response::GetClassData() const {
  return &_class_data_;
}

void xclDestroyQueue_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclDestroyQueue_response*>(&to_msg);
  auto& from = static_cast<const xclDestroyQueue_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclDestroyQueue_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclDestroyQueue_response::CopyFrom(const xclDestroyQueue_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclDestroyQueue_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclDestroyQueue_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclDestroyQueue_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclDestroyQueue_response::InternalSwap(xclDestroyQueue_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.success_, other->_impl_.success_);
}

::google::protobuf::Metadata xclDestroyQueue_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[59]);
}
// ===================================================================

class xclSetupInstance_call_argflowpair::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_call_argflowpair>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_._has_bits_);
  static void set_has_arg(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_flow(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_tag(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_call.argflowpair)
}
inline PROTOBUF_NDEBUG_INLINE xclSetupInstance_call_argflowpair::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        tag_(arena, from.tag_) {}

xclSetupInstance_call_argflowpair::xclSetupInstance_call_argflowpair(
    ::google::protobuf::Arena* arena,
    const xclSetupInstance_call_argflowpair& from)
    : ::google::protobuf::Message(arena) {
  xclSetupInstance_call_argflowpair* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, arg_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, arg_),
           offsetof(Impl_, flow_) -
               offsetof(Impl_, arg_) +
               sizeof(Impl_::flow_));

  // @@protoc_insertion_point(copy_constructor:xclSetupInstance_call.argflowpair)
}
inline PROTOBUF_NDEBUG_INLINE xclSetupInstance_call_argflowpair::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        tag_(arena) {}

inline void xclSetupInstance_call_argflowpair::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, arg_),
           0,
           offsetof(Impl_, flow_) -
               offsetof(Impl_, arg_) +
               sizeof(Impl_::flow_));
}
xclSetupInstance_call_argflowpair::~xclSetupInstance_call_argflowpair() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_call.argflowpair)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetupInstance_call_argflowpair::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.tag_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetupInstance_call_argflowpair::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_call.argflowpair)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.tag_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.arg_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.flow_) -
        reinterpret_cast<char*>(&_impl_.arg_)) + sizeof(_impl_.flow_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetupInstance_call_argflowpair::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 45, 2> xclSetupInstance_call_argflowpair::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_._has_bits_),
    0, // no _extensions_
    4, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967281,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSetupInstance_call_argflowpair_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional string tag = 4;
    {::_pbi::TcParser::FastSS1,
     {34, 0, 0, PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.tag_)}},
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 arg = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSetupInstance_call_argflowpair, _impl_.arg_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.arg_)}},
    // optional uint64 flow = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSetupInstance_call_argflowpair, _impl_.flow_), 2>(),
     {24, 2, 0, PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.flow_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 arg = 2;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.arg_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 flow = 3;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.flow_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional string tag = 4;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.tag_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\41\0\0\3\0\0\0\0"
    "xclSetupInstance_call.argflowpair"
    "tag"
  }},
};

::uint8_t* xclSetupInstance_call_argflowpair::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_call.argflowpair)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 arg = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        2, this->_internal_arg(), target);
  }

  // optional uint64 flow = 3;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_flow(), target);
  }

  // optional string tag = 4;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_tag();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclSetupInstance_call.argflowpair.tag");
    target = stream->WriteStringMaybeAliased(4, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_call.argflowpair)
  return target;
}

::size_t xclSetupInstance_call_argflowpair::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_call.argflowpair)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional string tag = 4;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_tag());
    }

    // optional uint64 arg = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_arg());
    }

    // optional uint64 flow = 3;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_flow());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetupInstance_call_argflowpair::_class_data_ = {
    xclSetupInstance_call_argflowpair::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetupInstance_call_argflowpair::GetClassData() const {
  return &_class_data_;
}

void xclSetupInstance_call_argflowpair::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_call_argflowpair*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_call_argflowpair&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_call.argflowpair)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_tag(from._internal_tag());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.arg_ = from._impl_.arg_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.flow_ = from._impl_.flow_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_call_argflowpair::CopyFrom(const xclSetupInstance_call_argflowpair& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_call.argflowpair)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetupInstance_call_argflowpair::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetupInstance_call_argflowpair::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetupInstance_call_argflowpair::InternalSwap(xclSetupInstance_call_argflowpair* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.tag_, &other->_impl_.tag_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.flow_)
      + sizeof(xclSetupInstance_call_argflowpair::_impl_.flow_)
      - PROTOBUF_FIELD_OFFSET(xclSetupInstance_call_argflowpair, _impl_.arg_)>(
          reinterpret_cast<char*>(&_impl_.arg_),
          reinterpret_cast<char*>(&other->_impl_.arg_));
}

::google::protobuf::Metadata xclSetupInstance_call_argflowpair::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[60]);
}
// ===================================================================

class xclSetupInstance_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSetupInstance_call, _impl_._has_bits_);
  static void set_has_route(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_call::xclSetupInstance_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSetupInstance_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        setup_{visibility, arena, from.setup_} {}

xclSetupInstance_call::xclSetupInstance_call(
    ::google::protobuf::Arena* arena,
    const xclSetupInstance_call& from)
    : ::google::protobuf::Message(arena) {
  xclSetupInstance_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.route_ = from._impl_.route_;

  // @@protoc_insertion_point(copy_constructor:xclSetupInstance_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSetupInstance_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        setup_{visibility, arena} {}

inline void xclSetupInstance_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.route_ = {};
}
xclSetupInstance_call::~xclSetupInstance_call() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetupInstance_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetupInstance_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.setup_.Clear();
  _impl_.route_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetupInstance_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 2, 1, 0, 2> xclSetupInstance_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSetupInstance_call, _impl_._has_bits_),
    0, // no _extensions_
    5, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967278,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclSetupInstance_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 route = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSetupInstance_call, _impl_.route_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSetupInstance_call, _impl_.route_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 route = 1;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_call, _impl_.route_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // repeated .xclSetupInstance_call.argflowpair setup = 5;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_call, _impl_.setup_), -1, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclSetupInstance_call_argflowpair>()},
  }}, {{
  }},
};

::uint8_t* xclSetupInstance_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 route = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_route(), target);
  }

  // repeated .xclSetupInstance_call.argflowpair setup = 5;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_setup_size()); i < n; i++) {
    const auto& repfield = this->_internal_setup().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(5, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_call)
  return target;
}

::size_t xclSetupInstance_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclSetupInstance_call.argflowpair setup = 5;
  total_size += 1UL * this->_internal_setup_size();
  for (const auto& msg : this->_internal_setup()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  // optional uint64 route = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_route());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetupInstance_call::_class_data_ = {
    xclSetupInstance_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetupInstance_call::GetClassData() const {
  return &_class_data_;
}

void xclSetupInstance_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_call*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_setup()->MergeFrom(
      from._internal_setup());
  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_route(from._internal_route());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_call::CopyFrom(const xclSetupInstance_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetupInstance_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetupInstance_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetupInstance_call::InternalSwap(xclSetupInstance_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  _impl_.setup_.InternalSwap(&other->_impl_.setup_);
        swap(_impl_.route_, other->_impl_.route_);
}

::google::protobuf::Metadata xclSetupInstance_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[61]);
}
// ===================================================================

class xclSetupInstance_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSetupInstance_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSetupInstance_response, _impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclSetupInstance_response::xclSetupInstance_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSetupInstance_response)
}
xclSetupInstance_response::xclSetupInstance_response(
    ::google::protobuf::Arena* arena, const xclSetupInstance_response& from)
    : xclSetupInstance_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclSetupInstance_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclSetupInstance_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.success_ = {};
}
xclSetupInstance_response::~xclSetupInstance_response() {
  // @@protoc_insertion_point(destructor:xclSetupInstance_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSetupInstance_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSetupInstance_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSetupInstance_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSetupInstance_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclSetupInstance_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSetupInstance_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSetupInstance_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool success = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclSetupInstance_response, _impl_.success_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSetupInstance_response, _impl_.success_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool success = 1;
    {PROTOBUF_FIELD_OFFSET(xclSetupInstance_response, _impl_.success_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSetupInstance_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSetupInstance_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSetupInstance_response)
  return target;
}

::size_t xclSetupInstance_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSetupInstance_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSetupInstance_response::_class_data_ = {
    xclSetupInstance_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSetupInstance_response::GetClassData() const {
  return &_class_data_;
}

void xclSetupInstance_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSetupInstance_response*>(&to_msg);
  auto& from = static_cast<const xclSetupInstance_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSetupInstance_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSetupInstance_response::CopyFrom(const xclSetupInstance_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSetupInstance_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSetupInstance_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclSetupInstance_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSetupInstance_response::InternalSwap(xclSetupInstance_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.success_, other->_impl_.success_);
}

::google::protobuf::Metadata xclSetupInstance_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[62]);
}
// ===================================================================

class swemuDriverVersion_call::_Internal {
 public:
  using HasBits = decltype(std::declval<swemuDriverVersion_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(swemuDriverVersion_call, _impl_._has_bits_);
  static void set_has_version(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

swemuDriverVersion_call::swemuDriverVersion_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:swemuDriverVersion_call)
}
inline PROTOBUF_NDEBUG_INLINE swemuDriverVersion_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        version_(arena, from.version_) {}

swemuDriverVersion_call::swemuDriverVersion_call(
    ::google::protobuf::Arena* arena,
    const swemuDriverVersion_call& from)
    : ::google::protobuf::Message(arena) {
  swemuDriverVersion_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:swemuDriverVersion_call)
}
inline PROTOBUF_NDEBUG_INLINE swemuDriverVersion_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        version_(arena) {}

inline void swemuDriverVersion_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
swemuDriverVersion_call::~swemuDriverVersion_call() {
  // @@protoc_insertion_point(destructor:swemuDriverVersion_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void swemuDriverVersion_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.version_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void swemuDriverVersion_call::Clear() {
// @@protoc_insertion_point(message_clear_start:swemuDriverVersion_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.version_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* swemuDriverVersion_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 39, 2> swemuDriverVersion_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(swemuDriverVersion_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_swemuDriverVersion_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required string version = 1;
    {::_pbi::TcParser::FastSS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(swemuDriverVersion_call, _impl_.version_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required string version = 1;
    {PROTOBUF_FIELD_OFFSET(swemuDriverVersion_call, _impl_.version_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\27\7\0\0\0\0\0\0"
    "swemuDriverVersion_call"
    "version"
  }},
};

::uint8_t* swemuDriverVersion_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:swemuDriverVersion_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string version = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_version();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "swemuDriverVersion_call.version");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:swemuDriverVersion_call)
  return target;
}

::size_t swemuDriverVersion_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:swemuDriverVersion_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required string version = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                    this->_internal_version());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData swemuDriverVersion_call::_class_data_ = {
    swemuDriverVersion_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* swemuDriverVersion_call::GetClassData() const {
  return &_class_data_;
}

void swemuDriverVersion_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<swemuDriverVersion_call*>(&to_msg);
  auto& from = static_cast<const swemuDriverVersion_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:swemuDriverVersion_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_version(from._internal_version());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void swemuDriverVersion_call::CopyFrom(const swemuDriverVersion_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:swemuDriverVersion_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool swemuDriverVersion_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* swemuDriverVersion_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void swemuDriverVersion_call::InternalSwap(swemuDriverVersion_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.version_, &other->_impl_.version_, arena);
}

::google::protobuf::Metadata swemuDriverVersion_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[63]);
}
// ===================================================================

class swemuDriverVersion_response::_Internal {
 public:
  using HasBits = decltype(std::declval<swemuDriverVersion_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(swemuDriverVersion_response, _impl_._has_bits_);
  static void set_has_success(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

swemuDriverVersion_response::swemuDriverVersion_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:swemuDriverVersion_response)
}
swemuDriverVersion_response::swemuDriverVersion_response(
    ::google::protobuf::Arena* arena, const swemuDriverVersion_response& from)
    : swemuDriverVersion_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE swemuDriverVersion_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void swemuDriverVersion_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.success_ = {};
}
swemuDriverVersion_response::~swemuDriverVersion_response() {
  // @@protoc_insertion_point(destructor:swemuDriverVersion_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void swemuDriverVersion_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void swemuDriverVersion_response::Clear() {
// @@protoc_insertion_point(message_clear_start:swemuDriverVersion_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.success_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* swemuDriverVersion_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> swemuDriverVersion_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(swemuDriverVersion_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_swemuDriverVersion_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional bool success = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(swemuDriverVersion_response, _impl_.success_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(swemuDriverVersion_response, _impl_.success_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool success = 1;
    {PROTOBUF_FIELD_OFFSET(swemuDriverVersion_response, _impl_.success_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* swemuDriverVersion_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:swemuDriverVersion_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool success = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_success(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:swemuDriverVersion_response)
  return target;
}

::size_t swemuDriverVersion_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:swemuDriverVersion_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // optional bool success = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData swemuDriverVersion_response::_class_data_ = {
    swemuDriverVersion_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* swemuDriverVersion_response::GetClassData() const {
  return &_class_data_;
}

void swemuDriverVersion_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<swemuDriverVersion_response*>(&to_msg);
  auto& from = static_cast<const swemuDriverVersion_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:swemuDriverVersion_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_success(from._internal_success());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void swemuDriverVersion_response::CopyFrom(const swemuDriverVersion_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:swemuDriverVersion_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool swemuDriverVersion_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* swemuDriverVersion_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void swemuDriverVersion_response::InternalSwap(swemuDriverVersion_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.success_, other->_impl_.success_);
}

::google::protobuf::Metadata swemuDriverVersion_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[64]);
}
// ===================================================================

class xclPerfMonReadCounters_Streaming_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_Streaming_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_call, _impl_._has_bits_);
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_Streaming_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_Streaming_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        slotname_(arena, from.slotname_) {}

xclPerfMonReadCounters_Streaming_call::xclPerfMonReadCounters_Streaming_call(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadCounters_Streaming_call& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadCounters_Streaming_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadCounters_Streaming_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_Streaming_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        slotname_(arena) {}

inline void xclPerfMonReadCounters_Streaming_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclPerfMonReadCounters_Streaming_call::~xclPerfMonReadCounters_Streaming_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_Streaming_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadCounters_Streaming_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.slotname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadCounters_Streaming_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_Streaming_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_Streaming_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 54, 2> xclPerfMonReadCounters_Streaming_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadCounters_Streaming_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required string slotname = 1;
    {::_pbi::TcParser::FastSS1,
     {10, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_call, _impl_.slotname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required string slotname = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_call, _impl_.slotname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\45\10\0\0\0\0\0\0"
    "xclPerfMonReadCounters_Streaming_call"
    "slotname"
  }},
};

::uint8_t* xclPerfMonReadCounters_Streaming_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_Streaming_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required string slotname = 1;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_slotname();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclPerfMonReadCounters_Streaming_call.slotname");
    target = stream->WriteStringMaybeAliased(1, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_Streaming_call)
  return target;
}

::size_t xclPerfMonReadCounters_Streaming_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_Streaming_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required string slotname = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                    this->_internal_slotname());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadCounters_Streaming_call::_class_data_ = {
    xclPerfMonReadCounters_Streaming_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadCounters_Streaming_call::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadCounters_Streaming_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_Streaming_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_Streaming_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_Streaming_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_slotname(from._internal_slotname());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_Streaming_call::CopyFrom(const xclPerfMonReadCounters_Streaming_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_Streaming_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadCounters_Streaming_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclPerfMonReadCounters_Streaming_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadCounters_Streaming_call::InternalSwap(xclPerfMonReadCounters_Streaming_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.slotname_, &other->_impl_.slotname_, arena);
}

::google::protobuf::Metadata xclPerfMonReadCounters_Streaming_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[65]);
}
// ===================================================================

class xclPerfMonReadCounters_Streaming_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadCounters_Streaming_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_._has_bits_);
  static void set_has_str_num_tranx(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_str_data_bytes(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_str_busy_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_str_stall_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_str_starve_cycles(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
};

xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadCounters_Streaming_response)
}
xclPerfMonReadCounters_Streaming_response::xclPerfMonReadCounters_Streaming_response(
    ::google::protobuf::Arena* arena, const xclPerfMonReadCounters_Streaming_response& from)
    : xclPerfMonReadCounters_Streaming_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadCounters_Streaming_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPerfMonReadCounters_Streaming_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, str_num_tranx_),
           0,
           offsetof(Impl_, str_starve_cycles_) -
               offsetof(Impl_, str_num_tranx_) +
               sizeof(Impl_::str_starve_cycles_));
}
xclPerfMonReadCounters_Streaming_response::~xclPerfMonReadCounters_Streaming_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadCounters_Streaming_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadCounters_Streaming_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadCounters_Streaming_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadCounters_Streaming_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    ::memset(&_impl_.str_num_tranx_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.str_starve_cycles_) -
        reinterpret_cast<char*>(&_impl_.str_num_tranx_)) + sizeof(_impl_.str_starve_cycles_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadCounters_Streaming_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclPerfMonReadCounters_Streaming_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_._has_bits_),
    0, // no _extensions_
    11, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294965311,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadCounters_Streaming_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint64 str_data_bytes = 8;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadCounters_Streaming_response, _impl_.str_data_bytes_), 1>(),
     {64, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_data_bytes_)}},
    // optional uint64 str_busy_cycles = 9;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadCounters_Streaming_response, _impl_.str_busy_cycles_), 2>(),
     {72, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_busy_cycles_)}},
    // optional uint64 str_stall_cycles = 10;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadCounters_Streaming_response, _impl_.str_stall_cycles_), 3>(),
     {80, 3, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_stall_cycles_)}},
    // optional uint64 str_starve_cycles = 11;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_), 4>(),
     {88, 4, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_)}},
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint64 str_num_tranx = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_), 0>(),
     {56, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint64 str_num_tranx = 7;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 str_data_bytes = 8;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_data_bytes_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 str_busy_cycles = 9;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_busy_cycles_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 str_stall_cycles = 10;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_stall_cycles_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // optional uint64 str_starve_cycles = 11;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPerfMonReadCounters_Streaming_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadCounters_Streaming_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint64 str_num_tranx = 7;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        7, this->_internal_str_num_tranx(), target);
  }

  // optional uint64 str_data_bytes = 8;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        8, this->_internal_str_data_bytes(), target);
  }

  // optional uint64 str_busy_cycles = 9;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        9, this->_internal_str_busy_cycles(), target);
  }

  // optional uint64 str_stall_cycles = 10;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        10, this->_internal_str_stall_cycles(), target);
  }

  // optional uint64 str_starve_cycles = 11;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        11, this->_internal_str_starve_cycles(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadCounters_Streaming_response)
  return target;
}

::size_t xclPerfMonReadCounters_Streaming_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadCounters_Streaming_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // optional uint64 str_num_tranx = 7;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_str_num_tranx());
    }

    // optional uint64 str_data_bytes = 8;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_str_data_bytes());
    }

    // optional uint64 str_busy_cycles = 9;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_str_busy_cycles());
    }

    // optional uint64 str_stall_cycles = 10;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_str_stall_cycles());
    }

    // optional uint64 str_starve_cycles = 11;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_str_starve_cycles());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadCounters_Streaming_response::_class_data_ = {
    xclPerfMonReadCounters_Streaming_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadCounters_Streaming_response::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadCounters_Streaming_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadCounters_Streaming_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadCounters_Streaming_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadCounters_Streaming_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.str_num_tranx_ = from._impl_.str_num_tranx_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.str_data_bytes_ = from._impl_.str_data_bytes_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.str_busy_cycles_ = from._impl_.str_busy_cycles_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.str_stall_cycles_ = from._impl_.str_stall_cycles_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.str_starve_cycles_ = from._impl_.str_starve_cycles_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadCounters_Streaming_response::CopyFrom(const xclPerfMonReadCounters_Streaming_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadCounters_Streaming_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadCounters_Streaming_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadCounters_Streaming_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadCounters_Streaming_response::InternalSwap(xclPerfMonReadCounters_Streaming_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_starve_cycles_)
      + sizeof(xclPerfMonReadCounters_Streaming_response::_impl_.str_starve_cycles_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadCounters_Streaming_response, _impl_.str_num_tranx_)>(
          reinterpret_cast<char*>(&_impl_.str_num_tranx_),
          reinterpret_cast<char*>(&other->_impl_.str_num_tranx_));
}

::google::protobuf::Metadata xclPerfMonReadCounters_Streaming_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[66]);
}
// ===================================================================

class xclPerfMonReadTrace_Streaming_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_Streaming_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_slotname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
};

xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_Streaming_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        slotname_(arena, from.slotname_) {}

xclPerfMonReadTrace_Streaming_call::xclPerfMonReadTrace_Streaming_call(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadTrace_Streaming_call& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadTrace_Streaming_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.ack_ = from._impl_.ack_;

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_Streaming_call)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_Streaming_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        slotname_(arena) {}

inline void xclPerfMonReadTrace_Streaming_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclPerfMonReadTrace_Streaming_call::~xclPerfMonReadTrace_Streaming_call() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_Streaming_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.slotname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_Streaming_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.slotname_.ClearNonDefaultToEmpty();
  }
  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 51, 2> xclPerfMonReadTrace_Streaming_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadTrace_Streaming_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional string slotname = 2;
    {::_pbi::TcParser::FastSS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_.slotname_)}},
    // optional bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclPerfMonReadTrace_Streaming_call, _impl_.ack_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_.ack_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // optional string slotname = 2;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_call, _impl_.slotname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kRawString | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
    "\42\0\10\0\0\0\0\0"
    "xclPerfMonReadTrace_Streaming_call"
    "slotname"
  }},
};

::uint8_t* xclPerfMonReadTrace_Streaming_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional bool ack = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  // optional string slotname = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_slotname();
    ::google::protobuf::internal::WireFormat::VerifyUTF8StringNamedField(_s.data(), static_cast<int>(_s.length()), ::google::protobuf::internal::WireFormat::SERIALIZE,
                                "xclPerfMonReadTrace_Streaming_call.slotname");
    target = stream->WriteStringMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_call)
  return target;
}

::size_t xclPerfMonReadTrace_Streaming_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // optional string slotname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::StringSize(
                                      this->_internal_slotname());
    }

    // optional bool ack = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_Streaming_call::_class_data_ = {
    xclPerfMonReadTrace_Streaming_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_Streaming_call::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_Streaming_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_call*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_slotname(from._internal_slotname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.ack_ = from._impl_.ack_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_call::CopyFrom(const xclPerfMonReadTrace_Streaming_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_Streaming_call::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_Streaming_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_Streaming_call::InternalSwap(xclPerfMonReadTrace_Streaming_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.slotname_, &other->_impl_.slotname_, arena);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclPerfMonReadTrace_Streaming_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[67]);
}
// ===================================================================

class xclPerfMonReadTrace_Streaming_response_events::_Internal {
 public:
  using HasBits = decltype(std::declval<xclPerfMonReadTrace_Streaming_response_events>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_._has_bits_);
  static void set_has_eventflags(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_host_timestamp(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
};

xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_response.events)
}
xclPerfMonReadTrace_Streaming_response_events::xclPerfMonReadTrace_Streaming_response_events(
    ::google::protobuf::Arena* arena, const xclPerfMonReadTrace_Streaming_response_events& from)
    : xclPerfMonReadTrace_Streaming_response_events(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_Streaming_response_events::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclPerfMonReadTrace_Streaming_response_events::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, eventflags_),
           0,
           offsetof(Impl_, host_timestamp_) -
               offsetof(Impl_, eventflags_) +
               sizeof(Impl_::host_timestamp_));
}
xclPerfMonReadTrace_Streaming_response_events::~xclPerfMonReadTrace_Streaming_response_events() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_response.events)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_Streaming_response_events::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_Streaming_response_events::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_response.events)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    ::memset(&_impl_.eventflags_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.host_timestamp_) -
        reinterpret_cast<char*>(&_impl_.eventflags_)) + sizeof(_impl_.host_timestamp_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_response_events::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 0, 2> xclPerfMonReadTrace_Streaming_response_events::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_._has_bits_),
    0, // no _extensions_
    5, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967267,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclPerfMonReadTrace_Streaming_response_events_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // optional uint32 timestamp = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_Streaming_response_events, _impl_.timestamp_), 1>(),
     {32, 1, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.timestamp_)}},
    // optional uint64 host_timestamp = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_), 2>(),
     {40, 2, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_)}},
    {::_pbi::TcParser::MiniParse, {}},
    // optional uint32 eventflags = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_), 0>(),
     {24, 0, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_)}},
  }}, {{
    65535, 65535
  }}, {{
    // optional uint32 eventflags = 3;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint32 timestamp = 4;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.timestamp_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // optional uint64 host_timestamp = 5;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclPerfMonReadTrace_Streaming_response_events::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_response.events)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // optional uint32 eventflags = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_eventflags(), target);
  }

  // optional uint32 timestamp = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        4, this->_internal_timestamp(), target);
  }

  // optional uint64 host_timestamp = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_host_timestamp(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_response.events)
  return target;
}

::size_t xclPerfMonReadTrace_Streaming_response_events::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_response.events)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // optional uint32 eventflags = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_eventflags());
    }

    // optional uint32 timestamp = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_timestamp());
    }

    // optional uint64 host_timestamp = 5;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_host_timestamp());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_Streaming_response_events::_class_data_ = {
    xclPerfMonReadTrace_Streaming_response_events::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_Streaming_response_events::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_Streaming_response_events::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_response_events*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_response_events&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_response.events)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.eventflags_ = from._impl_.eventflags_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.timestamp_ = from._impl_.timestamp_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.host_timestamp_ = from._impl_.host_timestamp_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_response_events::CopyFrom(const xclPerfMonReadTrace_Streaming_response_events& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_response.events)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_Streaming_response_events::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_Streaming_response_events::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_Streaming_response_events::InternalSwap(xclPerfMonReadTrace_Streaming_response_events* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.host_timestamp_)
      + sizeof(xclPerfMonReadTrace_Streaming_response_events::_impl_.host_timestamp_)
      - PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response_events, _impl_.eventflags_)>(
          reinterpret_cast<char*>(&_impl_.eventflags_),
          reinterpret_cast<char*>(&other->_impl_.eventflags_));
}

::google::protobuf::Metadata xclPerfMonReadTrace_Streaming_response_events::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[68]);
}
// ===================================================================

class xclPerfMonReadTrace_Streaming_response::_Internal {
 public:
};

xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclPerfMonReadTrace_Streaming_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_Streaming_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : output_data_{visibility, arena, from.output_data_},
        _cached_size_{0} {}

xclPerfMonReadTrace_Streaming_response::xclPerfMonReadTrace_Streaming_response(
    ::google::protobuf::Arena* arena,
    const xclPerfMonReadTrace_Streaming_response& from)
    : ::google::protobuf::Message(arena) {
  xclPerfMonReadTrace_Streaming_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclPerfMonReadTrace_Streaming_response)
}
inline PROTOBUF_NDEBUG_INLINE xclPerfMonReadTrace_Streaming_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : output_data_{visibility, arena},
        _cached_size_{0} {}

inline void xclPerfMonReadTrace_Streaming_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclPerfMonReadTrace_Streaming_response::~xclPerfMonReadTrace_Streaming_response() {
  // @@protoc_insertion_point(destructor:xclPerfMonReadTrace_Streaming_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclPerfMonReadTrace_Streaming_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclPerfMonReadTrace_Streaming_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclPerfMonReadTrace_Streaming_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.output_data_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclPerfMonReadTrace_Streaming_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 1, 0, 2> xclPerfMonReadTrace_Streaming_response::_table_ = {
  {
    0,  // no _has_bits_
    0, // no _extensions_
    8, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967167,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    1,  // num_aux_entries
    offsetof(decltype(_table_), aux_entries),
    &_xclPerfMonReadTrace_Streaming_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
    {::_pbi::TcParser::FastMtR1,
     {66, 63, 0, PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response, _impl_.output_data_)}},
  }}, {{
    65535, 65535
  }}, {{
    // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
    {PROTOBUF_FIELD_OFFSET(xclPerfMonReadTrace_Streaming_response, _impl_.output_data_), 0, 0,
    (0 | ::_fl::kFcRepeated | ::_fl::kMessage | ::_fl::kTvTable)},
  }}, {{
    {::_pbi::TcParser::GetTable<::xclPerfMonReadTrace_Streaming_response_events>()},
  }}, {{
  }},
};

::uint8_t* xclPerfMonReadTrace_Streaming_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclPerfMonReadTrace_Streaming_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
  for (unsigned i = 0,
      n = static_cast<unsigned>(this->_internal_output_data_size()); i < n; i++) {
    const auto& repfield = this->_internal_output_data().Get(i);
    target = ::google::protobuf::internal::WireFormatLite::
        InternalWriteMessage(8, repfield, repfield.GetCachedSize(), target, stream);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclPerfMonReadTrace_Streaming_response)
  return target;
}

::size_t xclPerfMonReadTrace_Streaming_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclPerfMonReadTrace_Streaming_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // repeated .xclPerfMonReadTrace_Streaming_response.events output_data = 8;
  total_size += 1UL * this->_internal_output_data_size();
  for (const auto& msg : this->_internal_output_data()) {
    total_size +=
      ::google::protobuf::internal::WireFormatLite::MessageSize(msg);
  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclPerfMonReadTrace_Streaming_response::_class_data_ = {
    xclPerfMonReadTrace_Streaming_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclPerfMonReadTrace_Streaming_response::GetClassData() const {
  return &_class_data_;
}

void xclPerfMonReadTrace_Streaming_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclPerfMonReadTrace_Streaming_response*>(&to_msg);
  auto& from = static_cast<const xclPerfMonReadTrace_Streaming_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclPerfMonReadTrace_Streaming_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  _this->_internal_mutable_output_data()->MergeFrom(
      from._internal_output_data());
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclPerfMonReadTrace_Streaming_response::CopyFrom(const xclPerfMonReadTrace_Streaming_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclPerfMonReadTrace_Streaming_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclPerfMonReadTrace_Streaming_response::IsInitialized() const {
  return true;
}

::_pbi::CachedSize* xclPerfMonReadTrace_Streaming_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclPerfMonReadTrace_Streaming_response::InternalSwap(xclPerfMonReadTrace_Streaming_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  _impl_.output_data_.InternalSwap(&other->_impl_.output_data_);
}

::google::protobuf::Metadata xclPerfMonReadTrace_Streaming_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[69]);
}
// ===================================================================

class xclSlaveReadReq_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveReadReq_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclSlaveReadReq_call::xclSlaveReadReq_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSlaveReadReq_call)
}
xclSlaveReadReq_call::xclSlaveReadReq_call(
    ::google::protobuf::Arena* arena, const xclSlaveReadReq_call& from)
    : xclSlaveReadReq_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveReadReq_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclSlaveReadReq_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, size_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::size_));
}
xclSlaveReadReq_call::~xclSlaveReadReq_call() {
  // @@protoc_insertion_point(destructor:xclSlaveReadReq_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSlaveReadReq_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSlaveReadReq_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveReadReq_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSlaveReadReq_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclSlaveReadReq_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSlaveReadReq_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclSlaveReadReq_call, _impl_.size_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.size_)}},
    // required uint64 addr = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSlaveReadReq_call, _impl_.addr_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.addr_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 1;
    {PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.addr_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSlaveReadReq_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveReadReq_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_addr(), target);
  }

  // required uint32 size = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveReadReq_call)
  return target;
}

::size_t xclSlaveReadReq_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveReadReq_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required uint64 addr = 1;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSlaveReadReq_call::_class_data_ = {
    xclSlaveReadReq_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSlaveReadReq_call::GetClassData() const {
  return &_class_data_;
}

void xclSlaveReadReq_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveReadReq_call*>(&to_msg);
  auto& from = static_cast<const xclSlaveReadReq_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveReadReq_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveReadReq_call::CopyFrom(const xclSlaveReadReq_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveReadReq_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSlaveReadReq_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSlaveReadReq_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSlaveReadReq_call::InternalSwap(xclSlaveReadReq_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.size_)
      + sizeof(xclSlaveReadReq_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclSlaveReadReq_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[70]);
}
// ===================================================================

class xclSlaveReadReq_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveReadReq_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclSlaveReadReq_response::xclSlaveReadReq_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSlaveReadReq_response)
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveReadReq_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclSlaveReadReq_response::xclSlaveReadReq_response(
    ::google::protobuf::Arena* arena,
    const xclSlaveReadReq_response& from)
    : ::google::protobuf::Message(arena) {
  xclSlaveReadReq_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.valid_ = from._impl_.valid_;

  // @@protoc_insertion_point(copy_constructor:xclSlaveReadReq_response)
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveReadReq_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclSlaveReadReq_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclSlaveReadReq_response::~xclSlaveReadReq_response() {
  // @@protoc_insertion_point(destructor:xclSlaveReadReq_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSlaveReadReq_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSlaveReadReq_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveReadReq_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSlaveReadReq_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclSlaveReadReq_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSlaveReadReq_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes data = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_.data_)}},
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclSlaveReadReq_response, _impl_.valid_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_.valid_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
    // required bytes data = 2;
    {PROTOBUF_FIELD_OFFSET(xclSlaveReadReq_response, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSlaveReadReq_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveReadReq_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  // required bytes data = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveReadReq_response)
  return target;
}

::size_t xclSlaveReadReq_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveReadReq_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required bytes data = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_data());
    }

    // required bool valid = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSlaveReadReq_response::_class_data_ = {
    xclSlaveReadReq_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSlaveReadReq_response::GetClassData() const {
  return &_class_data_;
}

void xclSlaveReadReq_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveReadReq_response*>(&to_msg);
  auto& from = static_cast<const xclSlaveReadReq_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveReadReq_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.valid_ = from._impl_.valid_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveReadReq_response::CopyFrom(const xclSlaveReadReq_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveReadReq_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSlaveReadReq_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSlaveReadReq_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSlaveReadReq_response::InternalSwap(xclSlaveReadReq_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclSlaveReadReq_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[71]);
}
// ===================================================================

class xclSlaveWriteReq_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveWriteReq_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_._has_bits_);
  static void set_has_addr(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_data(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000007) ^ 0x00000007) != 0;
  }
};

xclSlaveWriteReq_call::xclSlaveWriteReq_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSlaveWriteReq_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveWriteReq_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        data_(arena, from.data_) {}

xclSlaveWriteReq_call::xclSlaveWriteReq_call(
    ::google::protobuf::Arena* arena,
    const xclSlaveWriteReq_call& from)
    : ::google::protobuf::Message(arena) {
  xclSlaveWriteReq_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, addr_),
           offsetof(Impl_, size_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::size_));

  // @@protoc_insertion_point(copy_constructor:xclSlaveWriteReq_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveWriteReq_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        data_(arena) {}

inline void xclSlaveWriteReq_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, addr_),
           0,
           offsetof(Impl_, size_) -
               offsetof(Impl_, addr_) +
               sizeof(Impl_::size_));
}
xclSlaveWriteReq_call::~xclSlaveWriteReq_call() {
  // @@protoc_insertion_point(destructor:xclSlaveWriteReq_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSlaveWriteReq_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.data_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSlaveWriteReq_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveWriteReq_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.data_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x00000006u) {
    ::memset(&_impl_.addr_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.size_) -
        reinterpret_cast<char*>(&_impl_.addr_)) + sizeof(_impl_.size_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSlaveWriteReq_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 3, 0, 0, 2> xclSlaveWriteReq_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967288,  // skipmap
    offsetof(decltype(_table_), field_entries),
    3,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSlaveWriteReq_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    // required uint64 addr = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSlaveWriteReq_call, _impl_.addr_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.addr_)}},
    // required uint32 size = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclSlaveWriteReq_call, _impl_.size_), 2>(),
     {16, 2, 0, PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.size_)}},
    // required bytes data = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.data_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 addr = 1;
    {PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.addr_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint32 size = 2;
    {PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes data = 3;
    {PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.data_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSlaveWriteReq_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveWriteReq_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 addr = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_addr(), target);
  }

  // required uint32 size = 2;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_size(), target);
  }

  // required bytes data = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_data();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveWriteReq_call)
  return target;
}

::size_t xclSlaveWriteReq_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveWriteReq_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    // required bytes data = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_data());
    }

    // required uint64 addr = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_addr());
    }

    // required uint32 size = 2;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSlaveWriteReq_call::_class_data_ = {
    xclSlaveWriteReq_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSlaveWriteReq_call::GetClassData() const {
  return &_class_data_;
}

void xclSlaveWriteReq_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveWriteReq_call*>(&to_msg);
  auto& from = static_cast<const xclSlaveWriteReq_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveWriteReq_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_data(from._internal_data());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.addr_ = from._impl_.addr_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveWriteReq_call::CopyFrom(const xclSlaveWriteReq_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveWriteReq_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSlaveWriteReq_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSlaveWriteReq_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSlaveWriteReq_call::InternalSwap(xclSlaveWriteReq_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.data_, &other->_impl_.data_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.size_)
      + sizeof(xclSlaveWriteReq_call::_impl_.size_)
      - PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_call, _impl_.addr_)>(
          reinterpret_cast<char*>(&_impl_.addr_),
          reinterpret_cast<char*>(&other->_impl_.addr_));
}

::google::protobuf::Metadata xclSlaveWriteReq_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[72]);
}
// ===================================================================

class xclSlaveWriteReq_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSlaveWriteReq_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclSlaveWriteReq_response::xclSlaveWriteReq_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSlaveWriteReq_response)
}
xclSlaveWriteReq_response::xclSlaveWriteReq_response(
    ::google::protobuf::Arena* arena, const xclSlaveWriteReq_response& from)
    : xclSlaveWriteReq_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclSlaveWriteReq_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclSlaveWriteReq_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclSlaveWriteReq_response::~xclSlaveWriteReq_response() {
  // @@protoc_insertion_point(destructor:xclSlaveWriteReq_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSlaveWriteReq_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSlaveWriteReq_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSlaveWriteReq_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSlaveWriteReq_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclSlaveWriteReq_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSlaveWriteReq_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclSlaveWriteReq_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclSlaveWriteReq_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSlaveWriteReq_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSlaveWriteReq_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSlaveWriteReq_response)
  return target;
}

::size_t xclSlaveWriteReq_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSlaveWriteReq_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSlaveWriteReq_response::_class_data_ = {
    xclSlaveWriteReq_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSlaveWriteReq_response::GetClassData() const {
  return &_class_data_;
}

void xclSlaveWriteReq_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSlaveWriteReq_response*>(&to_msg);
  auto& from = static_cast<const xclSlaveWriteReq_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSlaveWriteReq_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSlaveWriteReq_response::CopyFrom(const xclSlaveWriteReq_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSlaveWriteReq_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSlaveWriteReq_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSlaveWriteReq_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSlaveWriteReq_response::InternalSwap(xclSlaveWriteReq_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclSlaveWriteReq_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[73]);
}
// ===================================================================

class xclInterruptOccured_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclInterruptOccured_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclInterruptOccured_call, _impl_._has_bits_);
  static void set_has_interrupt_line(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclInterruptOccured_call::xclInterruptOccured_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclInterruptOccured_call)
}
xclInterruptOccured_call::xclInterruptOccured_call(
    ::google::protobuf::Arena* arena, const xclInterruptOccured_call& from)
    : xclInterruptOccured_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclInterruptOccured_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclInterruptOccured_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.interrupt_line_ = {};
}
xclInterruptOccured_call::~xclInterruptOccured_call() {
  // @@protoc_insertion_point(destructor:xclInterruptOccured_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclInterruptOccured_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclInterruptOccured_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclInterruptOccured_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.interrupt_line_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclInterruptOccured_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclInterruptOccured_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclInterruptOccured_call, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclInterruptOccured_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 interrupt_line = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclInterruptOccured_call, _impl_.interrupt_line_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclInterruptOccured_call, _impl_.interrupt_line_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 interrupt_line = 1;
    {PROTOBUF_FIELD_OFFSET(xclInterruptOccured_call, _impl_.interrupt_line_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclInterruptOccured_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclInterruptOccured_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 interrupt_line = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_interrupt_line(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclInterruptOccured_call)
  return target;
}

::size_t xclInterruptOccured_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclInterruptOccured_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint32 interrupt_line = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_interrupt_line());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclInterruptOccured_call::_class_data_ = {
    xclInterruptOccured_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclInterruptOccured_call::GetClassData() const {
  return &_class_data_;
}

void xclInterruptOccured_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclInterruptOccured_call*>(&to_msg);
  auto& from = static_cast<const xclInterruptOccured_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclInterruptOccured_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_interrupt_line(from._internal_interrupt_line());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclInterruptOccured_call::CopyFrom(const xclInterruptOccured_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclInterruptOccured_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclInterruptOccured_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclInterruptOccured_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclInterruptOccured_call::InternalSwap(xclInterruptOccured_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.interrupt_line_, other->_impl_.interrupt_line_);
}

::google::protobuf::Metadata xclInterruptOccured_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[74]);
}
// ===================================================================

class xclInterruptOccured_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclInterruptOccured_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclInterruptOccured_response, _impl_._has_bits_);
  static void set_has_valid(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclInterruptOccured_response::xclInterruptOccured_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclInterruptOccured_response)
}
xclInterruptOccured_response::xclInterruptOccured_response(
    ::google::protobuf::Arena* arena, const xclInterruptOccured_response& from)
    : xclInterruptOccured_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclInterruptOccured_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclInterruptOccured_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.valid_ = {};
}
xclInterruptOccured_response::~xclInterruptOccured_response() {
  // @@protoc_insertion_point(destructor:xclInterruptOccured_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclInterruptOccured_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclInterruptOccured_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclInterruptOccured_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.valid_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclInterruptOccured_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclInterruptOccured_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclInterruptOccured_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclInterruptOccured_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool valid = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclInterruptOccured_response, _impl_.valid_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclInterruptOccured_response, _impl_.valid_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool valid = 1;
    {PROTOBUF_FIELD_OFFSET(xclInterruptOccured_response, _impl_.valid_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclInterruptOccured_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclInterruptOccured_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool valid = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_valid(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclInterruptOccured_response)
  return target;
}

::size_t xclInterruptOccured_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclInterruptOccured_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool valid = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclInterruptOccured_response::_class_data_ = {
    xclInterruptOccured_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclInterruptOccured_response::GetClassData() const {
  return &_class_data_;
}

void xclInterruptOccured_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclInterruptOccured_response*>(&to_msg);
  auto& from = static_cast<const xclInterruptOccured_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclInterruptOccured_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_valid(from._internal_valid());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclInterruptOccured_response::CopyFrom(const xclInterruptOccured_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclInterruptOccured_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclInterruptOccured_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclInterruptOccured_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclInterruptOccured_response::InternalSwap(xclInterruptOccured_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.valid_, other->_impl_.valid_);
}

::google::protobuf::Metadata xclInterruptOccured_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[75]);
}
// ===================================================================

class xclGraphInit_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphInit_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_graphname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphInit_call::xclGraphInit_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphInit_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphInit_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        graphname_(arena, from.graphname_) {}

xclGraphInit_call::xclGraphInit_call(
    ::google::protobuf::Arena* arena,
    const xclGraphInit_call& from)
    : ::google::protobuf::Message(arena) {
  xclGraphInit_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.gh_ = from._impl_.gh_;

  // @@protoc_insertion_point(copy_constructor:xclGraphInit_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphInit_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        graphname_(arena) {}

inline void xclGraphInit_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.gh_ = {};
}
xclGraphInit_call::~xclGraphInit_call() {
  // @@protoc_insertion_point(destructor:xclGraphInit_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphInit_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.graphname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphInit_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphInit_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.graphname_.ClearNonDefaultToEmpty();
  }
  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphInit_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGraphInit_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphInit_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphInit_call, _impl_.gh_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_.gh_)}},
    // required bytes graphname = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_.graphname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_.gh_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes graphname = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphInit_call, _impl_.graphname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphInit_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphInit_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required bytes graphname = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_graphname();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphInit_call)
  return target;
}

::size_t xclGraphInit_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphInit_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required bytes graphname = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_graphname());
    }

    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphInit_call::_class_data_ = {
    xclGraphInit_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphInit_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphInit_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphInit_call*>(&to_msg);
  auto& from = static_cast<const xclGraphInit_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphInit_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_graphname(from._internal_graphname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphInit_call::CopyFrom(const xclGraphInit_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphInit_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphInit_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphInit_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphInit_call::InternalSwap(xclGraphInit_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.graphname_, &other->_impl_.graphname_, arena);
        swap(_impl_.gh_, other->_impl_.gh_);
}

::google::protobuf::Metadata xclGraphInit_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[76]);
}
// ===================================================================

class xclGraphInit_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphInit_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphInit_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphInit_response::xclGraphInit_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphInit_response)
}
xclGraphInit_response::xclGraphInit_response(
    ::google::protobuf::Arena* arena, const xclGraphInit_response& from)
    : xclGraphInit_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphInit_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphInit_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphInit_response::~xclGraphInit_response() {
  // @@protoc_insertion_point(destructor:xclGraphInit_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphInit_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphInit_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphInit_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphInit_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphInit_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphInit_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphInit_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphInit_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphInit_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphInit_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphInit_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphInit_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphInit_response)
  return target;
}

::size_t xclGraphInit_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphInit_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphInit_response::_class_data_ = {
    xclGraphInit_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphInit_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphInit_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphInit_response*>(&to_msg);
  auto& from = static_cast<const xclGraphInit_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphInit_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphInit_response::CopyFrom(const xclGraphInit_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphInit_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphInit_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphInit_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphInit_response::InternalSwap(xclGraphInit_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphInit_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[77]);
}
// ===================================================================

class xclGraphRun_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphRun_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_iterations(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphRun_call::xclGraphRun_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphRun_call)
}
xclGraphRun_call::xclGraphRun_call(
    ::google::protobuf::Arena* arena, const xclGraphRun_call& from)
    : xclGraphRun_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphRun_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphRun_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, gh_),
           0,
           offsetof(Impl_, iterations_) -
               offsetof(Impl_, gh_) +
               sizeof(Impl_::iterations_));
}
xclGraphRun_call::~xclGraphRun_call() {
  // @@protoc_insertion_point(destructor:xclGraphRun_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphRun_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphRun_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphRun_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.gh_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.iterations_) -
        reinterpret_cast<char*>(&_impl_.gh_)) + sizeof(_impl_.iterations_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphRun_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGraphRun_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphRun_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphRun_call, _impl_.gh_), 0>(),
     {16, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.gh_)}},
    // required uint32 iterations = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphRun_call, _impl_.iterations_), 1>(),
     {24, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.iterations_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.gh_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint32 iterations = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.iterations_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphRun_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphRun_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required uint32 iterations = 3;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_iterations(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphRun_call)
  return target;
}

::size_t xclGraphRun_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphRun_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

    // required uint32 iterations = 3;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_iterations());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphRun_call::_class_data_ = {
    xclGraphRun_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphRun_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphRun_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphRun_call*>(&to_msg);
  auto& from = static_cast<const xclGraphRun_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphRun_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.iterations_ = from._impl_.iterations_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphRun_call::CopyFrom(const xclGraphRun_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphRun_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphRun_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphRun_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphRun_call::InternalSwap(xclGraphRun_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.iterations_)
      + sizeof(xclGraphRun_call::_impl_.iterations_)
      - PROTOBUF_FIELD_OFFSET(xclGraphRun_call, _impl_.gh_)>(
          reinterpret_cast<char*>(&_impl_.gh_),
          reinterpret_cast<char*>(&other->_impl_.gh_));
}

::google::protobuf::Metadata xclGraphRun_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[78]);
}
// ===================================================================

class xclGraphRun_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphRun_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphRun_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphRun_response::xclGraphRun_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphRun_response)
}
xclGraphRun_response::xclGraphRun_response(
    ::google::protobuf::Arena* arena, const xclGraphRun_response& from)
    : xclGraphRun_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphRun_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphRun_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphRun_response::~xclGraphRun_response() {
  // @@protoc_insertion_point(destructor:xclGraphRun_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphRun_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphRun_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphRun_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphRun_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphRun_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphRun_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphRun_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphRun_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphRun_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphRun_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphRun_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphRun_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphRun_response)
  return target;
}

::size_t xclGraphRun_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphRun_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphRun_response::_class_data_ = {
    xclGraphRun_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphRun_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphRun_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphRun_response*>(&to_msg);
  auto& from = static_cast<const xclGraphRun_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphRun_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphRun_response::CopyFrom(const xclGraphRun_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphRun_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphRun_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphRun_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphRun_response::InternalSwap(xclGraphRun_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphRun_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[79]);
}
// ===================================================================

class xclGraphWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphWait_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphWait_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphWait_call::xclGraphWait_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphWait_call)
}
xclGraphWait_call::xclGraphWait_call(
    ::google::protobuf::Arena* arena, const xclGraphWait_call& from)
    : xclGraphWait_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphWait_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphWait_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.gh_ = {};
}
xclGraphWait_call::~xclGraphWait_call() {
  // @@protoc_insertion_point(destructor:xclGraphWait_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphWait_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphWait_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphWait_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphWait_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphWait_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967293,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphWait_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphWait_call, _impl_.gh_), 0>(),
     {16, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphWait_call, _impl_.gh_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphWait_call, _impl_.gh_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphWait_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphWait_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphWait_call)
  return target;
}

::size_t xclGraphWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphWait_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint32 gh = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_gh());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphWait_call::_class_data_ = {
    xclGraphWait_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphWait_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphWait_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphWait_call*>(&to_msg);
  auto& from = static_cast<const xclGraphWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphWait_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphWait_call::CopyFrom(const xclGraphWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphWait_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphWait_call::InternalSwap(xclGraphWait_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.gh_, other->_impl_.gh_);
}

::google::protobuf::Metadata xclGraphWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[80]);
}
// ===================================================================

class xclGraphWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphWait_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphWait_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphWait_response::xclGraphWait_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphWait_response)
}
xclGraphWait_response::xclGraphWait_response(
    ::google::protobuf::Arena* arena, const xclGraphWait_response& from)
    : xclGraphWait_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphWait_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphWait_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphWait_response::~xclGraphWait_response() {
  // @@protoc_insertion_point(destructor:xclGraphWait_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphWait_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphWait_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphWait_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphWait_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphWait_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphWait_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphWait_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphWait_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphWait_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphWait_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphWait_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphWait_response)
  return target;
}

::size_t xclGraphWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphWait_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphWait_response::_class_data_ = {
    xclGraphWait_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphWait_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphWait_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphWait_response*>(&to_msg);
  auto& from = static_cast<const xclGraphWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphWait_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphWait_response::CopyFrom(const xclGraphWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphWait_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphWait_response::InternalSwap(xclGraphWait_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[81]);
}
// ===================================================================

class xclGraphEnd_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphEnd_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphEnd_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphEnd_call::xclGraphEnd_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphEnd_call)
}
xclGraphEnd_call::xclGraphEnd_call(
    ::google::protobuf::Arena* arena, const xclGraphEnd_call& from)
    : xclGraphEnd_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphEnd_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphEnd_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.gh_ = {};
}
xclGraphEnd_call::~xclGraphEnd_call() {
  // @@protoc_insertion_point(destructor:xclGraphEnd_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphEnd_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphEnd_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphEnd_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphEnd_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphEnd_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphEnd_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967293,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphEnd_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphEnd_call, _impl_.gh_), 0>(),
     {16, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphEnd_call, _impl_.gh_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphEnd_call, _impl_.gh_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphEnd_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphEnd_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphEnd_call)
  return target;
}

::size_t xclGraphEnd_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphEnd_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint32 gh = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_gh());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphEnd_call::_class_data_ = {
    xclGraphEnd_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphEnd_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphEnd_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphEnd_call*>(&to_msg);
  auto& from = static_cast<const xclGraphEnd_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphEnd_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphEnd_call::CopyFrom(const xclGraphEnd_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphEnd_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphEnd_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphEnd_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphEnd_call::InternalSwap(xclGraphEnd_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.gh_, other->_impl_.gh_);
}

::google::protobuf::Metadata xclGraphEnd_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[82]);
}
// ===================================================================

class xclGraphEnd_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphEnd_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphEnd_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphEnd_response::xclGraphEnd_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphEnd_response)
}
xclGraphEnd_response::xclGraphEnd_response(
    ::google::protobuf::Arena* arena, const xclGraphEnd_response& from)
    : xclGraphEnd_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphEnd_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphEnd_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphEnd_response::~xclGraphEnd_response() {
  // @@protoc_insertion_point(destructor:xclGraphEnd_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphEnd_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphEnd_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphEnd_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphEnd_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphEnd_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphEnd_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphEnd_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphEnd_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphEnd_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphEnd_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphEnd_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphEnd_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphEnd_response)
  return target;
}

::size_t xclGraphEnd_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphEnd_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint32 ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_ack());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphEnd_response::_class_data_ = {
    xclGraphEnd_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphEnd_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphEnd_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphEnd_response*>(&to_msg);
  auto& from = static_cast<const xclGraphEnd_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphEnd_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphEnd_response::CopyFrom(const xclGraphEnd_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphEnd_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphEnd_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphEnd_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphEnd_response::InternalSwap(xclGraphEnd_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphEnd_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[83]);
}
// ===================================================================

class xclGraphUpdateRTP_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphUpdateRTP_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_portname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000f) ^ 0x0000000f) != 0;
  }
};

xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphUpdateRTP_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphUpdateRTP_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        portname_(arena, from.portname_),
        buffer_(arena, from.buffer_) {}

xclGraphUpdateRTP_call::xclGraphUpdateRTP_call(
    ::google::protobuf::Arena* arena,
    const xclGraphUpdateRTP_call& from)
    : ::google::protobuf::Message(arena) {
  xclGraphUpdateRTP_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, size_),
           offsetof(Impl_, gh_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::gh_));

  // @@protoc_insertion_point(copy_constructor:xclGraphUpdateRTP_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphUpdateRTP_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        portname_(arena),
        buffer_(arena) {}

inline void xclGraphUpdateRTP_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, gh_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::gh_));
}
xclGraphUpdateRTP_call::~xclGraphUpdateRTP_call() {
  // @@protoc_insertion_point(destructor:xclGraphUpdateRTP_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphUpdateRTP_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.portname_.Destroy();
  _impl_.buffer_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphUpdateRTP_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphUpdateRTP_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.portname_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.buffer_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphUpdateRTP_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 0, 0, 2> xclGraphUpdateRTP_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_._has_bits_),
    0, // no _extensions_
    5, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967265,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphUpdateRTP_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes buffer = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.buffer_)}},
    // required uint64 size = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphUpdateRTP_call, _impl_.size_), 2>(),
     {40, 2, 0, PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.size_)}},
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphUpdateRTP_call, _impl_.gh_), 3>(),
     {16, 3, 0, PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.gh_)}},
    // required bytes portname = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.portname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.gh_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes portname = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.portname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes buffer = 4;
    {PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.buffer_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 size = 5;
    {PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphUpdateRTP_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphUpdateRTP_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required bytes portname = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_portname();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // required bytes buffer = 4;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_buffer();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphUpdateRTP_call)
  return target;
}

::size_t xclGraphUpdateRTP_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphUpdateRTP_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    // required bytes portname = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_portname());
    }

    // required bytes buffer = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_buffer());
    }

    // required uint64 size = 5;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphUpdateRTP_call::_class_data_ = {
    xclGraphUpdateRTP_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphUpdateRTP_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphUpdateRTP_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphUpdateRTP_call*>(&to_msg);
  auto& from = static_cast<const xclGraphUpdateRTP_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphUpdateRTP_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_portname(from._internal_portname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphUpdateRTP_call::CopyFrom(const xclGraphUpdateRTP_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphUpdateRTP_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphUpdateRTP_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphUpdateRTP_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphUpdateRTP_call::InternalSwap(xclGraphUpdateRTP_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.portname_, &other->_impl_.portname_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.buffer_, &other->_impl_.buffer_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.gh_)
      + sizeof(xclGraphUpdateRTP_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata xclGraphUpdateRTP_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[84]);
}
// ===================================================================

class xclGraphUpdateRTP_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphUpdateRTP_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_response, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphUpdateRTP_response)
}
xclGraphUpdateRTP_response::xclGraphUpdateRTP_response(
    ::google::protobuf::Arena* arena, const xclGraphUpdateRTP_response& from)
    : xclGraphUpdateRTP_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphUpdateRTP_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphUpdateRTP_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclGraphUpdateRTP_response::~xclGraphUpdateRTP_response() {
  // @@protoc_insertion_point(destructor:xclGraphUpdateRTP_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphUpdateRTP_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphUpdateRTP_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphUpdateRTP_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphUpdateRTP_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphUpdateRTP_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphUpdateRTP_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint64 size = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphUpdateRTP_response, _impl_.size_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_response, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 size = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphUpdateRTP_response, _impl_.size_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphUpdateRTP_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphUpdateRTP_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphUpdateRTP_response)
  return target;
}

::size_t xclGraphUpdateRTP_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphUpdateRTP_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint64 size = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
        this->_internal_size());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphUpdateRTP_response::_class_data_ = {
    xclGraphUpdateRTP_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphUpdateRTP_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphUpdateRTP_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphUpdateRTP_response*>(&to_msg);
  auto& from = static_cast<const xclGraphUpdateRTP_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphUpdateRTP_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_size(from._internal_size());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphUpdateRTP_response::CopyFrom(const xclGraphUpdateRTP_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphUpdateRTP_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphUpdateRTP_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphUpdateRTP_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphUpdateRTP_response::InternalSwap(xclGraphUpdateRTP_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclGraphUpdateRTP_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[85]);
}
// ===================================================================

class xclGraphReadRTP_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphReadRTP_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_portname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000000f) ^ 0x0000000f) != 0;
  }
};

xclGraphReadRTP_call::xclGraphReadRTP_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphReadRTP_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphReadRTP_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        portname_(arena, from.portname_),
        buffer_(arena, from.buffer_) {}

xclGraphReadRTP_call::xclGraphReadRTP_call(
    ::google::protobuf::Arena* arena,
    const xclGraphReadRTP_call& from)
    : ::google::protobuf::Message(arena) {
  xclGraphReadRTP_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, size_),
           offsetof(Impl_, gh_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::gh_));

  // @@protoc_insertion_point(copy_constructor:xclGraphReadRTP_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphReadRTP_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        portname_(arena),
        buffer_(arena) {}

inline void xclGraphReadRTP_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, gh_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::gh_));
}
xclGraphReadRTP_call::~xclGraphReadRTP_call() {
  // @@protoc_insertion_point(destructor:xclGraphReadRTP_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphReadRTP_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.portname_.Destroy();
  _impl_.buffer_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphReadRTP_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphReadRTP_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.portname_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.buffer_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x0000000cu) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphReadRTP_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<2, 4, 0, 0, 2> xclGraphReadRTP_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_._has_bits_),
    0, // no _extensions_
    5, 24,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967265,  // skipmap
    offsetof(decltype(_table_), field_entries),
    4,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphReadRTP_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes buffer = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.buffer_)}},
    // required uint64 size = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphReadRTP_call, _impl_.size_), 2>(),
     {40, 2, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.size_)}},
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphReadRTP_call, _impl_.gh_), 3>(),
     {16, 3, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.gh_)}},
    // required bytes portname = 3;
    {::_pbi::TcParser::FastBS1,
     {26, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.portname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.gh_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required bytes portname = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.portname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required bytes buffer = 4;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.buffer_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 size = 5;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.size_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphReadRTP_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphReadRTP_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required bytes portname = 3;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_portname();
    target = stream->WriteBytesMaybeAliased(3, _s, target);
  }

  // required bytes buffer = 4;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_buffer();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // required uint64 size = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_size(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphReadRTP_call)
  return target;
}

::size_t xclGraphReadRTP_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphReadRTP_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    // required bytes portname = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_portname());
    }

    // required bytes buffer = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_buffer());
    }

    // required uint64 size = 5;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphReadRTP_call::_class_data_ = {
    xclGraphReadRTP_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphReadRTP_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphReadRTP_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphReadRTP_call*>(&to_msg);
  auto& from = static_cast<const xclGraphReadRTP_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphReadRTP_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000000fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_portname(from._internal_portname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphReadRTP_call::CopyFrom(const xclGraphReadRTP_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphReadRTP_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphReadRTP_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphReadRTP_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphReadRTP_call::InternalSwap(xclGraphReadRTP_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.portname_, &other->_impl_.portname_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.buffer_, &other->_impl_.buffer_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.gh_)
      + sizeof(xclGraphReadRTP_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata xclGraphReadRTP_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[86]);
}
// ===================================================================

class xclGraphReadRTP_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphReadRTP_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_._has_bits_);
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_buffer(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphReadRTP_response::xclGraphReadRTP_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphReadRTP_response)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphReadRTP_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        buffer_(arena, from.buffer_) {}

xclGraphReadRTP_response::xclGraphReadRTP_response(
    ::google::protobuf::Arena* arena,
    const xclGraphReadRTP_response& from)
    : ::google::protobuf::Message(arena) {
  xclGraphReadRTP_response* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  _impl_.size_ = from._impl_.size_;

  // @@protoc_insertion_point(copy_constructor:xclGraphReadRTP_response)
}
inline PROTOBUF_NDEBUG_INLINE xclGraphReadRTP_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        buffer_(arena) {}

inline void xclGraphReadRTP_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.size_ = {};
}
xclGraphReadRTP_response::~xclGraphReadRTP_response() {
  // @@protoc_insertion_point(destructor:xclGraphReadRTP_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphReadRTP_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.buffer_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphReadRTP_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphReadRTP_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.buffer_.ClearNonDefaultToEmpty();
  }
  _impl_.size_ = ::uint64_t{0u};
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphReadRTP_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGraphReadRTP_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_._has_bits_),
    0, // no _extensions_
    2, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967292,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphReadRTP_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes buffer = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_.buffer_)}},
    // required uint64 size = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphReadRTP_response, _impl_.size_), 1>(),
     {8, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_.size_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint64 size = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bytes buffer = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphReadRTP_response, _impl_.buffer_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphReadRTP_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphReadRTP_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint64 size = 1;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        1, this->_internal_size(), target);
  }

  // required bytes buffer = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_buffer();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphReadRTP_response)
  return target;
}

::size_t xclGraphReadRTP_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphReadRTP_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required bytes buffer = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_buffer());
    }

    // required uint64 size = 1;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphReadRTP_response::_class_data_ = {
    xclGraphReadRTP_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphReadRTP_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphReadRTP_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphReadRTP_response*>(&to_msg);
  auto& from = static_cast<const xclGraphReadRTP_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphReadRTP_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_buffer(from._internal_buffer());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphReadRTP_response::CopyFrom(const xclGraphReadRTP_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphReadRTP_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphReadRTP_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphReadRTP_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphReadRTP_response::InternalSwap(xclGraphReadRTP_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.buffer_, &other->_impl_.buffer_, arena);
        swap(_impl_.size_, other->_impl_.size_);
}

::google::protobuf::Metadata xclGraphReadRTP_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[87]);
}
// ===================================================================

class xclSyncBOAIENB_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSyncBOAIENB_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_._has_bits_);
  static void set_has_gmioname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_dir(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_size(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_offset(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_boh(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000001f) ^ 0x0000001f) != 0;
  }
};

xclSyncBOAIENB_call::xclSyncBOAIENB_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSyncBOAIENB_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSyncBOAIENB_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        gmioname_(arena, from.gmioname_) {}

xclSyncBOAIENB_call::xclSyncBOAIENB_call(
    ::google::protobuf::Arena* arena,
    const xclSyncBOAIENB_call& from)
    : ::google::protobuf::Message(arena) {
  xclSyncBOAIENB_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, size_),
           offsetof(Impl_, dir_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::dir_));

  // @@protoc_insertion_point(copy_constructor:xclSyncBOAIENB_call)
}
inline PROTOBUF_NDEBUG_INLINE xclSyncBOAIENB_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        gmioname_(arena) {}

inline void xclSyncBOAIENB_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, size_),
           0,
           offsetof(Impl_, dir_) -
               offsetof(Impl_, size_) +
               sizeof(Impl_::dir_));
}
xclSyncBOAIENB_call::~xclSyncBOAIENB_call() {
  // @@protoc_insertion_point(destructor:xclSyncBOAIENB_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSyncBOAIENB_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.gmioname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSyncBOAIENB_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSyncBOAIENB_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.gmioname_.ClearNonDefaultToEmpty();
  }
  if (cached_has_bits & 0x0000001eu) {
    ::memset(&_impl_.size_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.dir_) -
        reinterpret_cast<char*>(&_impl_.size_)) + sizeof(_impl_.dir_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSyncBOAIENB_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 5, 0, 0, 2> xclSyncBOAIENB_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_._has_bits_),
    0, // no _extensions_
    6, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967233,  // skipmap
    offsetof(decltype(_table_), field_entries),
    5,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSyncBOAIENB_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    {::_pbi::TcParser::MiniParse, {}},
    {::_pbi::TcParser::MiniParse, {}},
    // required bytes gmioname = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.gmioname_)}},
    // required uint32 dir = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclSyncBOAIENB_call, _impl_.dir_), 4>(),
     {24, 4, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.dir_)}},
    // required uint64 size = 4;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSyncBOAIENB_call, _impl_.size_), 1>(),
     {32, 1, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.size_)}},
    // required uint64 offset = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSyncBOAIENB_call, _impl_.offset_), 2>(),
     {40, 2, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.offset_)}},
    // required uint64 boh = 6;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclSyncBOAIENB_call, _impl_.boh_), 3>(),
     {48, 3, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.boh_)}},
    {::_pbi::TcParser::MiniParse, {}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes gmioname = 2;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.gmioname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint32 dir = 3;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.dir_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint64 size = 4;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.size_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 offset = 5;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.offset_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required uint64 boh = 6;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.boh_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSyncBOAIENB_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSyncBOAIENB_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes gmioname = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_gmioname();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // required uint32 dir = 3;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        3, this->_internal_dir(), target);
  }

  // required uint64 size = 4;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        4, this->_internal_size(), target);
  }

  // required uint64 offset = 5;
  if (cached_has_bits & 0x00000004u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_offset(), target);
  }

  // required uint64 boh = 6;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        6, this->_internal_boh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSyncBOAIENB_call)
  return target;
}

::size_t xclSyncBOAIENB_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSyncBOAIENB_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    // required bytes gmioname = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_gmioname());
    }

    // required uint64 size = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_size());
    }

    // required uint64 offset = 5;
    if (cached_has_bits & 0x00000004u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_offset());
    }

    // required uint64 boh = 6;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_boh());
    }

    // required uint32 dir = 3;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_dir());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSyncBOAIENB_call::_class_data_ = {
    xclSyncBOAIENB_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSyncBOAIENB_call::GetClassData() const {
  return &_class_data_;
}

void xclSyncBOAIENB_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSyncBOAIENB_call*>(&to_msg);
  auto& from = static_cast<const xclSyncBOAIENB_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSyncBOAIENB_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000001fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_gmioname(from._internal_gmioname());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.size_ = from._impl_.size_;
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_impl_.offset_ = from._impl_.offset_;
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.boh_ = from._impl_.boh_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.dir_ = from._impl_.dir_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSyncBOAIENB_call::CopyFrom(const xclSyncBOAIENB_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSyncBOAIENB_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSyncBOAIENB_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSyncBOAIENB_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSyncBOAIENB_call::InternalSwap(xclSyncBOAIENB_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.gmioname_, &other->_impl_.gmioname_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.dir_)
      + sizeof(xclSyncBOAIENB_call::_impl_.dir_)
      - PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_call, _impl_.size_)>(
          reinterpret_cast<char*>(&_impl_.size_),
          reinterpret_cast<char*>(&other->_impl_.size_));
}

::google::protobuf::Metadata xclSyncBOAIENB_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[88]);
}
// ===================================================================

class xclSyncBOAIENB_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclSyncBOAIENB_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclSyncBOAIENB_response::xclSyncBOAIENB_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclSyncBOAIENB_response)
}
xclSyncBOAIENB_response::xclSyncBOAIENB_response(
    ::google::protobuf::Arena* arena, const xclSyncBOAIENB_response& from)
    : xclSyncBOAIENB_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclSyncBOAIENB_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclSyncBOAIENB_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclSyncBOAIENB_response::~xclSyncBOAIENB_response() {
  // @@protoc_insertion_point(destructor:xclSyncBOAIENB_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclSyncBOAIENB_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclSyncBOAIENB_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclSyncBOAIENB_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclSyncBOAIENB_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclSyncBOAIENB_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclSyncBOAIENB_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclSyncBOAIENB_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclSyncBOAIENB_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclSyncBOAIENB_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclSyncBOAIENB_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclSyncBOAIENB_response)
  return target;
}

::size_t xclSyncBOAIENB_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclSyncBOAIENB_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclSyncBOAIENB_response::_class_data_ = {
    xclSyncBOAIENB_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclSyncBOAIENB_response::GetClassData() const {
  return &_class_data_;
}

void xclSyncBOAIENB_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclSyncBOAIENB_response*>(&to_msg);
  auto& from = static_cast<const xclSyncBOAIENB_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclSyncBOAIENB_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclSyncBOAIENB_response::CopyFrom(const xclSyncBOAIENB_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclSyncBOAIENB_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclSyncBOAIENB_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclSyncBOAIENB_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclSyncBOAIENB_response::InternalSwap(xclSyncBOAIENB_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclSyncBOAIENB_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[89]);
}
// ===================================================================

class xclGMIOWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGMIOWait_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGMIOWait_call, _impl_._has_bits_);
  static void set_has_gmioname(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGMIOWait_call::xclGMIOWait_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGMIOWait_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGMIOWait_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        gmioname_(arena, from.gmioname_) {}

xclGMIOWait_call::xclGMIOWait_call(
    ::google::protobuf::Arena* arena,
    const xclGMIOWait_call& from)
    : ::google::protobuf::Message(arena) {
  xclGMIOWait_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);

  // @@protoc_insertion_point(copy_constructor:xclGMIOWait_call)
}
inline PROTOBUF_NDEBUG_INLINE xclGMIOWait_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        gmioname_(arena) {}

inline void xclGMIOWait_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
}
xclGMIOWait_call::~xclGMIOWait_call() {
  // @@protoc_insertion_point(destructor:xclGMIOWait_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGMIOWait_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.gmioname_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGMIOWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGMIOWait_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    _impl_.gmioname_.ClearNonDefaultToEmpty();
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGMIOWait_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGMIOWait_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGMIOWait_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967293,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGMIOWait_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bytes gmioname = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclGMIOWait_call, _impl_.gmioname_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes gmioname = 2;
    {PROTOBUF_FIELD_OFFSET(xclGMIOWait_call, _impl_.gmioname_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGMIOWait_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGMIOWait_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes gmioname = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_gmioname();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGMIOWait_call)
  return target;
}

::size_t xclGMIOWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGMIOWait_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bytes gmioname = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                    this->_internal_gmioname());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGMIOWait_call::_class_data_ = {
    xclGMIOWait_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGMIOWait_call::GetClassData() const {
  return &_class_data_;
}

void xclGMIOWait_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGMIOWait_call*>(&to_msg);
  auto& from = static_cast<const xclGMIOWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGMIOWait_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_gmioname(from._internal_gmioname());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGMIOWait_call::CopyFrom(const xclGMIOWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGMIOWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGMIOWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGMIOWait_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGMIOWait_call::InternalSwap(xclGMIOWait_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.gmioname_, &other->_impl_.gmioname_, arena);
}

::google::protobuf::Metadata xclGMIOWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[90]);
}
// ===================================================================

class xclGMIOWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGMIOWait_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGMIOWait_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGMIOWait_response::xclGMIOWait_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGMIOWait_response)
}
xclGMIOWait_response::xclGMIOWait_response(
    ::google::protobuf::Arena* arena, const xclGMIOWait_response& from)
    : xclGMIOWait_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGMIOWait_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGMIOWait_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGMIOWait_response::~xclGMIOWait_response() {
  // @@protoc_insertion_point(destructor:xclGMIOWait_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGMIOWait_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGMIOWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGMIOWait_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGMIOWait_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGMIOWait_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGMIOWait_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGMIOWait_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGMIOWait_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGMIOWait_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGMIOWait_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGMIOWait_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGMIOWait_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGMIOWait_response)
  return target;
}

::size_t xclGMIOWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGMIOWait_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGMIOWait_response::_class_data_ = {
    xclGMIOWait_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGMIOWait_response::GetClassData() const {
  return &_class_data_;
}

void xclGMIOWait_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGMIOWait_response*>(&to_msg);
  auto& from = static_cast<const xclGMIOWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGMIOWait_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGMIOWait_response::CopyFrom(const xclGMIOWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGMIOWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGMIOWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGMIOWait_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGMIOWait_response::InternalSwap(xclGMIOWait_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGMIOWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[91]);
}
// ===================================================================

class xclGraphTimedWait_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedWait_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_cycletimeout(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphTimedWait_call::xclGraphTimedWait_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedWait_call)
}
xclGraphTimedWait_call::xclGraphTimedWait_call(
    ::google::protobuf::Arena* arena, const xclGraphTimedWait_call& from)
    : xclGraphTimedWait_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphTimedWait_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphTimedWait_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, cycletimeout_),
           0,
           offsetof(Impl_, gh_) -
               offsetof(Impl_, cycletimeout_) +
               sizeof(Impl_::gh_));
}
xclGraphTimedWait_call::~xclGraphTimedWait_call() {
  // @@protoc_insertion_point(destructor:xclGraphTimedWait_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphTimedWait_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphTimedWait_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedWait_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.cycletimeout_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphTimedWait_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGraphTimedWait_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphTimedWait_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphTimedWait_call, _impl_.gh_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.gh_)}},
    // required uint64 cycletimeout = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphTimedWait_call, _impl_.cycletimeout_), 0>(),
     {24, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.cycletimeout_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.gh_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint64 cycletimeout = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.cycletimeout_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphTimedWait_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedWait_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required uint64 cycletimeout = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_cycletimeout(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedWait_call)
  return target;
}

::size_t xclGraphTimedWait_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedWait_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required uint64 cycletimeout = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_cycletimeout());
    }

    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphTimedWait_call::_class_data_ = {
    xclGraphTimedWait_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphTimedWait_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphTimedWait_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedWait_call*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedWait_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedWait_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.cycletimeout_ = from._impl_.cycletimeout_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedWait_call::CopyFrom(const xclGraphTimedWait_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedWait_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphTimedWait_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphTimedWait_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphTimedWait_call::InternalSwap(xclGraphTimedWait_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.gh_)
      + sizeof(xclGraphTimedWait_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_call, _impl_.cycletimeout_)>(
          reinterpret_cast<char*>(&_impl_.cycletimeout_),
          reinterpret_cast<char*>(&other->_impl_.cycletimeout_));
}

::google::protobuf::Metadata xclGraphTimedWait_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[92]);
}
// ===================================================================

class xclGraphTimedWait_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedWait_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphTimedWait_response::xclGraphTimedWait_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedWait_response)
}
xclGraphTimedWait_response::xclGraphTimedWait_response(
    ::google::protobuf::Arena* arena, const xclGraphTimedWait_response& from)
    : xclGraphTimedWait_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphTimedWait_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphTimedWait_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphTimedWait_response::~xclGraphTimedWait_response() {
  // @@protoc_insertion_point(destructor:xclGraphTimedWait_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphTimedWait_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphTimedWait_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedWait_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphTimedWait_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphTimedWait_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphTimedWait_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphTimedWait_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedWait_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphTimedWait_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedWait_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedWait_response)
  return target;
}

::size_t xclGraphTimedWait_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedWait_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphTimedWait_response::_class_data_ = {
    xclGraphTimedWait_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphTimedWait_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphTimedWait_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedWait_response*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedWait_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedWait_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedWait_response::CopyFrom(const xclGraphTimedWait_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedWait_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphTimedWait_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphTimedWait_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphTimedWait_response::InternalSwap(xclGraphTimedWait_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphTimedWait_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[93]);
}
// ===================================================================

class xclGraphTimedEnd_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedEnd_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_cycletimeout(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000003) ^ 0x00000003) != 0;
  }
};

xclGraphTimedEnd_call::xclGraphTimedEnd_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedEnd_call)
}
xclGraphTimedEnd_call::xclGraphTimedEnd_call(
    ::google::protobuf::Arena* arena, const xclGraphTimedEnd_call& from)
    : xclGraphTimedEnd_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphTimedEnd_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphTimedEnd_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, cycletimeout_),
           0,
           offsetof(Impl_, gh_) -
               offsetof(Impl_, cycletimeout_) +
               sizeof(Impl_::gh_));
}
xclGraphTimedEnd_call::~xclGraphTimedEnd_call() {
  // @@protoc_insertion_point(destructor:xclGraphTimedEnd_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphTimedEnd_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphTimedEnd_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedEnd_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    ::memset(&_impl_.cycletimeout_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.gh_) -
        reinterpret_cast<char*>(&_impl_.cycletimeout_)) + sizeof(_impl_.gh_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphTimedEnd_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<1, 2, 0, 0, 2> xclGraphTimedEnd_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_._has_bits_),
    0, // no _extensions_
    3, 8,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967289,  // skipmap
    offsetof(decltype(_table_), field_entries),
    2,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphTimedEnd_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphTimedEnd_call, _impl_.gh_), 1>(),
     {16, 1, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.gh_)}},
    // required uint64 cycletimeout = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclGraphTimedEnd_call, _impl_.cycletimeout_), 0>(),
     {24, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.cycletimeout_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.gh_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
    // required uint64 cycletimeout = 3;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.cycletimeout_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphTimedEnd_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedEnd_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000002u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  // required uint64 cycletimeout = 3;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_cycletimeout(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedEnd_call)
  return target;
}

::size_t xclGraphTimedEnd_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedEnd_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    // required uint64 cycletimeout = 3;
    if (cached_has_bits & 0x00000001u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_cycletimeout());
    }

    // required uint32 gh = 2;
    if (cached_has_bits & 0x00000002u) {
      total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
          this->_internal_gh());
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphTimedEnd_call::_class_data_ = {
    xclGraphTimedEnd_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphTimedEnd_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphTimedEnd_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedEnd_call*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedEnd_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedEnd_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x00000003u) {
    if (cached_has_bits & 0x00000001u) {
      _this->_impl_.cycletimeout_ = from._impl_.cycletimeout_;
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_impl_.gh_ = from._impl_.gh_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedEnd_call::CopyFrom(const xclGraphTimedEnd_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedEnd_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphTimedEnd_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphTimedEnd_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphTimedEnd_call::InternalSwap(xclGraphTimedEnd_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.gh_)
      + sizeof(xclGraphTimedEnd_call::_impl_.gh_)
      - PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_call, _impl_.cycletimeout_)>(
          reinterpret_cast<char*>(&_impl_.cycletimeout_),
          reinterpret_cast<char*>(&other->_impl_.cycletimeout_));
}

::google::protobuf::Metadata xclGraphTimedEnd_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[94]);
}
// ===================================================================

class xclGraphTimedEnd_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphTimedEnd_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphTimedEnd_response::xclGraphTimedEnd_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphTimedEnd_response)
}
xclGraphTimedEnd_response::xclGraphTimedEnd_response(
    ::google::protobuf::Arena* arena, const xclGraphTimedEnd_response& from)
    : xclGraphTimedEnd_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphTimedEnd_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphTimedEnd_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphTimedEnd_response::~xclGraphTimedEnd_response() {
  // @@protoc_insertion_point(destructor:xclGraphTimedEnd_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphTimedEnd_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphTimedEnd_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphTimedEnd_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphTimedEnd_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphTimedEnd_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphTimedEnd_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphTimedEnd_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphTimedEnd_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphTimedEnd_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphTimedEnd_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphTimedEnd_response)
  return target;
}

::size_t xclGraphTimedEnd_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphTimedEnd_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphTimedEnd_response::_class_data_ = {
    xclGraphTimedEnd_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphTimedEnd_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphTimedEnd_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphTimedEnd_response*>(&to_msg);
  auto& from = static_cast<const xclGraphTimedEnd_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphTimedEnd_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphTimedEnd_response::CopyFrom(const xclGraphTimedEnd_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphTimedEnd_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphTimedEnd_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphTimedEnd_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphTimedEnd_response::InternalSwap(xclGraphTimedEnd_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphTimedEnd_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[95]);
}
// ===================================================================

class xclGraphResume_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphResume_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphResume_call, _impl_._has_bits_);
  static void set_has_gh(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphResume_call::xclGraphResume_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphResume_call)
}
xclGraphResume_call::xclGraphResume_call(
    ::google::protobuf::Arena* arena, const xclGraphResume_call& from)
    : xclGraphResume_call(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphResume_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphResume_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.gh_ = {};
}
xclGraphResume_call::~xclGraphResume_call() {
  // @@protoc_insertion_point(destructor:xclGraphResume_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphResume_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphResume_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphResume_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.gh_ = 0u;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphResume_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphResume_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphResume_call, _impl_._has_bits_),
    0, // no _extensions_
    2, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967293,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphResume_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required uint32 gh = 2;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint32_t, offsetof(xclGraphResume_call, _impl_.gh_), 0>(),
     {16, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphResume_call, _impl_.gh_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required uint32 gh = 2;
    {PROTOBUF_FIELD_OFFSET(xclGraphResume_call, _impl_.gh_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt32)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphResume_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphResume_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required uint32 gh = 2;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt32ToArray(
        2, this->_internal_gh(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphResume_call)
  return target;
}

::size_t xclGraphResume_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphResume_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required uint32 gh = 2;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += ::_pbi::WireFormatLite::UInt32SizePlusOne(
        this->_internal_gh());
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphResume_call::_class_data_ = {
    xclGraphResume_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphResume_call::GetClassData() const {
  return &_class_data_;
}

void xclGraphResume_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphResume_call*>(&to_msg);
  auto& from = static_cast<const xclGraphResume_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphResume_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_gh(from._internal_gh());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphResume_call::CopyFrom(const xclGraphResume_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphResume_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphResume_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphResume_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphResume_call::InternalSwap(xclGraphResume_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.gh_, other->_impl_.gh_);
}

::google::protobuf::Metadata xclGraphResume_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[96]);
}
// ===================================================================

class xclGraphResume_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclGraphResume_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclGraphResume_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclGraphResume_response::xclGraphResume_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclGraphResume_response)
}
xclGraphResume_response::xclGraphResume_response(
    ::google::protobuf::Arena* arena, const xclGraphResume_response& from)
    : xclGraphResume_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclGraphResume_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclGraphResume_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclGraphResume_response::~xclGraphResume_response() {
  // @@protoc_insertion_point(destructor:xclGraphResume_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclGraphResume_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclGraphResume_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclGraphResume_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclGraphResume_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclGraphResume_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclGraphResume_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclGraphResume_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclGraphResume_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclGraphResume_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclGraphResume_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclGraphResume_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclGraphResume_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclGraphResume_response)
  return target;
}

::size_t xclGraphResume_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclGraphResume_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclGraphResume_response::_class_data_ = {
    xclGraphResume_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclGraphResume_response::GetClassData() const {
  return &_class_data_;
}

void xclGraphResume_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclGraphResume_response*>(&to_msg);
  auto& from = static_cast<const xclGraphResume_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclGraphResume_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclGraphResume_response::CopyFrom(const xclGraphResume_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclGraphResume_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclGraphResume_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclGraphResume_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclGraphResume_response::InternalSwap(xclGraphResume_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclGraphResume_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[97]);
}
// ===================================================================

class xclLoadXclbinContent_call::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadXclbinContent_call>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_._has_bits_);
  static void set_has_xmlbuff(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static void set_has_xmlbuffsize(HasBits* has_bits) {
    (*has_bits)[0] |= 8u;
  }
  static void set_has_sharedbin(HasBits* has_bits) {
    (*has_bits)[0] |= 2u;
  }
  static void set_has_sharedbinsize(HasBits* has_bits) {
    (*has_bits)[0] |= 16u;
  }
  static void set_has_emuldata(HasBits* has_bits) {
    (*has_bits)[0] |= 4u;
  }
  static void set_has_emuldatasize(HasBits* has_bits) {
    (*has_bits)[0] |= 32u;
  }
  static void set_has_keepdir(HasBits* has_bits) {
    (*has_bits)[0] |= 64u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x0000007f) ^ 0x0000007f) != 0;
  }
};

xclLoadXclbinContent_call::xclLoadXclbinContent_call(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclLoadXclbinContent_call)
}
inline PROTOBUF_NDEBUG_INLINE xclLoadXclbinContent_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility, ::google::protobuf::Arena* arena,
    const Impl_& from)
      : _has_bits_{from._has_bits_},
        _cached_size_{0},
        xmlbuff_(arena, from.xmlbuff_),
        sharedbin_(arena, from.sharedbin_),
        emuldata_(arena, from.emuldata_) {}

xclLoadXclbinContent_call::xclLoadXclbinContent_call(
    ::google::protobuf::Arena* arena,
    const xclLoadXclbinContent_call& from)
    : ::google::protobuf::Message(arena) {
  xclLoadXclbinContent_call* const _this = this;
  (void)_this;
  _internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(
      from._internal_metadata_);
  new (&_impl_) Impl_(internal_visibility(), arena, from._impl_);
  ::memcpy(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, xmlbuffsize_),
           reinterpret_cast<const char *>(&from._impl_) +
               offsetof(Impl_, xmlbuffsize_),
           offsetof(Impl_, keepdir_) -
               offsetof(Impl_, xmlbuffsize_) +
               sizeof(Impl_::keepdir_));

  // @@protoc_insertion_point(copy_constructor:xclLoadXclbinContent_call)
}
inline PROTOBUF_NDEBUG_INLINE xclLoadXclbinContent_call::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0},
        xmlbuff_(arena),
        sharedbin_(arena),
        emuldata_(arena) {}

inline void xclLoadXclbinContent_call::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  ::memset(reinterpret_cast<char *>(&_impl_) +
               offsetof(Impl_, xmlbuffsize_),
           0,
           offsetof(Impl_, keepdir_) -
               offsetof(Impl_, xmlbuffsize_) +
               sizeof(Impl_::keepdir_));
}
xclLoadXclbinContent_call::~xclLoadXclbinContent_call() {
  // @@protoc_insertion_point(destructor:xclLoadXclbinContent_call)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclLoadXclbinContent_call::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.xmlbuff_.Destroy();
  _impl_.sharedbin_.Destroy();
  _impl_.emuldata_.Destroy();
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclLoadXclbinContent_call::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadXclbinContent_call)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000007u) {
    if (cached_has_bits & 0x00000001u) {
      _impl_.xmlbuff_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000002u) {
      _impl_.sharedbin_.ClearNonDefaultToEmpty();
    }
    if (cached_has_bits & 0x00000004u) {
      _impl_.emuldata_.ClearNonDefaultToEmpty();
    }
  }
  if (cached_has_bits & 0x00000078u) {
    ::memset(&_impl_.xmlbuffsize_, 0, static_cast<::size_t>(
        reinterpret_cast<char*>(&_impl_.keepdir_) -
        reinterpret_cast<char*>(&_impl_.xmlbuffsize_)) + sizeof(_impl_.keepdir_));
  }
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclLoadXclbinContent_call::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<3, 7, 0, 0, 2> xclLoadXclbinContent_call::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_._has_bits_),
    0, // no _extensions_
    8, 56,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967041,  // skipmap
    offsetof(decltype(_table_), field_entries),
    7,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclLoadXclbinContent_call_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool keepdir = 8;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclLoadXclbinContent_call, _impl_.keepdir_), 6>(),
     {64, 6, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.keepdir_)}},
    {::_pbi::TcParser::MiniParse, {}},
    // required bytes xmlbuff = 2;
    {::_pbi::TcParser::FastBS1,
     {18, 0, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuff_)}},
    // required uint64 xmlbuffsize = 3;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclLoadXclbinContent_call, _impl_.xmlbuffsize_), 3>(),
     {24, 3, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuffsize_)}},
    // required bytes sharedbin = 4;
    {::_pbi::TcParser::FastBS1,
     {34, 1, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.sharedbin_)}},
    // required uint64 sharedbinsize = 5;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclLoadXclbinContent_call, _impl_.sharedbinsize_), 4>(),
     {40, 4, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.sharedbinsize_)}},
    // required bytes emuldata = 6;
    {::_pbi::TcParser::FastBS1,
     {50, 2, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.emuldata_)}},
    // required uint64 emuldatasize = 7;
    {::_pbi::TcParser::SingularVarintNoZag1<::uint64_t, offsetof(xclLoadXclbinContent_call, _impl_.emuldatasize_), 5>(),
     {56, 5, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.emuldatasize_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bytes xmlbuff = 2;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuff_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 xmlbuffsize = 3;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuffsize_), _Internal::kHasBitsOffset + 3, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bytes sharedbin = 4;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.sharedbin_), _Internal::kHasBitsOffset + 1, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 sharedbinsize = 5;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.sharedbinsize_), _Internal::kHasBitsOffset + 4, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bytes emuldata = 6;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.emuldata_), _Internal::kHasBitsOffset + 2, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBytes | ::_fl::kRepAString)},
    // required uint64 emuldatasize = 7;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.emuldatasize_), _Internal::kHasBitsOffset + 5, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kUInt64)},
    // required bool keepdir = 8;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.keepdir_), _Internal::kHasBitsOffset + 6, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclLoadXclbinContent_call::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadXclbinContent_call)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bytes xmlbuff = 2;
  if (cached_has_bits & 0x00000001u) {
    const std::string& _s = this->_internal_xmlbuff();
    target = stream->WriteBytesMaybeAliased(2, _s, target);
  }

  // required uint64 xmlbuffsize = 3;
  if (cached_has_bits & 0x00000008u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        3, this->_internal_xmlbuffsize(), target);
  }

  // required bytes sharedbin = 4;
  if (cached_has_bits & 0x00000002u) {
    const std::string& _s = this->_internal_sharedbin();
    target = stream->WriteBytesMaybeAliased(4, _s, target);
  }

  // required uint64 sharedbinsize = 5;
  if (cached_has_bits & 0x00000010u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        5, this->_internal_sharedbinsize(), target);
  }

  // required bytes emuldata = 6;
  if (cached_has_bits & 0x00000004u) {
    const std::string& _s = this->_internal_emuldata();
    target = stream->WriteBytesMaybeAliased(6, _s, target);
  }

  // required uint64 emuldatasize = 7;
  if (cached_has_bits & 0x00000020u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteUInt64ToArray(
        7, this->_internal_emuldatasize(), target);
  }

  // required bool keepdir = 8;
  if (cached_has_bits & 0x00000040u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        8, this->_internal_keepdir(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadXclbinContent_call)
  return target;
}

::size_t xclLoadXclbinContent_call::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadXclbinContent_call)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    // required bytes xmlbuff = 2;
    if (cached_has_bits & 0x00000001u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_xmlbuff());
    }

    // required bytes sharedbin = 4;
    if (cached_has_bits & 0x00000002u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_sharedbin());
    }

    // required bytes emuldata = 6;
    if (cached_has_bits & 0x00000004u) {
      total_size += 1 + ::google::protobuf::internal::WireFormatLite::BytesSize(
                                      this->_internal_emuldata());
    }

    // required uint64 xmlbuffsize = 3;
    if (cached_has_bits & 0x00000008u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_xmlbuffsize());
    }

    // required uint64 sharedbinsize = 5;
    if (cached_has_bits & 0x00000010u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_sharedbinsize());
    }

    // required uint64 emuldatasize = 7;
    if (cached_has_bits & 0x00000020u) {
      total_size += ::_pbi::WireFormatLite::UInt64SizePlusOne(
          this->_internal_emuldatasize());
    }

    // required bool keepdir = 8;
    if (cached_has_bits & 0x00000040u) {
      total_size += 2;
    }

  }
  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclLoadXclbinContent_call::_class_data_ = {
    xclLoadXclbinContent_call::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclLoadXclbinContent_call::GetClassData() const {
  return &_class_data_;
}

void xclLoadXclbinContent_call::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclLoadXclbinContent_call*>(&to_msg);
  auto& from = static_cast<const xclLoadXclbinContent_call&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadXclbinContent_call)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  cached_has_bits = from._impl_._has_bits_[0];
  if (cached_has_bits & 0x0000007fu) {
    if (cached_has_bits & 0x00000001u) {
      _this->_internal_set_xmlbuff(from._internal_xmlbuff());
    }
    if (cached_has_bits & 0x00000002u) {
      _this->_internal_set_sharedbin(from._internal_sharedbin());
    }
    if (cached_has_bits & 0x00000004u) {
      _this->_internal_set_emuldata(from._internal_emuldata());
    }
    if (cached_has_bits & 0x00000008u) {
      _this->_impl_.xmlbuffsize_ = from._impl_.xmlbuffsize_;
    }
    if (cached_has_bits & 0x00000010u) {
      _this->_impl_.sharedbinsize_ = from._impl_.sharedbinsize_;
    }
    if (cached_has_bits & 0x00000020u) {
      _this->_impl_.emuldatasize_ = from._impl_.emuldatasize_;
    }
    if (cached_has_bits & 0x00000040u) {
      _this->_impl_.keepdir_ = from._impl_.keepdir_;
    }
    _this->_impl_._has_bits_[0] |= cached_has_bits;
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadXclbinContent_call::CopyFrom(const xclLoadXclbinContent_call& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadXclbinContent_call)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclLoadXclbinContent_call::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclLoadXclbinContent_call::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclLoadXclbinContent_call::InternalSwap(xclLoadXclbinContent_call* PROTOBUF_RESTRICT other) {
  using std::swap;
  auto* arena = GetArena();
  ABSL_DCHECK_EQ(arena, other->GetArena());
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.xmlbuff_, &other->_impl_.xmlbuff_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.sharedbin_, &other->_impl_.sharedbin_, arena);
  ::_pbi::ArenaStringPtr::InternalSwap(&_impl_.emuldata_, &other->_impl_.emuldata_, arena);
  ::google::protobuf::internal::memswap<
      PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.keepdir_)
      + sizeof(xclLoadXclbinContent_call::_impl_.keepdir_)
      - PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_call, _impl_.xmlbuffsize_)>(
          reinterpret_cast<char*>(&_impl_.xmlbuffsize_),
          reinterpret_cast<char*>(&other->_impl_.xmlbuffsize_));
}

::google::protobuf::Metadata xclLoadXclbinContent_call::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[98]);
}
// ===================================================================

class xclLoadXclbinContent_response::_Internal {
 public:
  using HasBits = decltype(std::declval<xclLoadXclbinContent_response>()._impl_._has_bits_);
  static constexpr ::int32_t kHasBitsOffset =
    8 * PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_response, _impl_._has_bits_);
  static void set_has_ack(HasBits* has_bits) {
    (*has_bits)[0] |= 1u;
  }
  static bool MissingRequiredFields(const HasBits& has_bits) {
    return ((has_bits[0] & 0x00000001) ^ 0x00000001) != 0;
  }
};

xclLoadXclbinContent_response::xclLoadXclbinContent_response(::google::protobuf::Arena* arena)
    : ::google::protobuf::Message(arena) {
  SharedCtor(arena);
  // @@protoc_insertion_point(arena_constructor:xclLoadXclbinContent_response)
}
xclLoadXclbinContent_response::xclLoadXclbinContent_response(
    ::google::protobuf::Arena* arena, const xclLoadXclbinContent_response& from)
    : xclLoadXclbinContent_response(arena) {
  MergeFrom(from);
}
inline PROTOBUF_NDEBUG_INLINE xclLoadXclbinContent_response::Impl_::Impl_(
    ::google::protobuf::internal::InternalVisibility visibility,
    ::google::protobuf::Arena* arena)
      : _cached_size_{0} {}

inline void xclLoadXclbinContent_response::SharedCtor(::_pb::Arena* arena) {
  new (&_impl_) Impl_(internal_visibility(), arena);
  _impl_.ack_ = {};
}
xclLoadXclbinContent_response::~xclLoadXclbinContent_response() {
  // @@protoc_insertion_point(destructor:xclLoadXclbinContent_response)
  _internal_metadata_.Delete<::google::protobuf::UnknownFieldSet>();
  SharedDtor();
}
inline void xclLoadXclbinContent_response::SharedDtor() {
  ABSL_DCHECK(GetArena() == nullptr);
  _impl_.~Impl_();
}

PROTOBUF_NOINLINE void xclLoadXclbinContent_response::Clear() {
// @@protoc_insertion_point(message_clear_start:xclLoadXclbinContent_response)
  PROTOBUF_TSAN_WRITE(&_impl_._tsan_detect_race);
  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  _impl_.ack_ = false;
  _impl_._has_bits_.Clear();
  _internal_metadata_.Clear<::google::protobuf::UnknownFieldSet>();
}

const char* xclLoadXclbinContent_response::_InternalParse(
    const char* ptr, ::_pbi::ParseContext* ctx) {
  ptr = ::_pbi::TcParser::ParseLoop(this, ptr, ctx, &_table_.header);
  return ptr;
}


PROTOBUF_CONSTINIT PROTOBUF_ATTRIBUTE_INIT_PRIORITY1
const ::_pbi::TcParseTable<0, 1, 0, 0, 2> xclLoadXclbinContent_response::_table_ = {
  {
    PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_response, _impl_._has_bits_),
    0, // no _extensions_
    1, 0,  // max_field_number, fast_idx_mask
    offsetof(decltype(_table_), field_lookup_table),
    4294967294,  // skipmap
    offsetof(decltype(_table_), field_entries),
    1,  // num_field_entries
    0,  // num_aux_entries
    offsetof(decltype(_table_), field_names),  // no aux_entries
    &_xclLoadXclbinContent_response_default_instance_._instance,
    ::_pbi::TcParser::GenericFallback,  // fallback
  }, {{
    // required bool ack = 1;
    {::_pbi::TcParser::SingularVarintNoZag1<bool, offsetof(xclLoadXclbinContent_response, _impl_.ack_), 0>(),
     {8, 0, 0, PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_response, _impl_.ack_)}},
  }}, {{
    65535, 65535
  }}, {{
    // required bool ack = 1;
    {PROTOBUF_FIELD_OFFSET(xclLoadXclbinContent_response, _impl_.ack_), _Internal::kHasBitsOffset + 0, 0,
    (0 | ::_fl::kFcOptional | ::_fl::kBool)},
  }},
  // no aux_entries
  {{
  }},
};

::uint8_t* xclLoadXclbinContent_response::_InternalSerialize(
    ::uint8_t* target,
    ::google::protobuf::io::EpsCopyOutputStream* stream) const {
  // @@protoc_insertion_point(serialize_to_array_start:xclLoadXclbinContent_response)
  ::uint32_t cached_has_bits = 0;
  (void)cached_has_bits;

  cached_has_bits = _impl_._has_bits_[0];
  // required bool ack = 1;
  if (cached_has_bits & 0x00000001u) {
    target = stream->EnsureSpace(target);
    target = ::_pbi::WireFormatLite::WriteBoolToArray(
        1, this->_internal_ack(), target);
  }

  if (PROTOBUF_PREDICT_FALSE(_internal_metadata_.have_unknown_fields())) {
    target =
        ::_pbi::WireFormat::InternalSerializeUnknownFieldsToArray(
            _internal_metadata_.unknown_fields<::google::protobuf::UnknownFieldSet>(::google::protobuf::UnknownFieldSet::default_instance), target, stream);
  }
  // @@protoc_insertion_point(serialize_to_array_end:xclLoadXclbinContent_response)
  return target;
}

::size_t xclLoadXclbinContent_response::ByteSizeLong() const {
// @@protoc_insertion_point(message_byte_size_start:xclLoadXclbinContent_response)
  ::size_t total_size = 0;

  ::uint32_t cached_has_bits = 0;
  // Prevent compiler warnings about cached_has_bits being unused
  (void) cached_has_bits;

  // required bool ack = 1;
  cached_has_bits = _impl_._has_bits_[0];
  if (cached_has_bits & 0x00000001u) {
    total_size += 2;
  }

  return MaybeComputeUnknownFieldsSize(total_size, &_impl_._cached_size_);
}

const ::google::protobuf::Message::ClassData xclLoadXclbinContent_response::_class_data_ = {
    xclLoadXclbinContent_response::MergeImpl,
    nullptr,  // OnDemandRegisterArenaDtor
};
const ::google::protobuf::Message::ClassData* xclLoadXclbinContent_response::GetClassData() const {
  return &_class_data_;
}

void xclLoadXclbinContent_response::MergeImpl(::google::protobuf::Message& to_msg, const ::google::protobuf::Message& from_msg) {
  auto* const _this = static_cast<xclLoadXclbinContent_response*>(&to_msg);
  auto& from = static_cast<const xclLoadXclbinContent_response&>(from_msg);
  // @@protoc_insertion_point(class_specific_merge_from_start:xclLoadXclbinContent_response)
  ABSL_DCHECK_NE(&from, _this);
  ::uint32_t cached_has_bits = 0;
  (void) cached_has_bits;

  if ((from._impl_._has_bits_[0] & 0x00000001u) != 0) {
    _this->_internal_set_ack(from._internal_ack());
  }
  _this->_internal_metadata_.MergeFrom<::google::protobuf::UnknownFieldSet>(from._internal_metadata_);
}

void xclLoadXclbinContent_response::CopyFrom(const xclLoadXclbinContent_response& from) {
// @@protoc_insertion_point(class_specific_copy_from_start:xclLoadXclbinContent_response)
  if (&from == this) return;
  Clear();
  MergeFrom(from);
}

PROTOBUF_NOINLINE bool xclLoadXclbinContent_response::IsInitialized() const {
  if (_Internal::MissingRequiredFields(_impl_._has_bits_)) {
    return false;
  }
  return true;
}

::_pbi::CachedSize* xclLoadXclbinContent_response::AccessCachedSize() const {
  return &_impl_._cached_size_;
}
void xclLoadXclbinContent_response::InternalSwap(xclLoadXclbinContent_response* PROTOBUF_RESTRICT other) {
  using std::swap;
  _internal_metadata_.InternalSwap(&other->_internal_metadata_);
  swap(_impl_._has_bits_[0], other->_impl_._has_bits_[0]);
        swap(_impl_.ack_, other->_impl_.ack_);
}

::google::protobuf::Metadata xclLoadXclbinContent_response::GetMetadata() const {
  return ::_pbi::AssignDescriptors(
      &descriptor_table_rpc_5fmessages_2eproto_getter, &descriptor_table_rpc_5fmessages_2eproto_once,
      file_level_metadata_rpc_5fmessages_2eproto[99]);
}
// @@protoc_insertion_point(namespace_scope)
namespace google {
namespace protobuf {
}  // namespace protobuf
}  // namespace google
// @@protoc_insertion_point(global_scope)
#include "google/protobuf/port_undef.inc"
